#ifndef _ANALYSIS_H
#define _ANALYSIS_H

#include "ldock.H"
#include <map>
using std::map;
#include <vector>
using std::vector;
#include <cstdio>


class elementType;
class prioritizeNodes;

class Path {
public :

  Path(int i, int j, int numberOfNodes, nodeData ** allNodes); 

  /* constructs the first node of the tree, 
     which has the best path from node i to node j.
  */
  Path(int deviatedEdgeIndex); /* 
				  almost same as above.
				  except : start and endIndex are now known.
				  deviatedEdgeIndex set by user.
			       */
  Path();                      // same as above. everything set by default.
    

  //  Path(float _pathWeight, int _numIndices, int * _indices,nodeData ** allNodes);

  Path(float _pathWeight, const vector<int>& _indices, nodeData ** allNodes);
  ~Path();

  void expand (int numberOfNodes, nodeData ** allNodes);
  //  void appendNode(int nodeIndex); not used as of now
  void appendPath(Path * path2Append,nodeData ** allNodes);
  void copyPath  (Path * headPath);
  void copyPath  (int * indices, int numIndices ,nodeData ** allNodes);
  void print     (nodeData ** allNodes);
  void print     (const char * filename,  nodeData ** allNodes);
  // prints x,y coordinates of every pt. on the best path.

  void printEnergyProfile(nodeData ** allNodes, FILE * fout);
  
  void append2Children(int numNewChildren, Path * newChildren, 
		   int inp_deviatedEdgeIndex);
  void findMoreDeviationsAndAddThemToChildren (vector<int> neighborIndices,
					       int inp_deviatedEdgeIndex, 
					       int numberOfNodes, 
					       nodeData ** allNodes);




  /* for debugging */
  friend void nextBestPath (Path ** desiredPath, Path * root);
  friend void  reportKBestPaths (int K, int startIndex, int endIndex, int numberOfNodes);

  /* end for debugging */
  
  float  getPathWeight()  const;
  int    getNumChildren() const;
  Path * getChildPtr(int) const; 
  int    getIndex(int) const;
  int    getNumIndices() const;
  int    getDeviatedEdgeIndex() const;

  void setDeviatedEdgeIndex(int);
  bool isFound() const;
  void setFound(); /* sets found to true */
  void updatePathWeight(nodeData ** allNodes); /* somehow path weihts computed by runbfs are not correct, this function corrects it. Also used by copypath version 2. */
  bool checkPathWeights(nodeData ** allNodes);


  static const int uninitializedDeviatedEdgeIndexValue;
  
  /* 
     For each node in the path, increments the
     corresponding element in the numVisited array. 
  */
  //  void markVisitedNodes(visitsClass * visits, int numNodes,float closeEnoughRMSD,  nodeData ** allNodes);

  float distance(Path *p, nodeData ** allNodes); // distance between two paths.
  float findNearestNodeDistance(int * closestNodeIndex, int tNI,
				nodeData ** allNodes);
  // looking for the closest node in "this" to target node index (tNI).
  // returns the min. distance, also the closest node index.
  void cluster (vector<Path *> &clusteredPaths,
	        vector<int>    &clusteredPathIndices,
	        int pathIndex, nodeData ** allNodes);
    
private : 

  float pathWeight;
  int  * indices; /* indices[0] == startIndex for elements in the formed tree
		     indices[numIndices - 1] == endIndex ""
		     but not all elements are in the tree!
		   */
  int numIndices;
  static int startIndex, endIndex, numElementsInTree;
  Path * children;
  int numChildren;
  int deviatedEdgeIndex; /*  AtWhichThisPathDeviatesFromItsParent;
			     0 .. numEdges = numIndices - 1?
			 */
  bool found; /*  true if a path is among k best paths and 
		  has been found before.
		  
		  useful if a node is found 
		  but has no children, hence it is found again and again
	      */
  static const float clusterRadius; // distance for which two nodes will be assumed
  // to be close. 

};

class elementType 
{
  // each element of the pq. used in bfs.
public:
  friend class prioritizeNodes;
  elementType (float _time, int _index) : time(_time), index(_index) {};
  float getTime()  const { return time;}
  int   getIndex() const { return index;}
private:
    float time;
    int index;

};

// friend class with function object to compare elementTypes.
class prioritizeNodes
{
public:
  int operator() (const elementType & x, const elementType & y)
    {
      return x.time > y.time;
    }
};



void newFindBestPath (Path ** bestPath, int startIndex, int endIndex,
		      int numberOfNodes, nodeData ** allNodes);

/* 
   returns forward path weight between allNodes[startIndex] and 
   allNodes[endIndex].
 */
float fPathWeight(int startIndex, int endIndex, nodeData ** allNodes);

bool BFS (int sI, int eI, nodeData ** allNodes,
	  map<int,int> &pred, float * pW);

void findNeighbors (vector<int> &neighborIndices, int startNode, 
		    nodeData ** allNodes);

void printEnergyProfiles2BindingSites( const vector<int> & 
				       potBindingSiteIndices, 
				       const char * filename, 
				       nodeData ** allNodes, int numNodes);


void filterIndices  (vector<int> & filteredIndices, int numSampledAround, 
		     neighbourType * sampledAroundList, nodeData ** allNodes,
		     float protSurfDistThreshold, float pairwiseRMSDThreshold);

void findNodesClose2Node(vector<int> & v, int origIndex, int numNodes, 
			 nodeData ** allNodes, float rmsdRadius);

void findNeighborsAndEdgeWeights (vector<int>   &neighborIndices,
					 vector<float> &edgeWeights,
					 int startNode,int numNodes, 
					 nodeData ** allNodes);

unsigned int findNumNeighbors (int i, int numNodes, nodeData ** allNodes);

void analyzePairwiseAtomDistances(int globalNumAtoms, const atomData *adata);

#endif
