#include <string.h>
#include <math.h>

int aa_index_table[] =
/*A  B  C  D  E  F  G  H  I  J  K  L  M  N  O  P  Q  R  S  T  U  V  W  X  Y  Z*/
{0, 20, 1, 2, 3, 4, 5, 6, 7,-1, 8, 9,10,11,-1,12,13,14,15,16,-1,17,18,-1,19,21};

char index_aa_table[23] =
/*0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21*/
"ACDEFGHIKLMNPQRSTVWYBZ";

char *aa3_index_table[] =
{"ALA","CYS","ASP","GLU","PHE","GLY","HIS","ILE","LYS","LEU",
 "MET","ASN","PRO","GLN","ARG","SER","THR","VAL","TRP","TYR"};

char aa3_aa (char *aa) {
  int index;

  for (index = 0; index < 20; ++index) {
    if (strcmp(aa,aa3_index_table[index]) == 0) {
      return index_aa_table[index];
    }
  }
  return 'X';
}

int aa_index (char c) {
  return aa_index_table[c - 'A'];
}


float N_charges[] = 
/* A     C     D     E     F     G     H     I     K     L */
{ 9.69,10.46, 9.60, 9.67, 9.18, 9.78, 9.17, 9.68, 9.18, 9.60,
  9.21, 8.80,10.64, 9.13, 8.99, 9.21, 9.10, 9.62, 9.44, 9.11};
/* M     N     P     Q     R     S     T     V     W     Y */

float C_charges[] =
/* A     C     D     E     F     G     H     I     K     L */
{ 2.34, 1.92, 1.88, 2.19, 2.16, 2.35, 1.82, 2.36, 2.16, 2.36, 
  2.28, 2.02, 1.95, 2.17, 1.82, 2.19, 2.09, 2.32, 2.43, 2.20};
/* M     N     P     Q     R     S     T     V     W     Y */

float side_charges[] =
/* A     C     D     E     F     G     H     I     K     L */
{ 0.00, 8.35, 3.65, 4.25, 0.00, 0.00, 6.00, 0.00,10.79, 0.00, 
  0.00, 0.00, 0.00, 0.00,12.48, 0.00, 0.00, 0.00, 0.00,10.13};
/* M     N     P     Q     R     S     T     V     W     Y */

int side_charge_types[] =
/*A  C  D  E  F  G   H  I   K  L */
{ 0,-1,-1,-1, 0, 0, +1, 0, +1, 0,
  0, 0, 0, 0, +1, 0, 0, 0, 0,-1};
/*M  N  P  Q   R  S  T  V  W  Y */


float compute_N_charge (char aa, float pH) {
  int index = aa_index(aa);
  float pK, percent;

  if ((index < 0) || (index >= 20)) {
    return 0.0;
  } else {
    pK = N_charges[index];
    percent = pow(10,pK-pH)/(1+pow(10,pK-pH));
      return percent;
  }
}

float compute_C_charge (char aa, float pH) {
  int index = aa_index(aa);
  float pK, percent;

  if ((index < 0) || (index >= 20)) {
    return 0.0;
  } else {
    pK = C_charges[index];
    percent = pow(10,pK-pH)/(1+pow(10,pK-pH));
    return -1.0+percent;
  }
}

float compute_side_charge_aux (char aa, float pH) {
  int index = aa_index(aa);
  float pK, percent;
  int chg;
    
  if ((index < 0) || (index >= 20)) {
    return 0.0;
  } else if ((chg = side_charge_types[index]) == 0) {
    return 0.0;
  } else {
    pK = side_charges[index];
    percent = pow(10,pK-pH)/(1+pow(10,pK-pH));
    if (chg == +1) {
      return percent;
    } else {
      return -1.0+percent;
    }
  }
}

float compute_side_charge (char *atom, char residue, float pH) {
  switch (residue) {
  case 'C': 
    if (strcmp(atom,"SG") == 0) {
      return compute_side_charge_aux('C',pH);
    } else {
      return 0.0;
    }
  case 'D':
    if (strncmp(atom,"OD",2) == 0) {
      return compute_side_charge_aux('D',pH)/2.0;
    } else {
      return 0.0;
    }
  case 'E':
    if (strncmp(atom,"OE",2) == 0) {
      return compute_side_charge_aux('E',pH)/2.0;
    } else {
      return 0.0;
    }
  case 'K':
    if (strcmp(atom,"NZ") == 0) {
      return compute_side_charge_aux('K',pH);
    } else {
      return 0.0;
    }
  case 'R':
    if (strcmp(atom,"NH2") == 0) {
      return compute_side_charge_aux('R',pH);
    } else {
      return 0.0;
    }
  case 'Y':
    if (strcmp(atom,"OH") == 0) {
      return compute_side_charge_aux('Y',pH);
    } else {
      return 0.0;
    }
  default: return 0.0;
  }
}
