#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <stdio.h>
#include "display.h"
#include "tk.h"
#include "ldock.H"
#include <vector.h>
#include <iostream.h>

extern void setupGLWindow(char *);
extern int My_Tk_Init(int, char**, Tcl_AppInitProc*);
extern void InitPosition(int, int, int, int);
extern void addReshapeFunc(void (*Func)(GLsizei, GLsizei));
extern void addKeyFunc(int key, void (*Func)(void));
extern void addMouseFunc(int mouse, int mode, void (*Func)(AUX_EVENTREC *));
extern void tellTcl(char*);
extern void randomTest();
extern int mycompare(const void*, const void*);

extern Display *mydisplay;
extern Window GLwin;

extern receptorClass myReceptor;
extern float MAX_XYZ;
extern int globalNumAtoms;
extern int globalRootAtom;
extern ligAtom *myLigand;
extern ligAtom **globalLigAtomPtrs;
extern atomData *origAtomData;
extern nodeData* allNodes[MAXNODES];
extern neighbourType nodeEnergyList[MAXNODES];
extern neighbourType sampledAroundList[MIN_REGIONS+1];
extern int numSampledAround;
extern int nodeEnergyListCount;
extern int rootSampleIndex;
extern int rootSampleMinIndex;
extern int numNodes;
extern int currentGoalIndex;
float colorTable[500][3];

static GLfloat LdownXold = -1;
static GLfloat LdownYold = -1;
static GLfloat MdownXold = -1;
static GLfloat MdownYold = -1;
static GLfloat RdownXold = -1;
static GLfloat RdownYold = -1;


GLfloat currentTransformMat[16];

int SOLID_SPHERE_LIST;
int currentLigandList;
int goalLigandList;
int displayNodesList = -1;
int displayCurrentLigand = 0;
int displayGridList = -1;
int displayAnimation = 0;
int animatedLigandList = -1;
int currentAnimationIndex = -1;
int displayNodesOn = 0;
int displayGridOn = 0;
int displayMyGoalList = -1;
int displayMyGoalOn = 0;
int displayBestNodesList = -1;
int displayBestNodesOn = 0;
int numBestNodes = 0;
int displaySpecificNodesList = -1;
int displaySpecificNodesOn = 0;
int numSpecificNodes = 0;
int displayHSDNOn = 0; // high stat dist nodes
int displayHSDNList = -1;


int currentDisplaySample = -1;
neighbourType tempNeighbourList[SAMPLES_PER_REGION + 10];


void display(void);
void init(void);
void myReshape(GLsizei, GLsizei);
void setBindings(void);
void mouseLeftDown(AUX_EVENTREC*);
void mouseLeftUp(AUX_EVENTREC*);
void mouseMiddleDown(AUX_EVENTREC*);
void mouseMiddleUp(AUX_EVENTREC*);
void mouseRightDown(AUX_EVENTREC*);
void mouseRightUp(AUX_EVENTREC*);

void applyMtoV(GLfloat*, GLfloat*, GLfloat*);
void applyRottoV(GLfloat*, GLfloat*, GLfloat*);
void printMat_draw(GLfloat*);
void loadFunnel(vector<int> & funnel1, char * filename);


void receptorClass::createObjList() {

  //GLfloat tempmat[16];
  int i;
  GLfloat t0 = 0.0, t1 = 0.0, t2 = 0.0, tempfloat;

  for (i = 0; i < numReceptorAtoms; i++) {
    t0 += recAtom[i][0];
    t1 += recAtom[i][1];
    t2 += recAtom[i][2];
  }

  tempfloat = -1.0/((GLfloat)(numReceptorAtoms));

  t0 *= tempfloat;
  t1 *= tempfloat;
  t2 *= tempfloat;

  printf("%f %f %f\n", t0, t1, t2);

  glPushMatrix();
  glLoadIdentity();
  glTranslatef(t0, t1, t2);
  glGetFloatv(GL_MODELVIEW_MATRIX, objMat);
  glPopMatrix();

  modelNumber = glGenLists(1);
  center[0] = 0.0;
  center[1] = 0.0;
  center[2] = 0.0;

  glNewList(modelNumber, GL_COMPILE);

  for (i = 0; i < numReceptorAtoms; i++) {  
    glPushMatrix();
    if (recAtomCharge[i] > 0.0001) {
      glColor3f(1.0, 0.0, 1.0);
    }
    else if (recAtomCharge[i] < -0.0001) {
      glColor3f(0.0, 1.0, 1.0);
    }
    else {
      glColor3f(0.0, 0.0, 1.0);
    }
    glTranslatef(recAtom[i][0], recAtom[i][1], recAtom[i][2]);
    glCallList(SOLID_SPHERE_LIST);
    glPopMatrix();
  }
  glEndList();

}


  
void myMatMult(GLfloat *m1, GLfloat *m2) {

  int i;
  GLfloat t[3];

  for (i = 0; i < 4; i++) {
    t[0] = m1[i*4] * m2[0] + m1[i*4 + 1] * m2[4] + m1[i*4 + 2] * m2[8] + m2[12];
    t[1] = m1[i*4] * m2[1] + m1[i*4 + 1] * m2[5] + m1[i*4 + 2] * m2[9] + m2[13];
    t[2] = m1[i*4] * m2[2] + m1[i*4 + 1] * m2[6] + m1[i*4 + 2] * m2[10] + m2[14];

    m1[i*4] = t[0];
    m1[i*4 + 1] = t[1];
    m1[i*4 + 2] = t[2];

  }
}







void applyMtoV(GLfloat *vout, GLfloat *vin, GLfloat *matin) {

  GLfloat t[3];

  t[0] = (vin[0])* matin[0] + (vin[1])* matin[4] + (vin[2])* matin[8] + matin[12];
  t[1] = (vin[0])* matin[1] + (vin[1])* matin[5] + (vin[2])* matin[9] + matin[13];
  t[2] = (vin[0])* matin[2] + (vin[1])* matin[6] + (vin[2])* matin[10] + matin[14];

  vout[0] = t[0];
  vout[1] = t[1];
  vout[2] = t[2];

}


void applyRottoV(GLfloat *vout, GLfloat *vin, GLfloat *matin) {

  GLfloat t[3];

  t[0] = (vin[0])* matin[0] + (vin[1])* matin[4] + (vin[2])* matin[8];
  t[1] = (vin[0])* matin[1] + (vin[1])* matin[5] + (vin[2])* matin[9];
  t[2] = (vin[0])* matin[2] + (vin[1])* matin[6] + (vin[2])* matin[10];

  vout[0] = t[0];
  vout[1] = t[1];
  vout[2] = t[2];

}

void printMat_draw(GLfloat *m) {
  
  int i;
  
  for (i = 0; i<16; i++) {
    if (i%4 == 0) 
      printf("\n");
    printf ("%f  ", m[i]);
  }

  printf("\n");  
  
}


void display(void) {

  int i;
  
  glClearColor(0.0, 0.0, 0.0, 1.0); 
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 
  
  glPushMatrix();
  glMultMatrixf(myReceptor.objMat);
  glCallList(myReceptor.modelNumber);
  glCallList(goalLigandList);
  if (displayCurrentLigand) 
    glCallList(currentLigandList);
  if (displayAnimation) {
    for (i = animatedLigandList; i < currentAnimationIndex; i++) {
      glCallList(i);
    }
  }
  if (displayBestNodesOn) {
    for (i = displayBestNodesList; i < displayBestNodesList + numBestNodes; i++) {
      glCallList(i);
    }
  }
  if (displaySpecificNodesOn) {
    for (i = displaySpecificNodesList; i < displaySpecificNodesList + numSpecificNodes; i++) {
      glCallList(i);
    }
  }

  if (displayNodesOn) {
    glCallList(displayNodesList);
  }
  if (displayGridOn) {
    glCallList(displayGridList);
  }
  if (displayMyGoalOn) 
    {
      glCallList(displayMyGoalList);
    }
  if (displayHSDNOn) 
    {
      glCallList(displayHSDNList);
    }


  glPopMatrix();
  glFlush();

  glXSwapBuffers(mydisplay, GLwin); 

}


void init(void) {

  GLfloat light_position[] = { 0.0, 0.0, 1.0, 0.0 }; 
  GLfloat light_diffuse[] = { 1.0, 1.0, 1.0, 0.0 }; 
  GLfloat mat_ambient[] = { 0.5, 0.5, 0.5, 0.0 }; 
  //  GLfloat lmodel_ambient[] = { 0.2, 0.2, 0.2, 1.0 };

  glMatrixMode(GL_MODELVIEW); 
  glShadeModel(GL_SMOOTH);
  glLoadIdentity();

  glLightfv(GL_LIGHT0, GL_POSITION, light_position); 
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);

  //  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_FALSE);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glDepthFunc(GL_LESS);
  glEnable(GL_DEPTH_TEST);
  glColorMaterial(GL_FRONT, GL_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);
  glFlush();

}

void myReshape(GLsizei w, GLsizei h) {

  float max_d;

  glViewport(0, 0, w, h); 
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  //glFrustum(-MAX_XYZ/2.0, MAX_XYZ/2.0, -MAX_XYZ/2.0, MAX_XYZ/2.0, 20.0, 90.0);

  //  max_d = sqrt(MAX_XYZ * MAX_XYZ * 3.0);
  max_d = sqrt(MAX_XYZ * MAX_XYZ * 1.0);

  glOrtho(-(max_d)/2.0, (max_d)/2.0, -(max_d)/2.0, (max_d)/2.0, -(max_d)/2.0, (max_d)/2.0);
  glMatrixMode(GL_MODELVIEW); 

  display();

}


void setBindings(void) {

  addMouseFunc(AUX_LEFTBUTTON, AUX_MOUSELOC, mouseLeftDown);
  addMouseFunc(AUX_LEFTBUTTON, AUX_MOUSEUP, mouseLeftUp);
  addMouseFunc(AUX_MIDDLEBUTTON, AUX_MOUSELOC, mouseMiddleDown);
  addMouseFunc(AUX_MIDDLEBUTTON, AUX_MOUSEUP, mouseMiddleUp);
  addMouseFunc(AUX_RIGHTBUTTON, AUX_MOUSELOC, mouseRightDown);
  addMouseFunc(AUX_RIGHTBUTTON, AUX_MOUSEUP, mouseRightUp);
}


void mouseLeftDown(AUX_EVENTREC *mouseEvent) {

  GLfloat Xval, Yval, Xt, Yt;
  GLfloat transValue[3];
  //int i;  

  Xt = (float)(mouseEvent->data[AUX_MOUSEX]);
  Yt = (float)(mouseEvent->data[AUX_MOUSEY]);

  if ((LdownXold == -1.0) && (LdownYold == -1.0)) {
    LdownXold = Xt;
    LdownYold = Yt;
  }
  else {

    Xval = (Xt - LdownXold)*0.5;
    Yval = (Yt - LdownYold)*0.5;

    transValue[0] = myReceptor.center[0];
    transValue[1] = myReceptor.center[1];
    transValue[2] = myReceptor.center[2];

    
    glPushMatrix();
    
    glLoadIdentity();
    glTranslatef(transValue[0], transValue[1], transValue[2]);
    glRotatef(Xval, 0.0, 1.0, 0.0);
    glRotatef(Yval, 1.0, 0.0, 0.0);
    glTranslatef(-transValue[0], -transValue[1], -transValue[2]);
    
    glGetFloatv(GL_MODELVIEW_MATRIX, currentTransformMat);
    
    glLoadMatrixf(currentTransformMat);
    glMultMatrixf(myReceptor.objMat);
    glGetFloatv(GL_MODELVIEW_MATRIX, myReceptor.objMat);
    
    applyMtoV(myReceptor.center, myReceptor.center, currentTransformMat); 

    glPopMatrix();

    LdownXold = Xt;
    LdownYold = Yt;

    display();
  }
}


  
void mouseLeftUp(AUX_EVENTREC *mouseEvent) {

  LdownXold = -1.0; 
  LdownYold = -1.0; 

}


void mouseMiddleDown(AUX_EVENTREC *mouseEvent) {

  GLfloat Xval, Yval, Xt, Yt;
  //int i;

  Xt = (float)(mouseEvent->data[AUX_MOUSEX]);
  Yt = (float)(mouseEvent->data[AUX_MOUSEY]);

  if ((MdownXold == -1.0) && (MdownYold == -1.0)) {
    MdownXold = Xt;
    MdownYold = Yt;
  }
  else {
    
    Xval = ((Xt - MdownXold)/500.0)*MAX_XYZ;
    Yval = ((MdownYold - Yt)/500.0)*MAX_XYZ;

    glPushMatrix();
    
    glLoadIdentity();
    glTranslatef(Xval, Yval, 0.0);
    glGetFloatv(GL_MODELVIEW_MATRIX, currentTransformMat);
    
    glLoadMatrixf(currentTransformMat);
    glMultMatrixf(myReceptor.objMat);
    glGetFloatv(GL_MODELVIEW_MATRIX, myReceptor.objMat);
    
    myReceptor.center[0] += Xval;
    myReceptor.center[1] += Yval;

    glPopMatrix();

    MdownXold = Xt;
    MdownYold = Yt;

    display();
  }
}
    
  
void mouseMiddleUp(AUX_EVENTREC *mouseEvent) {
  
  MdownXold = -1.0; 
  MdownYold = -1.0; 

}    
  
void mouseRightDown(AUX_EVENTREC *mouseEvent) {

  GLfloat Yval, Xt, Yt;
  //int i;

  Xt = (float)(mouseEvent->data[AUX_MOUSEX]);
  Yt = (float)(mouseEvent->data[AUX_MOUSEY]);

  if ((RdownXold == -1.0) && (RdownYold == -1.0)) {    
    RdownXold = Xt;
    RdownYold = Yt;
  }
  else {

    Yval = ((RdownYold - Yt)/500.0)*MAX_XYZ;

    glPushMatrix();
    
    glLoadIdentity();
    glTranslatef(0.0, 0.0, Yval);
    glGetFloatv(GL_MODELVIEW_MATRIX, currentTransformMat);
    
    glLoadMatrixf(currentTransformMat);
    glMultMatrixf(myReceptor.objMat);
    glGetFloatv(GL_MODELVIEW_MATRIX, myReceptor.objMat);

    myReceptor.center[2] += Yval;

    glPopMatrix();

    RdownXold = Xt;
    RdownYold = Yt;

    display();
  }
}
    
  
void mouseRightUp(AUX_EVENTREC *mouseEvent) {
  
  RdownXold = -1.0; 
  RdownYold = -1.0; 

}    




void makeSolidSphere (GLdouble radius)
{
  GLUquadricObj *quadObj;

  quadObj = gluNewQuadric();
  gluQuadricDrawStyle (quadObj, GLU_FILL);
  gluQuadricNormals (quadObj, GLU_SMOOTH);
  gluQuadricTexture (quadObj, GL_FALSE);
  glNewList(SOLID_SPHERE_LIST, GL_COMPILE);
  gluSphere (quadObj, radius, 8, 8);
  glEndList();

}


void makeGoalLigand(atomData *adata) {

  int i, k, tempatomnum;

  glNewList(goalLigandList, GL_COMPILE);
  
  glDisable(GL_LIGHTING);
  glLineWidth(2.0);
  glColor3f(1.0,0.0,0.0);

  glBegin(GL_LINES);

  for (i = 0; i < globalNumAtoms; i++) {
    for (k = 0; k < globalLigAtomPtrs[i]->numBonds; k++) {
      tempatomnum = globalLigAtomPtrs[i]->bonds[k]->atomNum;
      if (tempatomnum > i) {
	glVertex3fv(adata[i].xyz);
	glVertex3fv(adata[tempatomnum].xyz);
      }
    }
  }

  glEnd();

  glEnable(GL_LIGHTING);

  glEndList();

  displayAnimation = 0;

}


void makeCurrentLigand(atomData *adata) {

  int i, k, tempatomnum;

  glNewList(currentLigandList, GL_COMPILE);
  
  glDisable(GL_LIGHTING);
  glLineWidth(2.0);
  glColor3f(0.0,1.0,0.0);

  glBegin(GL_LINES);

  for (i = 0; i < globalNumAtoms; i++) {
    for (k = 0; k < globalLigAtomPtrs[i]->numBonds; k++) {
      tempatomnum = globalLigAtomPtrs[i]->bonds[k]->atomNum;
      if (tempatomnum > i) {
	glVertex3fv(adata[i].xyz);
	glVertex3fv(adata[tempatomnum].xyz);
      }
    }
  }

  glEnd();

  glEnable(GL_LIGHTING);

  glEndList();

  displayCurrentLigand = 1.0;
  displayAnimation = 0;

  display();

}


void addAnimatedLigand(atomData *adata) {

  int i, k, tempatomnum;

  if (currentAnimationIndex < animatedLigandList + 1000) {

    glNewList(currentAnimationIndex, GL_COMPILE);
  
    glDisable(GL_LIGHTING);
    //    glLineWidth(1.0);
    glLineWidth(2.0);
    glColor3f(1.0,1.0,1.0);
    
    glBegin(GL_LINES);
    
    for (i = 0; i < globalNumAtoms; i++) {
      for (k = 0; k < globalLigAtomPtrs[i]->numBonds; k++) {
	tempatomnum = globalLigAtomPtrs[i]->bonds[k]->atomNum;
	if (tempatomnum > i) {
	  glVertex3fv(adata[i].xyz);
	  glVertex3fv(adata[tempatomnum].xyz);
	}
      }
    }

    glEnd();
    
    glEnable(GL_LIGHTING);
    
    glEndList();

    currentAnimationIndex++;
  }
  else {
    cout << "Increase animation slots" << endl;
  }

}


void initAnimation() {

  if (animatedLigandList == -1) {
    animatedLigandList = glGenLists(1000);
  }

  currentAnimationIndex = animatedLigandList;
  displayAnimation = 0;

}

void doAnimation() {

  displayAnimation = 1;

  display();  


}

void clearAnimation() {

  displayAnimation = 0;

  display();  



#if 0    //only for saving images
  char temppixels[1000000];
  char imagename[200];
  int i, count = 1;
  FILE *datafile;
  
  for (i = animatedLigandList; i < currentAnimationIndex; i++) {
    glClearColor(0.0, 0.0, 0.0, 1.0); 
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 
    
    glPushMatrix();
    glMultMatrixf(myReceptor.objMat);
    glCallList(myReceptor.modelNumber);
    glCallList(goalLigandList);
    glCallList(i);
    glPopMatrix();
    glFlush();
    glXSwapBuffers(mydisplay, GLwin); 

    glReadPixels(0, 0, 400, 400, GL_RGB, GL_UNSIGNED_BYTE, temppixels);
  
    sprintf(imagename, "image%d.ppm", count);
    count++;

    datafile = fopen(imagename, "w");
    
    fprintf(datafile, "P6\n400 400\n255\n");
    
    fwrite(temppixels, sizeof(char), 480000, datafile);

    fclose(datafile);

  }
#endif
}

void fillColorTable() {

  int i;

  for (i = 0; i < 500; i++) {

    colorTable[i][0] = ((float)(random()))/((float)(MY_RAND_MAX));
    colorTable[i][1] = ((float)(random()))/((float)(MY_RAND_MAX));
    colorTable[i][2] = ((float)(random()))/((float)(MY_RAND_MAX));
  }

}


void colorAllNodes() {
  
  int i;
  vector<int> funnel1, funnel2,funnel3, funnel4, funnel5, funnel6;
  
  loadFunnel(funnel1, "funnel1.txt");
  loadFunnel(funnel2, "funnel2.txt");
  loadFunnel(funnel3, "funnel3.txt");
  loadFunnel(funnel4, "funnel4.txt");
  loadFunnel(funnel5, "funnel5.txt");
  loadFunnel(funnel6, "funnel6.txt");
  
  glNewList(displayNodesList, GL_COMPILE);

  glDisable(GL_LIGHTING);

  glBegin(GL_POINTS);
  glColor3f(1.0, 1.0, 1.0);
  for (i = 0; i < funnel1.size(); i++) {
    //    if (allNodes[i]->myTree->numNodesInTree < 100) {
    //      glColor3fv(colorTable[allNodes[i]->myTree->myTreeName]);
    assert (funnel1[i] <numNodes);
    
    glVertex3fv(allNodes[funnel1[i]] ->cog);
    //    }
  }
  glColor3f(1.0, 0.0, 0.0);
  for (i = 0; i < funnel2.size(); i++) {
    //    if (allNodes[i]->myTree->numNodesInTree < 100) {
    //      glColor3fv(colorTable[allNodes[i]->myTree->myTreeName]);
    assert (funnel2[i] <numNodes);

    glVertex3fv(allNodes[funnel2[i]]->cog);
    //    }
  }
  glColor3f(0.0, 1.0, 0.0);
  for (i = 0; i < funnel3.size(); i++) {
    //    if (allNodes[i]->myTree->numNodesInTree < 100) {
    //      glColor3fv(colorTable[allNodes[i]->myTree->myTreeName]);
    assert (funnel3[i] <numNodes);

    glVertex3fv(allNodes[funnel3[i]]->cog);
    //    }
  }
  glColor3f(0.0, 0.0, 1.0);
  for (i = 0; i < funnel4.size(); i++) {
    //    if (allNodes[i]->myTree->numNodesInTree < 100) {
    //      glColor3fv(colorTable[allNodes[i]->myTree->myTreeName]);
    assert (funnel4[i] <numNodes);

    glVertex3fv(allNodes[funnel4[i]]->cog);
    //    }
  }
  glColor3f(1.0, 1.0, 0.0);
  for (i = 0; i < funnel5.size(); i++) {
    //    if (allNodes[i]->myTree->numNodesInTree < 100) {
    //      glColor3fv(colorTable[allNodes[i]->myTree->myTreeName]);
    assert (funnel5[i] <numNodes);

    glVertex3fv(allNodes[funnel5[i]]->cog);
    //    }
  }
  glColor3f(0.0, 1.0, 1.0);
  for (i = 0; i < funnel6.size(); i++) {
    //    if (allNodes[i]->myTree->numNodesInTree < 100) {
    //      glColor3fv(colorTable[allNodes[i]->myTree->myTreeName]);
    assert (funnel6[i] <numNodes);

    glVertex3fv(allNodes[funnel6[i]]->cog);
    //    }
  }
  glEnd();
  
  glEnable(GL_LIGHTING);
  
  glEndList();

  displayNodesOn = 1;

}

void colorAllNodes2() {
  
  int i;

  glNewList(displayNodesList, GL_COMPILE);

  glDisable(GL_LIGHTING);

  glBegin(GL_POINTS);
  glColor3f(1.0, 1.0, 1.0);
  for (i = 0; i < numNodes; i++) {
    //    if (allNodes[i]->myTree->numNodesInTree < 100) {
    //      glColor3fv(colorTable[allNodes[i]->myTree->myTreeName]);
      glVertex3fv(allNodes[i]->cog);
      //    }
  }

  glEnd();
  
  glEnable(GL_LIGHTING);
  
  glEndList();

  displayNodesOn = 1;

}

void displayGrid(float threshold) {
  int x, y, z, zindex, yindex;
  VEC tvec;
  float divFactor, divFactorBy2;


  if (displayGridList == -1) {
    displayGridList = glGenLists(1);
    
  }
  glNewList(displayGridList, GL_COMPILE);
  
  glDisable(GL_LIGHTING);
  
  glBegin(GL_POINTS);
  
  glColor3f(1.0, 1.0, 1.0);
  
  divFactor = 1.0/(myReceptor.potMatMultFact);
  divFactorBy2 = divFactor/2.0;
  
  for (z = 0; z < myReceptor.potMatSize; z++) {
      
    zindex = z * myReceptor.potMatSize * myReceptor.potMatSize;
    tvec[2] = ((float)(z)) * divFactor + divFactorBy2;
    
    for (y = 0; y < myReceptor.potMatSize; y++) {
      
      yindex = y * myReceptor.potMatSize;
      tvec[1] = ((float)(y)) * divFactor + divFactorBy2;
      
      for (x = 0; x < myReceptor.potMatSize; x++) {
	tvec[0] = ((float)(x)) * divFactor + divFactorBy2;
	
	if (myReceptor.vanPot[zindex+yindex+x] - myReceptor.elecPot[zindex+yindex+x] < threshold) {     
	  glVertex3fv(tvec);
	}
      }
    }
  }
  
  glEnd();
  
  glEnable(GL_LIGHTING);
  
  glEndList();

    
  displayGridOn = 1;
  display();

}

// funnel contain the list of node indices that will be added to
// the display.
// num is the # of prev. inserted nodes.

void insertToNodeList(const vector<int> & funnel, int num, float r, float g, 
		      float b)
{
  int j, k, tempatomnum;

  for (unsigned i = 0; i < funnel.size(); i++) 
    {
      int myindex = funnel[i];

      glNewList(displaySpecificNodesList + i + num, GL_COMPILE);
  
      glDisable(GL_LIGHTING);
      glLineWidth(1.0);
      //    red = (maxE - allNodes[myindex]->nodeEnergy)/
      //    (maxE - minE);
      //    glColor3f(red,1.0 - red,0.0);
      
      glColor3f(r, g, b);
      glBegin(GL_LINES);

      for (j = 0; j < globalNumAtoms; j++) {
	for (k = 0; k < globalLigAtomPtrs[j]->numBonds;
	     k++) 
	  {
	    tempatomnum = globalLigAtomPtrs[j]->bonds[k]->atomNum;
	    if (tempatomnum > j) {
	      glVertex3fv(allNodes[myindex]->myData[j].xyz);
	      glVertex3fv(allNodes[myindex]->myData[tempatomnum].xyz);
	    }
	  }
      }
      
      glEnd();
      
      glEnable(GL_LIGHTING);
      
      glEndList();
    }
}


void dontDisplayGrid() {
  
  displayGridOn = 0;
  display();

}


void displayNodes() {

  if (displayNodesList == -1) {
    displayNodesList = glGenLists(1);
  }

  //  fillColorTable();
  colorAllNodes();

  displayNodesOn = 1;
  display();

}

void dontDisplayNodes() {
  
  displayNodesOn = 0;
  display();

}

void displayMyGoal(int g) {

  int tempatomnum, i, k, temp;

  cout << "attention! rootSampleIndex  may not be set." << endl;

  if (currentDisplaySample >= 0) {

    if (currentDisplaySample == 0) {     
      temp = rootSampleIndex;
    }
    else {
      temp = sampledAroundList[currentDisplaySample-1].linkedBit;
    }
    
    for (i = 0; i < numBestNodes; i++) {

      tempNeighbourList[i].index = temp+i;
      tempNeighbourList[i].dist = allNodes[temp+i]->nodeEnergy;

    }

    if (currentDisplaySample == 0) {
      tempNeighbourList[numBestNodes].index = 0;
      tempNeighbourList[numBestNodes].dist = allNodes[0]->nodeEnergy;
    }
    else {
      tempNeighbourList[numBestNodes].index = sampledAroundList[currentDisplaySample-1].index;
      tempNeighbourList[numBestNodes].dist = allNodes[sampledAroundList[currentDisplaySample-1].index]->nodeEnergy;
    }

    qsort(tempNeighbourList, numBestNodes+1, sizeof(neighbourType), mycompare); 

    if (displayMyGoalList == -1) {
      displayMyGoalList = glGenLists(1);
    }
   
    currentGoalIndex = tempNeighbourList[g].index;

    cout << "CURRENT GOAL ENERGY = " << tempNeighbourList[g].dist << endl;

    glNewList(displayMyGoalList, GL_COMPILE);
  
    glDisable(GL_LIGHTING);
    glLineWidth(2.0);
    glColor3f(1.0,1.0,1.0);

    glBegin(GL_LINES);

    for (i = 0; i < globalNumAtoms; i++) {
      for (k = 0; k < globalLigAtomPtrs[i]->numBonds; k++) {
	tempatomnum = globalLigAtomPtrs[i]->bonds[k]->atomNum;
	if (tempatomnum > i) {
	  glVertex3fv(allNodes[currentGoalIndex]->myData[i].xyz);
	  glVertex3fv(allNodes[currentGoalIndex]->myData[tempatomnum].xyz);
	}
      }
    }

    glEnd();
    
    glEnable(GL_LIGHTING);
    
    glEndList();
    
    displayMyGoalOn = 1;
    display();
  }

}

void dontDisplayHighStationaryDistNodes()
{
  displayHSDNOn = 0;
  display();
}

void dontDisplayMyGoal() {
  
  displayMyGoalOn = 0;
  display();

}


#if 0
void displayMyGoal(int g) {

  int tempatomnum, i, k;

  if ((g >= 0) && (g < nodeEnergyListCount)) {

    if (displayMyGoalList == -1) {
      displayMyGoalList = glGenLists(1);
    }
   
    currentGoalIndex = nodeEnergyList[g].index;

    cout << "CURRENT ENERGY = " << nodeEnergyList[g].dist << endl;

    glNewList(displayMyGoalList, GL_COMPILE);
  
    glDisable(GL_LIGHTING);
    glLineWidth(2.0);
    glColor3f(1.0,1.0,1.0);

    glBegin(GL_LINES);

    for (i = 0; i < globalNumAtoms; i++) {
      for (k = 0; k < globalLigAtomPtrs[i]->numBonds; k++) {
	tempatomnum = globalLigAtomPtrs[i]->bonds[k]->atomNum;
	if (tempatomnum > i) {
	  glVertex3fv(allNodes[currentGoalIndex]->myData[i].xyz);
	  glVertex3fv(allNodes[currentGoalIndex]->myData[tempatomnum].xyz);
	}
      }
    }

    glEnd();
    
    glEnable(GL_LIGHTING);
    
    glEndList();
    
    displayMyGoalOn = 1;
    display();
  }

}
#endif




void displayBestNodes(int g) {

  int tempatomnum, myindex, i,j, k;
  float maxE, minE, red;
  int temp;


  cout << "attention! rootSampleIndex and rootSampleMinIndex may not be set." << endl;

  if (displayBestNodesList == -1) {
    displayBestNodesList = glGenLists(1000);
  }

  numBestNodes = SAMPLES_PER_REGION;

  currentDisplaySample = g;

  maxE = sampledAroundList[numSampledAround].dist;
  minE = sampledAroundList[0].dist;

  if (g > 0) {
    cout << "CURRENT ENERGY = " << sampledAroundList[g-1].dist << endl;    
    
    temp = sampledAroundList[g-1].linkedBit;
  }
  else if (g == 0) {
    cout << "CURRENT ENERGY = " << allNodes[rootSampleMinIndex]->nodeEnergy << endl;    
    
    temp = rootSampleIndex;
  }

  for (i = 0; i < numBestNodes; i++) {

    myindex = temp + i;

    glNewList(displayBestNodesList+i, GL_COMPILE);
  
    glDisable(GL_LIGHTING);
    glLineWidth(1.0);
    red = (maxE - allNodes[myindex]->nodeEnergy)/(maxE - minE);

    glColor3f(red,1.0 - red,0.0);

    glBegin(GL_LINES);

    for (j = 0; j < globalNumAtoms; j++) {
      for (k = 0; k < globalLigAtomPtrs[j]->numBonds; k++) {
	tempatomnum = globalLigAtomPtrs[j]->bonds[k]->atomNum;
	if (tempatomnum > j) {
	  glVertex3fv(allNodes[myindex]->myData[j].xyz);
	  glVertex3fv(allNodes[myindex]->myData[tempatomnum].xyz);
	}
      }
    }

    glEnd();
    
    glEnable(GL_LIGHTING);
    
    glEndList();
  }

    
  displayBestNodesOn = 1;
  display();


}

#if 0
// displays a specific set of nodes in r,g,b.
void displaySpecificNodes(vector<int> nodeList, float r, 
			  float g, float b) 
{
  int tempatomnum, myindex, i,j,k;
  //  float maxE, minE, red;
  int temp;
  
 
  //  currentDisplaySample = g;

  //  maxE = sampledAroundList[numSampledAround].dist;
  //  minE = sampledAroundList[0].dist;
  /*
    if (g > 0) {
    cout << "CURRENT ENERGY = " << sampledAroundList[g-1].dist << endl;    
    
    temp = sampledAroundList[g-1].linkedBit;
    }
    else if (g == 0) {
    cout << "CURRENT ENERGY = " << allNodes[rootSampleMinIndex]->nodeEnergy << endl;    
    
    temp = rootSampleIndex;
    }
  */
  for (i = 0; i < numSpecificNodes; i++) {

    myindex = nodeList[i];

    glNewList(displaySpecificNodesList+i, GL_COMPILE);
  
    glDisable(GL_LIGHTING);
    glLineWidth(1.0);
    //    red = (maxE - allNodes[myindex]->nodeEnergy)/(maxE - minE);

    //    glColor3f(red,1.0 - red,0.0);
    glColor3f(r, g, b);

    glBegin(GL_LINES);

    for (j = 0; j < globalNumAtoms; j++) {
      for (k = 0; k < globalLigAtomPtrs[j]->numBonds; k++) {
	tempatomnum = globalLigAtomPtrs[j]->bonds[k]->atomNum;
	if (tempatomnum > j) {
	  glVertex3fv(allNodes[myindex]->myData[j].xyz);
	  glVertex3fv(allNodes[myindex]->myData[tempatomnum].xyz);
	}
      }
    }

    glEnd();
    
    glEnable(GL_LIGHTING);
    
    glEndList();
  }

    
  displaySpecificNodesOn = 1;
  display();
}
#endif

void dontDisplayBestNodes() {
  
  displayBestNodesOn = 0;
  display();

}
  
void dontDisplaySpecificNodes() {
  
  displaySpecificNodesOn = 0;
  display();

}

#if 0
void displayBestNodes(int g) {

  int tempatomnum, myindex, i,j, k;
  float maxE, minE, red;

  if (displayBestNodesList == -1) {
    displayBestNodesList = glGenLists(1000);
  }
  
  numBestNodes = g;
    
  maxE = allNodes[nodeEnergyList[g].index]->nodeEnergy;
  minE = allNodes[nodeEnergyList[0].index]->nodeEnergy;

  for (i = 0; i < numBestNodes; i++) {

    myindex = nodeEnergyList[i].index;

    glNewList(displayBestNodesList+i, GL_COMPILE);
  
    glDisable(GL_LIGHTING);
    glLineWidth(1.0);
    red = (maxE - allNodes[myindex]->nodeEnergy)/(maxE - minE);

    glColor3f(red,1.0 - red,0.0);

    glBegin(GL_LINES);

    for (j = 0; j < globalNumAtoms; j++) {
      for (k = 0; k < globalLigAtomPtrs[j]->numBonds; k++) {
	tempatomnum = globalLigAtomPtrs[j]->bonds[k]->atomNum;
	if (tempatomnum > j) {
	  glVertex3fv(allNodes[myindex]->myData[j].xyz);
	  glVertex3fv(allNodes[myindex]->myData[tempatomnum].xyz);
	}
      }
    }

    glEnd();
    
    glEnable(GL_LIGHTING);
    
    glEndList();
  }

    
  displayBestNodesOn = 1;
  display();


}
#endif

/*
  loads the list of nodes in funnel.
*/
void loadFunnel(vector<int> & funnel1, char * filename)
{
  int nodeIndex;
  
  funnel1.resize(0);
  FILE * fin = fopen(filename,"r");
  if (fin == NULL)
      {
	cerr << "couldn't read " << filename << endl;
	return;
      }
  cout << "reading " << filename << endl;
  fscanf(fin, "%d\n", &nodeIndex);

  while (!feof(fin))
    {
      funnel1.push_back(nodeIndex);
      fscanf(fin, "%d\n", &nodeIndex);
    }
  fclose(fin);
}

void addFunnelNodesAndRootToDisplayList
(const vector<int>& funnel1, int rootNode)
{

  int j, k;

  glNewList(displaySpecificNodesList, GL_COMPILE);

  glDisable(GL_LIGHTING);

  glBegin  (GL_POINTS);
  
  glColor3f(1.0, 1.0, 1.0);
  
  for (unsigned int i = 0; i < funnel1.size(); i++) 
    {
      //if (allNodes[i]->myTree->numNodesInTree < 100) {
      //glColor3fv(colorTable[allNodes[i]->myTree->myTreeName]);
      assert (funnel1[i] < numNodes);
      
      glVertex3fv(allNodes[funnel1[i]] ->cog);
      //    }
    }

  glEnd();

  glLineWidth(1.0);
      //    red = (maxE - allNodes[myindex]->nodeEnergy)/
      //    (maxE - minE);
      //    glColor3f(red,1.0 - red,0.0);
      
  glColor3f(1.0, 1.0, 1.0);
  glBegin(GL_LINES);
  
  for (j = 0; j < globalNumAtoms; j++) {
    for (k = 0; k < globalLigAtomPtrs[j]->numBonds;
	 k++) 
      {
	int tempatomnum = globalLigAtomPtrs[j]->bonds[k]->atomNum;
	if (tempatomnum > j) {
	  glVertex3fv(allNodes[rootNode]->myData[j].xyz);
	  glVertex3fv(allNodes[rootNode]->myData[tempatomnum].xyz);
	}
      }
  }
  
  glEnd();
  
  glEnable(GL_LIGHTING);
  
  glEndList();
  
}

void displayFunnelNum(int g)
{
  vector<int> roots, funnel;
  char filename[100];

  loadFunnel(roots, "G.txt");

  //  if (displaySpecificNodesList == -1) 
  //    {
  cerr << "beware : i am not deleting prev. lists" << endl;
  displaySpecificNodesList =     glGenLists(1);
  numSpecificNodes         = 1;
      //    }

  if ((g >= 0) && (g <= 5))
    {
      sprintf(filename, "funnel%d.txt", g+1);
    }
  else
    {
      cout << "error. funnel # = " << g << " too big. should be <= 5" << endl;
      return;
    }
  
  loadFunnel                        (funnel, filename);
  addFunnelNodesAndRootToDisplayList(funnel, roots[g]);
  displaySpecificNodesOn = 1;
  display();
}

void  dontDisplayFunnel()
{
  displaySpecificNodesOn = 0;
  display();
}

void displayHighStatDistNodes()
{
  cout << "this is disabled." << endl;

  vector<int> nodeList;
  loadFunnel(nodeList, "highestStatDist.txt");

  if (displayHSDNList == -1) 
    {
      displayHSDNList =
	glGenLists(1);
    }
  glNewList(displayHSDNList, GL_COMPILE);

  glDisable(GL_LIGHTING);

  glBegin(GL_POINTS);
  glColor3f(1.0, 1.0, 1.0);
  
  for (unsigned i = 0; i < nodeList.size(); i++) 
    {
      //if (allNodes[i]->myTree->numNodesInTree < 100) {
      //glColor3fv(colorTable[allNodes[i]->myTree->myTreeName]);
      assert (nodeList[i] <numNodes);
      
      glVertex3fv(allNodes[nodeList[i]] ->cog);
      //    }
    }
  
  glEnd();
  
  glEnable(GL_LIGHTING);
  
  glEndList();
  //insertToNodeList(funnel2, funnel1.size(), 0.0, 1.0, 0.0);
  
  displayHSDNOn = 1;
  display();

}


void drawMain(char **argv) {

  //char str[200];

  //  InitPosition(65, 281, 720, 720);
  InitPosition(65, 281, 400, 400);
  My_Tk_Init(1, argv, Tcl_AppInit);   
  setupGLWindow("Bring on the dancing proteins!"); 

  init();
  
  SOLID_SPHERE_LIST = glGenLists(1);
  goalLigandList = glGenLists(1);
  currentLigandList = glGenLists(1);
  
  makeSolidSphere(1.25);
  myReceptor.createObjList();
  makeGoalLigand(origAtomData);  

  addReshapeFunc(myReshape);
  setBindings();


  //displaySpecificNodes(funnel1, 1.0, 0.0, 0.0);
  //displaySpecificNodes(funnel2, 0.0, 1.0, 0.0);
  
  display();
  
  Tk_MainLoop(); 
} 




  
