#include <iostream>
#include <iomanip>
#include <fstream>
using std::ifstream;
using std::ios;
#include <sstream>
#include <cstring>
#include <cmath>
#include <cstdlib>
#include <cstdio>
#include <time.h>
#include <float.h>
#include "ldock.H"
#include "ioRoutines_uniform.H"
#include <vector>
using std::vector;
#include "analysis.H"
#include <ctype.h>
#include "roadmapConstruction.H"
#include <iostream>
using std::cout;
using std::endl;
#include "debug.H"
#include "ligand.H"

extern ligAtom *myLigand;
extern int globalNumAtoms;
extern receptorClass myReceptor;
extern bitmapClass myBitmap;
extern atomData *origAtomData;
extern int globalRootAtom;
extern float MAX_XYZ;
extern float MIN_X, MIN_Y, MIN_Z;
extern float MAX_X, MAX_Y, MAX_Z;
extern int globalNumDOF;
extern int *globalConnections[3];
extern int globalNumConnects;

extern float distanceBetweenVecs(VEC, VEC);
extern void computeCOG(nodeData * node);
extern void sampleCOG(VEC retval, float samplingRadius, VEC origin);
extern void  VecSub(VEC Result, const VEC A, const VEC B);
extern int VecClose (const VEC A, const VEC B, float epsilon);
extern void printVec(VEC);
float findRMSD(atomData *a1, atomData *a2);



int MAXLINKS;
int INIT_NODES;
int MIN_REGIONS = 0;
const int maximumMIN_REGIONS = 30;

nodeData* allNodes[MAXNODES];
int linkTimeArray[MAXNODES];
int numNodes = 0;
treeClass *globalFirstTree;
treeClass *globalLastTree;
int numTrees = 0;
int globalTreeName = 0;
neighbourType neighbours[MAXNODES];
neighbourType nodeEnergyList[MAXNODES];
neighbourType sampledAroundList[maximumMIN_REGIONS + 1];
int nodeEnergyListCount;
int numSampledAround;
int rootSampleIndex;
int numNeighbours;
float globalGridResolution;
atomData* pathConfigurations[MAXNODES];
queType *myQue;
float currentTime;
int globalPathLength;
int globaldebugCheckPath = 0;
int totalExtraNodes = 0;

bool WRITE_TRANSITION_PROBS;



void initNodeColor(int);
void initQue(float, nodeClass*, nodeClass*);
void addToQue(float, nodeClass*, nodeClass*);
void runBFS(nodeClass*);
void findMinEnergyLigs();


/** reads the grid coordinates, from .grd.info file. The grid file format is:
    minX ignored_number
    minY ignored_number
    minZ ignored_number
    max_XYZ ignored_number ignored_number
    grid_resolution

    the grid is supposed to be cubical, and with the same resolution in each
    dimension. all coordinates read from the .atm and .lig files are converted
    later using MINX, MINY and MINZ into the relative reference frame of the grid.
*/
    
void fixMaxMinValues(char *filename) {
  
  ifstream gridfile;
  float temp;

  gridfile.open(filename, ios::in);

  gridfile >> MIN_X >> temp;
  gridfile >> MIN_Y >> temp;
  gridfile >> MIN_Z >> temp;
  
  MIN_X = -MIN_X;
  MIN_Y = -MIN_Y;
  MIN_Z = -MIN_Z;

  gridfile >> MAX_XYZ;
  gridfile >> temp;
  gridfile >> temp;

  gridfile >> globalGridResolution;

  gridfile.close();

  cout << MIN_X << "\t" << MIN_Y << "\t" << MIN_Z << endl;
  cout << MAX_XYZ << endl;


}

///
void initialize(char ** argv)
{
  char tempStr[200];

  srandom(time(NULL));
  cout.precision(6);
  cout.setf(ios::fixed, ios::floatfield);
  
  myLigand = new ligAtom;

  strcpy(tempStr, argv[1]);
  strcat(tempStr, ".lig");
  cout << "reading " << tempStr << endl;
  myLigand->readLigFile(tempStr);
  strcpy(tempStr, argv[1]);
  strcat(tempStr, ".atm");
  myReceptor.readReceptorFile(tempStr);
  strcpy(tempStr, argv[1]);
  strcat(tempStr, ".grd.info");
  fixMaxMinValues(tempStr);
  myLigand = myLigand->findGoodRoot(NULL);
  myLigand->fixLigand();
  myReceptor.fixReceptor();
  findNonBondedPairs();

  for (int i = 0; i < MAXNODES; i++) 
    {
      pathConfigurations[i] = NULL;
    }
  
  numSampledAround = 0;

  strcpy(tempStr, argv[1]);
  strcat(tempStr, ".grd.data");
  myReceptor.createPotential(globalGridResolution, tempStr);
  cout << "Bound State Energy = " << myReceptor.getEnergy(origAtomData) << endl;

}

///
int main(int argc, char **argv) 
{
  int regionCenterIndex;
  char complexName[200], roadmapFilename[100], 
    funnelCentersFile[100];
  vector<int> potBindingSiteIndices;
  const float samplingRadius = 15.0,
    //nodes sampled within this radius.
    rmsdRadius = 10.0; 
  //nodes within this rmsd of the catalytic site 
  //are going to be written to the output

  //  testLigandDOFs (argc, argv, &myLigand, allNodes);
  
  handleInput    (argc, argv, complexName,  roadmapFilename, 
		  funnelCentersFile,  &regionCenterIndex);

  initialize     (argv);
  //  analyzePairwiseAtomDistances
  //(globalNumAtoms, origAtomData);
  //  printLigandConnections
  //(globalNumConnects, globalConnections);
  assert ((regionCenterIndex >= 0) && (regionCenterIndex <= 4));

  if ((regionCenterIndex >= 1) && (regionCenterIndex <= 4))
    readGivenNode(funnelCentersFile,  allNodes, 
		  &numNodes, regionCenterIndex);
  else
    allNodes[0] = copyConfig(origAtomData);
  
  potBindingSiteIndices.resize(1); potBindingSiteIndices[0] = 0;

  createNewRoadmapRoutines(roadmapFilename, potBindingSiteIndices, 
			   samplingRadius, 0);

  writeRegions (potBindingSiteIndices, allNodes, numNodes, rmsdRadius);
  printRMSDs   (allNodes, numNodes, potBindingSiteIndices);
  printEnergies(allNodes, numNodes);
  writeTransitionProbabilities(numNodes, allNodes);

  return 0;
}

