#include "ioRoutines.H"
#include <cstdio>
#include <cstdlib>
#include <math.h>
#include "ldock.H"
#include <vector>
#include <iostream>
using std::cout;
using std::endl;
using std::cerr;
#include <cstdio>
#include <algorithm>
using std::min;
#include "analysis.H"
#include "roadmapConstruction.H"

#define SUCCESS 1
#define FAILURE 0

extern receptorClass myReceptor;
extern bool WRITE_TRANSITION_PROBS;
extern int globalNumAtoms;
extern int numNodes;
extern int numTrees;
extern int INIT_NODES;
extern int rootSampleIndex;
extern neighbourType sampledAroundList[];
extern int numSampledAround;
extern treeClass *globalFirstTree;
extern treeClass *globalLastTree;
extern int globalTreeName;
extern int totalExtraNodes;

extern float findRMSD            (atomData *a1, atomData *a2);
extern float distanceBetweenNodes(atomData *a1, atomData *a2);
extern int   MAXLINKS;
extern int   INIT_NODES;
extern int   MIN_REGIONS;



/*
  reads connectivity info from the file.
 */
void  readConnectivityInfo(int nodeIndex, vector<int> & adjacentNodeList, 
			   vector<float> & pTransArray, 
			   FILE * fin,  nodeData ** allNodes,
			   vector<vector< int> >   & connectedNodes,
			   vector<vector< float> > & edgeWeights,
			   int numEdges);
/*
  reads one line, and stores it in allNodes[*nodeIndex].
  the line format is described in readMyRoadmap.
*/

int readOneLineOfMyRoadmap (FILE * fin, int * nodeIndex, 
			    nodeData ** allNodes, 
			    vector<vector< int> >   & connectedNodes, 
			    vector<vector< float> > & edgeWeights)
			    //			    const char * relDir)
{
  float energy;
  int   numEdges;
  vector<float> pTransArray;
  vector<int> adjacentNodeList;

  
  fscanf(fin, "%d", nodeIndex);

  // in case eof not encountered yet..
  
  if (feof(fin))
    return FAILURE;
  
  // nodes should already be allocated in allocateNodes.
  assert (allNodes[(*nodeIndex)] != NULL);
  
  allNodes[(*nodeIndex)]->myData = new atomData[globalNumAtoms];
 
	  
  allNodes[(*nodeIndex)]->numLinks = 0;
  
  
  readAtomCoordsInfo(allNodes[*nodeIndex], fin);

  // connecting every node to other nodes, including node # 0.    
     
  int dummyInt1, dummyInt2; 
  fscanf  (fin, "%d %d", &dummyInt1, &dummyInt2);
  assert   ((dummyInt1 == -1) && (dummyInt2 == -1));
  
  fscanf   (fin, "%f %d", &energy, &numEdges);
  allNodes [(*nodeIndex)]->nodeEnergy = energy;
  
  readConnectivityInfo(*nodeIndex, adjacentNodeList, pTransArray, fin, 
		       allNodes, connectedNodes, edgeWeights, numEdges);

  //#if WRITE_TRANSITION_PROBS
  if (WRITE_TRANSITION_PROBS)
    {
      char filename[100];
      //      strcpy (filename, relDir);
      strcpy (filename, "transitionProbs.txt");

      FILE * fout = fopen(filename, "a");
      assert (adjacentNodeList.size() ==  pTransArray.size());

      for (unsigned int i = 0; i < pTransArray.size(); i++)
	{
	  //      pTransArray [i] /= normalizingFactor;
      
	  fprintf(fout, "%d %d %f\n", (*nodeIndex), 
		  adjacentNodeList[i],
		  pTransArray [i]);
	}
      fclose(fout);
    }
  //#endif  
  return SUCCESS;
}


// allocates and sets the initial values
void initializeMatrices(vector<vector< int> > & connectedNodes, 
			vector<vector< float> > & edgeWeights , 
			int numNodes)
{
  connectedNodes.resize(numNodes);
  edgeWeights   .resize(numNodes);
  
  for (int k = 0; k < numNodes; k++)
    {
      connectedNodes[k].resize(numNodes);
      edgeWeights[k]   .resize(numNodes);

      for (int l = 0; l < numNodes; l ++)
	{
	  connectedNodes[k][l] = 0;
	  edgeWeights[k][l]    = -1.0;
	}
    }
}

/*

  reads the roadmap file, which has the following format:
  
  the map format is as follows:

  randomNumNodes (i.e. 2000) (corresponds to INIT_NODES)
 
  numSampledAround (i.e. 20)
  total num. extra nodes sampled 
  for each region sampled around, the index, the linked bit, and the energy, followed by -1.

  then, in the next line:
     (# of nodes) 
  then, in the next line:
  num potential binding sites 
  indices of each of the potential binding sites    
  then followed by the detail information for each node, the format is:

  //(index) coords for each atom <5 floats x numAtomsInLigand> (cog (3 floats)); -1 -1 (potential) (# of nodes connecting to this node) (the adjacent list) 
    (index) coords for each atom <(5 floats -1) x numAtomsInLigand and >; (cog (3 floats)); -1 -1 (potential) (# of nodes connecting to this node) (the adjacent list) 

    the adjacent list format is:  (neighour node index) 1 (edge weight)
    
  establishes the nodes and the connections.

//  the first node is supposed to be already read and allocated.
 
  (index = 0..numNodes - 1).   

  establishes the nodes and the connections.

//  sets also rootsampleindex.
*/

void readMyRoadmap    (int * numberOfNodes, const char * filename, 
		       nodeData ** allNodes,
		       //		       const char * relDir, 
		       vector<int> & potBindingSiteIndices)
{

  FILE * fin =  fopen (filename,"r");
  vector<vector< int> > connectedNodes; // for debugging,
  vector<vector< float> > edgeWeights;
  int   nodeIndex;//, numAtomsInLigand;

  if (WRITE_TRANSITION_PROBS)
    {
      char    filename[100];
      strcpy (filename, "transitionProbs.txt");
      FILE * fout = fopen(filename, "w");
      fclose(fout);
      cout << "check " << filename << endl;
    }

  assert (*numberOfNodes == 0);
  
  if (fin == NULL)
    {
      cout << "couldn't open " << filename << ". exiting." << endl;
      exit(0);
    }
  else
    {
      cout << "reading " << filename << endl;
    }
  
  readRoadmapFileHeaderParameters (numberOfNodes, fin, potBindingSiteIndices);

  cout << "num. nodes to be read = " << *numberOfNodes << endl;
  cout << "-" << endl;
  numNodes = *numberOfNodes;
  initializeMatrices(connectedNodes, edgeWeights, *numberOfNodes);
  cout << "-" << endl;
  allocateNodes     (*numberOfNodes, allNodes, 0);
  cout << "-" << endl;

  int i = 0; // just to check the indices of the array.

  while (!feof(fin))
    {
      if (!readOneLineOfMyRoadmap (fin, &nodeIndex, allNodes, 
				   connectedNodes, edgeWeights))
								  //  relDir))
	break;
      
      assert (nodeIndex == i);
      i++;
      //cout << "i = " << i << endl;
    }
  
  cout << "-" << endl;
  if (*numberOfNodes != i)
    {
      cerr << "alert!. # nodes read = " << nodeIndex << " whereas the file is supposed to contain "   << numberOfNodes  << " nodes." << endl;
      assert (0);
    }
  
  //#if WRITE_TRANSITION_PROBS
  /*
    vector<int> indices;
    unsigned int  K = 10;
  
    findMinEnergyClusters    (indices,  K, 
    allNodes,   numNodes - 1000,
    5.0, 0);
    
  
    fout = fopen("transitionProbs.txt", "a");
    for (int i  = 0; i < K; i++)
    {
    fprintf(fout, "%d\n", indices[i]);
    }
    fclose(fout);
  */
  
  //  cout << "wrote transition probabilities.. exiting." << endl;
  //cout << "wrote edge weights and the lowest 10 energy configs." << endl;
  //  exit(0);
  //#endif
  
  doConnections4MyRoadmap(edgeWeights, *numberOfNodes, allNodes);
  cout << "-" << endl;
  fclose(fin);
  //  (*numberOfNodes) = numNodes;
  cout << "read a roadmap of " << *numberOfNodes << " nodes." << endl;
}

/*
  same format as in readMyRoadmap.
  writes the roadmap.

  also, if  WRITE_TRANSITION_PROBS is set, write them to
  transitionProbs.txt. These probs include self stay prob.
*/

void writeMyRoadmap    (int numNodes, const char * filename, 
			nodeData** allNodes, 
			int numRandomNodes, 
			int numSampledAround,
			int totalExtraNodes,
			//			char * relDir,
			const vector<int> & potBindingSiteIndices)
{
  FILE * fout =  fopen (filename,"w");
  FILE * f2;

  vector<float> edgeWeights;
  vector<int>   neighborIndices;
    
  cout << "check " << filename << endl;

  fprintf(fout, "%d\n", numRandomNodes);
  fprintf(fout, "%d\n", numSampledAround);
  fprintf(fout, "%d\n", totalExtraNodes);
  fflush (fout);

  for (int i = 0; i < numSampledAround; i++)
    {
      fprintf(fout, "%d %d %f -1 ", sampledAroundList[i].index,
	      sampledAroundList[i].linkedBit,
	      sampledAroundList[i].dist);
      fflush(fout);
    }
  fprintf(fout, "\n");
  
  fprintf(fout, "%d\n", numNodes);
  fflush(fout);

  fprintf(fout, "%d\n", potBindingSiteIndices.size());
  for (unsigned int i = 0; i < potBindingSiteIndices.size(); i++)
    {
      fprintf(fout, "%d ", potBindingSiteIndices[i]);
    }
  fprintf(fout, "\n");

  //#if WRITE_TRANSITION_PROBS
  if (WRITE_TRANSITION_PROBS)
    {
      char filename[100];
      //strcpy (filename, relDir);
      strcpy (filename, "transitionProbs.txt");
      f2 = fopen(filename, "w");
      assert (f2 != NULL);
      cout << "check " << filename << endl;
    }
  //#endif

  for (int i = 0; i < numNodes; i++)
    {
      fprintf(fout, "%d ", i);
      for (int j = 0; j < globalNumAtoms; j++)
	{
	  float x,y,z, angle1, angle2; 
	  // coordinates for each ligand atom

	  x      = allNodes[i]->myData[j].xyz[0];
	  y      = allNodes[i]->myData[j].xyz[1];
	  z      = allNodes[i]->myData[j].xyz[2];
	  angle1 = allNodes[i]->myData[j].angle[0];
	  angle2 = allNodes[i]->myData[j].angle[1];
	  //cout << " ioRoutines_uniform.C: i = " << i << " j = " << j << " angle2 = " << angle2 << endl;
	  fprintf(fout, "%f %f %f %f %f -1 ", x, y, z, angle1, angle2);
	}
      
      fprintf(fout, "%f %f %f ", allNodes[i]->cog[0], allNodes[i]->cog[1],
	      allNodes[i]->cog[2]);
      
      
      fprintf(fout, "-1 -1 %f ", allNodes[i]->nodeEnergy);
      
      findNeighborsAndEdgeWeights (neighborIndices,
				   edgeWeights,
				   i, 
				   numNodes, 
				   allNodes);
      
      
      int numNeighbors = neighborIndices.size();
      //#if WRITE_TRANSITION_PROBS
      
      if (WRITE_TRANSITION_PROBS)
	{
	  vector<float> pTransArray;
	  vector<int> adjacentNodeList;
      
	  pTransArray.resize(0);adjacentNodeList.resize(0);
	  float pTrans;

	  for (int j = 0; j < numNeighbors; j++)
	    {
	      pTrans                                        = exp(-edgeWeights[j]);
	      pTransArray.push_back(pTrans);
	      adjacentNodeList.push_back(neighborIndices[j]);


	    }
	  
      
	  assert (adjacentNodeList.size() ==  pTransArray.size());
	  for (unsigned int j = 0; j < pTransArray.size(); j++)
	    {

	      
	      fprintf(f2, "%d %d %f\n", i,  
		      //neighborIndices[j],
		      adjacentNodeList[j],
		      pTransArray [j]);
	    }
	}
      //#endif
      
      fprintf(fout, "%d ", numNeighbors);
      
      for (int j = 0; j <  numNeighbors; j++)
	{
	  fprintf(fout, "%d 1 %f ", neighborIndices[j], edgeWeights[j]);
	}
      fprintf(fout, "\n");
    }
  fclose(fout);
  //#if WRITE_TRANSITION_PROBS
  if (WRITE_TRANSITION_PROBS)
    {
      fclose(f2);
    }
  //#endif
}

void printEnergies(nodeData ** allNodes, int numberOfNodes)
{
  FILE * fout = fopen("energies.txt", "w");
  cout << "check energies.txt" << endl;
  for (int i = 0; i < numberOfNodes; i++)
    {
      fprintf(fout, "%f\n", allNodes[i]->nodeEnergy);
    }
  fclose(fout);
}

/*
  PRINTS THE RMSD OF EACH NODE TO NODEs in centers.
*/
void printRMSDs(nodeData ** allNodes, int numberOfNodes,  
		const vector<int> & centers)
{
  FILE * fout = fopen("rmsds.txt", "w");
  cout << "check rmsds.txt" << endl;

  for (int i = 0; i < numberOfNodes; i++)
    {
      for (unsigned int j = 0; j < centers.size(); j++)
	{
	  fprintf(fout, "%f ", findRMSD(allNodes[centers[j]]->myData, 
				       allNodes[i]->myData));
	}
      fprintf(fout, "\n");
    }

  fclose(fout);
}


/*
    prints funnel centers info (coords and energies) to the output file. (potBindingSiteIndices.txt).
   The first 3 lines contain: numFunnelCenters:
                              %d
			      ----------------
*/
void printFunnelCenterInfo (const vector<int> & potBindingSiteIndices,
			    nodeData ** allNodes)
			    
{
  char filename[100];

  float rmsd, minRMSD = 100.0;
  int   minRMSDindex = -1;

  
  /*
  if (strcmp(relDir, "") == 0)
    {
      strcpy(relDir, ".");
    }
    strcpy(filename, relDir);
  */
  strcpy(filename, "potBindingSiteIndices.txt");
  
  FILE * fout = fopen(filename, "w");
  cout << "check " << filename << endl;

  fprintf(fout, "num potential binding sites:\n");
  fprintf(fout, "%d\n", potBindingSiteIndices.size());
  fprintf(fout, "---------------\n");

  for (unsigned int i = 0; i < potBindingSiteIndices.size(); i++) 
    {
      int node1 = potBindingSiteIndices[i];
      fprintf(fout, "%d ", i);

      rmsd = distanceBetweenNodes(allNodes[0]->myData, 
				  allNodes[potBindingSiteIndices[i]]->myData);
      
      if (rmsd < minRMSD)
	{
	  minRMSD      = rmsd;
	  minRMSDindex = i;
	}
      
      for (int j = 0; j < globalNumAtoms; j++)
	{
	  float x,y,z, angle1, angle2; 
	  // coordinates for each ligand atom
	  
	  x      = allNodes[node1]->myData[j].xyz[0];
	  y      = allNodes[node1]->myData[j].xyz[1];
	  z      = allNodes[node1]->myData[j].xyz[2];
	  angle1 = allNodes[node1]->myData[j].angle[0];
	  angle2 = allNodes[node1]->myData[j].angle[1];
	  
	  fprintf(fout, "%f %f %f %f %f -1 ", x, y, z, angle1, angle2);
	}
      
      fprintf(fout, "%f %f %f ", allNodes[node1]->cog[0], allNodes[node1]->cog[1], allNodes[node1]->cog[2]);
      
      fprintf(fout, "-1 -1 %f \n", allNodes[node1]->nodeEnergy);
    }

  cout << "d(the pot. binding site closest to the catalytic site, catalytic site) =  ";
  cout << minRMSD << endl;
  cout << "its energy = " ;
  cout << allNodes[potBindingSiteIndices[minRMSDindex]]->nodeEnergy << endl;
  
}



/*
  reads atom coords from the opened file.
  does not read connection info.
*/
void readAtomCoordsInfo(nodeData * node,  FILE * fin)
{
  assert (fin  != NULL);
  assert (node != NULL);
  assert (node->myData != NULL);
  
  for (int i = 0; i < globalNumAtoms; i++)
    {
      float x,y,z, angle1, angle2; 
      int temp;
      // coordinates for each ligand atom

      fscanf(fin, "%f %f %f %f %f %d ", &x, &y, &z, &angle1, &angle2, &temp);
      //      cout << x<< " " << y<< " " <<z<< " " << angle1<< " " << angle2 << endl; 
      assert (temp == -1);
      node->myData[i].xyz[0] = x;
      node->myData[i].xyz[1] = y;
      node->myData[i].xyz[2] = z;
      node->myData[i].angle[0] = angle1;
      node->myData[i].angle[1] = angle2;
    }
  

  float cogx, cogy, cogz;
  fscanf(fin, "%f %f %f", &cogx, &cogy, &cogz);

  //  if ((*nodeIndex) != 0)
  //    {
  node->cog[0] = cogx;
  node->cog[1] = cogy;
  node->cog[2] = cogz;

}

/*
  reads and sorts the indices.
*/
void  readNSortIndices(vector<int> & indices, const char * indexFilename)
{


  FILE * fin2 = fopen(indexFilename, "r");

  if (fin2 == NULL)
    {
      cout << "couldn't open " << indexFilename << ". exiting." << endl;
      exit(0);
    }
  else
    {
      cout << "reading " << indexFilename << endl;
    }
  
  while (!feof(fin2))
    {
      int index;
      fscanf (fin2, "%d\n", &index);
      // indices are one based.
      indices.push_back(index);
      assert (index >= 1); 
    }
  
  fclose(fin2);

  for (unsigned int i = 0; i < indices.size(); i++)
    {
      for (unsigned int j = i+1; j < indices.size(); j++)
	{
	  if (indices[i] > indices[j])
	    {
	      int temp         = indices[i];
	      indices[i]       = indices[j];
	      indices[j]       = temp;
	    }
	}
    }
}

/*
  reads and stores node info. given in funnelCentersFile.
  node info format is assumed to be the same as the roadmap format, except 
  that the link info is omitted. (is not present in i/p file.)
  sets potBindingSiteIndices to 0..numPotBindingSites.
  potBindingSites are supposed to contain node # 0, the true binding site,
  and it is added automatically.
*/

void readGivenNodes(char * funnelCentersFile, nodeData ** allNodes, 
		    int  * numNodes, vector<int> & potBindingSiteIndices)
{
  int  numFunnelCenters, numNodesRead = 0, nodeIndex;
  char line[200];
  float energy, tempEnergy;

  FILE * fin =  fopen (funnelCentersFile,"r");
  if (fin == NULL)
    {
      cout << "couldn't open " << funnelCentersFile << ". exiting." << endl;
      exit(0);
    }
  else
    {
      cout << "reading " << funnelCentersFile << endl;
      fgets (line, 200, fin);
      fscanf(fin, "%d\n", &numFunnelCenters);
      fgets (line, 200, fin);
      cout << "will read " << numFunnelCenters << " centers. " << endl;
    }
  
  assert (*numNodes == 1);
  potBindingSiteIndices.resize(0);
  potBindingSiteIndices.push_back(0); // adding true binding site.
  for (int i = 0; i < numFunnelCenters; i++)
    {
      potBindingSiteIndices.push_back(i+1); // the first binding site is node # 1.
    }
  
  allocateNodes   (numFunnelCenters, allNodes, (*numNodes));
  
  while (numNodesRead < numFunnelCenters)
    {
      fscanf(fin, "%d", &nodeIndex);
 
      allNodes[(*numNodes)]->myData   = new atomData[globalNumAtoms];
      allNodes[(*numNodes)]->numLinks = 0;
	  
      readAtomCoordsInfo(allNodes[(*numNodes)], fin);
	  
      int dummyInt1, dummyInt2; 
      fscanf  (fin, "%d %d", &dummyInt1, &dummyInt2);
      assert   ((dummyInt1 == -1) && (dummyInt2 == -1));

      //	  read energy;
      fscanf   (fin, "%f ", &energy);
      allNodes [(*numNodes)]->nodeEnergy = energy;
	  
      cout << "reading node # " << (*numNodes) << " info. energy = ";
      cout << energy << endl;
	  
      //	  check energy;
      myReceptor.checkCollision(allNodes[*numNodes]->myData, &tempEnergy);
      if (fabs(allNodes[(*numNodes)]->nodeEnergy - tempEnergy) >= 0.5)
	{
	  cout << "read energy = " << energy << " computed : " << tempEnergy << endl;
	  assert (0);
	}
	  
      fscanf(fin, "\n");//, temp);

      // check for whether the node is the true binding site, if yes, abort
      if (findRMSD(allNodes[0]->myData, allNodes[*numNodes]->myData) < 0.01)
	{
	  cout << "error! pot. binding sites are not supposed to also include";
	  cout << endl << "true binding site. exiting..." << endl;
	  exit(-1);
	}
      numNodesRead ++;
      (*numNodes)++;
    }

  fclose(fin);
}

/*
  reads and stores node info. given in funnelCentersFile.
  node info format is assumed to be the same as the roadmap
  format, except that the link info is omitted. 
  (is not present in i/p file.)
  sets potBindingSiteIndices to funnelCenterIndex (between 1..4).

  note: funnelCenterIndex == 1 means first potential binding site is going to be read from funnelCentersFile.
1st pot binding site, etc.
*/

void readGivenNode(char * funnelCentersFile, nodeData ** allNodes, 
		   int  * numNodes, 
		   int funnelCenterIndex)
{
  int  numFunnelCenters,  nodeIndex;
  char line[2000];
  float energy, tempEnergy;

  assert ((funnelCenterIndex >= 1) && (funnelCenterIndex <= 4));

  FILE * fin =  fopen (funnelCentersFile,"r");
  if (fin == NULL)
    {
      cout << "couldn't open " << funnelCentersFile << ". exiting." << endl;
      exit(0);
    }
  else
    {
      cout << "reading " << funnelCentersFile << endl;
      fgets (line, 2000, fin);
      fscanf(fin, "%d\n", &numFunnelCenters);
      fgets (line, 2000, fin);
    }
  
  assert (numFunnelCenters == 4);

  cout << "funnel center index = " << funnelCenterIndex << endl;

  assert (*numNodes == 0);
  allocateNodes   (1, allNodes, 0);

  int numLinesRead = 0;
  bool readFunnelCenter = false;
  while ((!readFunnelCenter) &&  (numLinesRead < 100))
    {
      fscanf(fin, "%d", &nodeIndex);
	  
      cout << "node index = " << nodeIndex << endl;
      if ((nodeIndex + 1) == funnelCenterIndex) 
      	{
	  readFunnelCenter  = true;
	  cout << "found the node." << endl;
	  allNodes[(*numNodes)]->myData   = new atomData[globalNumAtoms];
	  allNodes[(*numNodes)]->numLinks = 0;
	  
	  readAtomCoordsInfo(allNodes[(*numNodes)], fin);
	  
	  int dummyInt1, dummyInt2; 
	  fscanf  (fin, "%d %d", &dummyInt1, &dummyInt2);
	  assert   ((dummyInt1 == -1) && (dummyInt2 == -1));
	      
	  //	  read energy;
	  fscanf   (fin, "%f ", &energy);
	  allNodes [(*numNodes)]->nodeEnergy = energy;
	      
	  cout << "reading node # " << (*numNodes) << " info. energy = ";
	  cout << energy << endl;
	  cout << "writing energy again = " <<  allNodes [(*numNodes)]->nodeEnergy << endl;
	  
	  //	  check energy;
	  myReceptor.checkCollision(allNodes[*numNodes]->myData, &tempEnergy);
	  if (fabs(allNodes[(*numNodes)]->nodeEnergy - tempEnergy) >= 0.05)
	    {
	      cout << "read energy = " << energy << " computed : " << tempEnergy << endl;
	      assert (0);
	    }
	  (*numNodes)++;
	  break;
	}
      else
	{
	  fgets (line, 2000, fin);
	}
      numLinesRead++;
    }
  assert (numLinesRead < 100);

  cout << "num nodes at end of reading nodes = " << *numNodes << endl;
  assert (*numNodes == 1);

  fclose(fin);
}

/*
    given a roadmap file, goes thru the unnecessary info
    and picks the nodes with given nodeIndices. Adds them to the 
    roadmap, starting from node # startNodeIndex. 
    Assumes the same format as in readMyRoadmap.
    updates the global variable numNodes.

    node indices are assumed to be 1 based.
*/
void readGivenListOfNodes(const char * roadmapFilename,
			  nodeData ** allNodes, 
			  const char * indexFilename,
			  vector<int> & indices, 
			  int startNodeIndex)
{
  FILE * fin =  fopen (roadmapFilename,"r");
  int  dummyInt, numEdges, N; // N is numsampledaround.
  char temp[1500];
  float energy;

  indices.resize(0);

  if (fin == NULL)
    {
      cout << "couldn't open " << roadmapFilename << ". exiting." << endl;
      exit(0);
    }
  else
    {
      cout << "reading " << roadmapFilename << endl;
    }
  
  readNSortIndices(indices, indexFilename);
  for (unsigned int i = 0; i < indices.size(); i++)
    {
      indices[i] = indices[i] - 1;
      assert     (indices[i] > 0);
    }

  assert          (startNodeIndex >= 0);

  allocateNodes   (indices.size(), allNodes, startNodeIndex);
  
  fscanf          (fin, "%d\n", &dummyInt);
  fscanf          (fin, "%d\n", &N); 
  // N is numsampled around, but it
  // is not utilized.
  fscanf          (fin, "%d\n", &dummyInt);
  
  for (int i = 0; i < N; i++)
    {
      float dist;
      int   index, linkedBit, temp;
      
      fscanf        (fin, "%d %d %f %d ", &index, &linkedBit, &dist, &temp);
      assert        (temp == -1);
    }
  
  fscanf   (fin, "\n");
  fscanf   (fin, "%d\n", &dummyInt);

  int prevNodeIndex = -1, nodeIndex, numNodesRead = 0;

  
  while (numNodesRead < indices.size())
    {
      fscanf(fin, "%d", &nodeIndex);
      //cout << "nodeIndex = " << nodeIndex << endl;
      assert (prevNodeIndex + 1 == nodeIndex);
      prevNodeIndex ++;

      if (nodeIndex == indices[numNodesRead])
	{
	  int newNodeIndex = numNodesRead + startNodeIndex;

	  allNodes[newNodeIndex]->myData   = new atomData[globalNumAtoms];
	  allNodes[newNodeIndex]->numLinks = 0;
	  
	  readAtomCoordsInfo(allNodes[newNodeIndex], fin);
	  
	  int dummyInt1, dummyInt2; 
	  fscanf  (fin, "%d %d", &dummyInt1, &dummyInt2);
	  assert   ((dummyInt1 == -1) && (dummyInt2 == -1));

	  //	  read energy;
	  fscanf   (fin, "%f %d", &energy, &numEdges);
	  allNodes [newNodeIndex]->nodeEnergy = energy;
	  
	  cout << "reading node # " << nodeIndex << " info. energy = ";
	  cout << energy << endl;
	  
	  //	  check energy;
	  if (fabs(allNodes[newNodeIndex]->nodeEnergy - energy) >= 0.5)
	    {
	      cout << "read energy = " << energy << " computed : " << allNodes[nodeIndex]->nodeEnergy << endl;
	      assert (0);
	    }
	  
	  // fscanf(fin, "%s", temp);
	  //fscanf(fin, "\n");//, temp);
	  fgets(temp, 1500, fin);
	  //	  cout << "discarding the following conn. info : " << temp << endl;
	  numNodesRead ++;
	}
      else
	{
	  // fscanf(fin, "%s", temp);
	  fgets(temp, 1500, fin);
	  // fscanf(fin, "\n");//, temp);
	  // cout << "discarding the following complete node info : " << temp << endl;
	}
    }
  
  numNodes += numNodesRead;
  cout << "read " << indices.size() << " nodes." << endl;
  fclose(fin);
}


/*
  reads the parameters written on top of the file header.
 */
void readRoadmapFileHeaderParameters (int * numberOfNodes, FILE * fin, 
				      vector<int> & potBindingSiteIndices)
{
  int numPotBindingSiteIndices;

  fscanf            (fin, "%d\n", &INIT_NODES);
  fscanf            (fin, "%d\n", &numSampledAround);
  fscanf            (fin, "%d\n", &totalExtraNodes);

  //  rootSampleIndex = INIT_NODES;

  for (int i = 0; i < numSampledAround; i++)
    {
      float dist;
      int   index, linkedBit, temp;
      
      fscanf        (fin, "%d %d %f %d ", &index, &linkedBit, &dist, &temp);
      assert        (temp == -1);
      sampledAroundList[i].index       = index;
      sampledAroundList[i].linkedBit   = linkedBit;
      sampledAroundList[i].dist        = dist;
    }
  
  fscanf   (fin, "\n");
  fscanf   (fin, "%d\n", numberOfNodes);
  fscanf   (fin, "%d\n", &numPotBindingSiteIndices);

  potBindingSiteIndices.resize(0);
  for (int i = 0; i < numPotBindingSiteIndices; i++)
    {
      int index;
      fscanf            (fin, "%d ", &index);
      assert            (index < (*numberOfNodes));
      potBindingSiteIndices.push_back   (index);
    }

  for (unsigned int i = 0; i < potBindingSiteIndices.size(); i++)
    {
      cout << "pot. binding site index # " << i << " = " << potBindingSiteIndices[i] << endl;
    }
  
  fscanf   (fin, "\n");
}

/*
  reads connectivity info from the file.
*/
void  readConnectivityInfo(int nodeIndex, vector<int> & adjacentNodeList, 
			   vector<float> & pTransArray, 
			   FILE * fin, nodeData ** allNodes,
			   vector<vector< int> >   & connectedNodes,
			   vector<vector< float> > & edgeWeights,
			   int numEdges)
{

  float pTrans;



  assert (fin!= NULL);
  adjacentNodeList.resize(0);
  pTransArray.resize(0);

  // now reading the adjacency list
  for (int k = 0; k < numEdges; k++)
    {
      int   otherNodeIndex;
      float pW;
      int   dummyInt;
	  
      fscanf         (fin, "%d %d %f", &otherNodeIndex,&dummyInt, &pW);
      
      assert         (dummyInt == 1);
      
      adjacentNodeList.push_back(otherNodeIndex);
      assert (nodeIndex      < connectedNodes.size());
      assert (otherNodeIndex < connectedNodes[0].size());

      connectedNodes [nodeIndex][otherNodeIndex] = 1;
      edgeWeights    [nodeIndex][otherNodeIndex] = pW;
      pTrans                                     = exp(-pW);
      pTransArray.push_back(pTrans);
    }
}

/*
  prints indices of nodes that are around each of the potential binding sites.
  potential binding sites are nodeEnergyList[0]..nodeEnergyList[numPotBindingSites-1].
  check for whether they are close 2 each other!
  the radiusRMSD is a parameter that gives the distance of the furthest distant node from the binding site center.
  the node indices are written in 0-based format, and to files
  close2_%d.txt, 0..numPotBindingSites-1.
  The first line of these files will be "\% nodes close 2 node # %d, nodeEnergyList[i]".
*/
void printCloseNodeIndices(const vector<int> & potBindingSites, 
			   neighbourType * nodeEnergyList,
			   int nodeEnergyListCount,
			   float rmsdRadius,
			   int numNodes, 
			   nodeData ** allNodes)
{
  int numPotBindingSites = potBindingSites.size();

  int N = min<int> (numPotBindingSites, nodeEnergyListCount); 
  vector<int> v; // contains indices of nodes to be printed for node # j.

  for (int j = 0; j < N;  j++)
    {
      v.resize(0);
      for (int i = 0; i < numNodes; i++)
	{
	  assert (nodeEnergyList[j].index < numNodes);
	  assert (nodeEnergyList[j].index >= 0);
	  if (findRMSD(allNodes[i]->myData, 
		       allNodes[nodeEnergyList[j].index]->myData) < rmsdRadius)
	    {
	      //   put node i in the vector to be printed for node # nodeEnergyList[j]]->index
	      v.push_back(i);
	    }
	}

       
      char filename[100];
      sprintf(filename, "close2PotBindingSite#%d.txt", j);
      FILE * fout = fopen(filename, "w");
      fprintf(fout, "%% nodes close (rmsd < %f) to node # %d\n", rmsdRadius, nodeEnergyList[j].index);
      
      for (unsigned int k = 0; k < v.size(); k++)
	{
	  fprintf(fout, "%d\n", v[k]);
	}
      fclose(fout);
    }
}

///
void writeRegions (const vector<int> & potBindingSiteIndices, 
		   nodeData ** allNodes, int numNodes, float rmsdRadius)
{
  // here printing all nodes within rmsdRadius of the pot. binding
  // sites
  vector<int> v;
  char filename[200];
  
  for (unsigned int i = 0; i < potBindingSiteIndices.size(); i++) 
    {
      v.resize(0);

      findNodesClose2Node(v,potBindingSiteIndices[i] , numNodes,
			  allNodes, rmsdRadius);

      sprintf(filename, "close2PotBindingSite#%d.txt", i);
      cout << "check " << filename << endl;
      FILE * fout = fopen(filename, "w");
      fprintf(fout, "%% nodes close (rmsd < %f) to node # %d\n",
	      rmsdRadius, potBindingSiteIndices[i]);
      
      for (unsigned int k = 0; k < v.size(); k++)
	{
	  assert ((v[k] >= 0) && (v[k] < numNodes));
	  
	  assert (findRMSD(allNodes[v[k]]->myData, 
			   allNodes[potBindingSiteIndices[i]]->myData) < rmsdRadius);

	  fprintf(fout, "%d\n", v[k]);
	}
      fclose(fout);
    }
  cout << "printed close nodes..." << endl;
}

 
/* 
   warning: roadmap edges contain transition probabilities, not edge weights.
 */
void writeTransitionProbabilities (int numNodes, 
				   nodeData ** allNodes)
{
  char filename[] = "transitionProbs.txt";
  FILE * fout     =  fopen (filename,"w");
  vector<float> edgeWeights;
  vector<int>   neighborIndices;
  
  cout << "check transitionProbs.txt" << endl;
  for (int i = 0; i < numNodes; i++)
    {
      findNeighborsAndEdgeWeights (neighborIndices,
				   edgeWeights,
				   i, 
				   numNodes, 
				   allNodes);
      assert (neighborIndices.size() == edgeWeights.size());
      
      for (unsigned int j = 0; j < neighborIndices.size() ; j++)
	{

	  assert (edgeWeights[j] >= 0);
	  assert (neighborIndices[j] != i);
	  
	  fprintf(fout, "%d %d %f \n", i, neighborIndices[j], edgeWeights[j]);
	}
    }
  fclose(fout);
}

/*
  reads simply roadmap and potential binding sites.
*/
void readRoadmapRoutines(//char relDir[], 
			 char roadmapFilename[], 
			 vector<int> & potBindingSiteIndices,
			 nodeData ** allNodes)
{
  //  how to specify the potential binding sites if you read a roadmap?

  readMyRoadmap      (&numNodes, roadmapFilename,  allNodes, //relDir,
		      potBindingSiteIndices);
  
  //set the numsampledaround, sampledaroundlist parameters from the read roadmap;

  /*
    // here check for the distance of the last 500 nodes from their center,
    // which are nodes 0..4.
  
    for (int i = 0; i < 5; i++) 
    {
  
      
    //      center at i
    float maxRMSD = 0.0;
    for (int j = numNodes - 500 + i * 100; j < numNodes - 400 + i * 100; j++)
    {
    // nodes at j
    float rmsd = findRMSD(allNodes[i]->myData, allNodes[j]->myData);
    cout << "rmsd between " << i << " and " << j << " is : " << rmsd << endl;
    if (rmsd > maxRMSD)
    {
    maxRMSD = rmsd;
    }
    }
    ofstream fout("maxRMSD.txt");
    fout << maxRMSD << endl;
    fout.close();
    }
  */
  
  
  

  //  setIndices(numNodes);
  if (globalFirstTree != NULL) 
    globalFirstTree->printTree();
  if (allNodes[0]->myTree != NULL) 
    {
      cout << "numNodes in the tree of native structure : " << allNodes[0]->myTree->numNodesInTree << endl;
    }

  //  analyzeHighStatDistNodes ("indices.txt");
  //  analyzeRMSDDistanceFromSeeds ("seeds.txt");
  
}




/* 
   sets variables from input, including INIT_NODES, MAXLINKS, MIN_REGIONS and WRITE_TRANSITION_PROBS.
*/
void handleInput (int argc, char ** argv, char * complexName,
		  char * roadmapFilename, 
		  char * funnelCentersFile,
		  int  * regionCenterIndex)
{
  if (argc != 7) 
    {
      cout << "Usage: " << argv[0] << " <complex name> <roadmap file name>  [#init nodes] [#dof] [funnelCentersFile] <regionCenterIndex> " << endl; 
      exit(0);
    }
  else
    {
      strcpy       (complexName    , argv[1]);
      strcpy       (roadmapFilename, argv[2]);
      INIT_NODES = (int) (atoi(argv[3]));
      MAXLINKS   = (int) (atoi(argv[4]));
      strcpy       (funnelCentersFile, argv[5]);
      cout << "will read funnel centers from " << funnelCentersFile << endl;
      *regionCenterIndex = (int)(atoi(argv[6]));
      cout << "region center index = " << *regionCenterIndex << endl;
    }
}

///
void printLigandConnections(int globalNumConnects,  int *globalConnections[3])
{
  cout << "there are total : " <<  globalNumConnects << " connections." << endl;
  for (int i = 0; i < globalNumConnects; i++) 
    {
      cout << globalConnections[i][0] << " " << globalConnections[i][1] << endl;
    }
}
