#ifndef _LDOCK_H
#define _LDOCK_H

#define PI 3.1415926535897
#define THETA_MAX 1.570796327
#define PHI_MAX 1.570796327
#define MY_RAND_MAX 2147483647.0
#define MAXNODES 20000
#define MAXNEIGHBOURS 40
#define MAXDOF 50
#define DISTTHRESHOLD 15.0
#define MAX_ATOMS 5500 //3000
#define WORDSIZE 32
#define EPSILON_DIST 1.0
#define XYZ_EPSILON 3.0
#define ANGLE_EPSILON 0.5 
#define XYZ_UNIT_STEP 1.0
#define ANGLE_UNIT_STEP 0.5 
#define MY_PH 7.4
#define MAX_ENERGY 5.0
#define MAX_PATH_LENGTH 10000.0
#define BOLTZMAN_K 0.596131676
#define INV_BOLTZMAN_K 1.677481737
#define ENERGY_WEIGHT 1.0
#define LENGTH_WEIGHT 0.0
//#define MIN_REGIONS 30
#define SAMPLES_PER_REGION 100

//define MAXLINKS 7
//define INIT_NODES 4000

#include <vector>
using std::vector;
#include <float.h>

typedef float MAT[4][4];
typedef float VEC[3];

typedef unsigned int bitVecType;


class ptrList;
class nodeClass;
class treeClass;
class ligAtom;
class receptorClass;
class bitClass;
class bitmapClass;
///
struct atomDataStruct {
  VEC xyz;
  float angle[2];
};
///
typedef struct atomDataStruct atomData;
///
struct neighbourStruct {
  int index;
  float dist;
  int linkedBit;
};
///
typedef struct neighbourStruct neighbourType;

///
class ligAtom {

public:
  int numBonds;
  int colorBit;
  ligAtom *bonds[4];
  float *bondLength;
  float *bondAngle;
  float *torAngle;
  int atomNum;

  ligAtom();
  ~ligAtom();

  void findAngles_root(atomData*);
  void findAngles(atomData*, int, int);
  void findXYZ_root(atomData*);
  void findXYZ_root(atomData *adata, VEC translationVec);
  void findXYZ(atomData*, int, int);
  void linkTree();
  void addBond(ligAtom*);
  void readLigFile(const char *);
  void fixLigand();
  void createRandomLigand_root(atomData*);
  void createRandomLigand(atomData*, int);
  void findNumDof_root();
  void findNumDof(int);
  void createRandomLigFromLig_root(atomData*, atomData*);
  void createRandomLigFromLig(atomData*, int, atomData*);
  void clearColors();
  ligAtom* findGoodRoot(ligAtom*);
};

///
struct nodeDataStruct {
  atomData *myData;
  VEC cog;
  nodeClass *myNode;
  treeClass *myTree;
  float nodeEnergy;
  int numLinks;
  int index;
};
///
typedef struct nodeDataStruct nodeData;

///
class ptrList {

public:
  ptrList *nextPtr;
  nodeClass *node;
  float pathWeight;
  ptrList();
};

///
class nodeClass {

public:
  ptrList *firstLink;
  ptrList *lastLink;
  nodeData *myConfig;
  nodeClass *pathPtr;
  char nodeColor;
  float nodeTime;

  nodeClass();
  void addNodeLink(nodeClass*, float);
  void replaceTreePtr(treeClass*, nodeClass*);
  void setPath(nodeClass*, float);
  void setPath_reverse(nodeClass*, float);
  float reverseWeight(nodeClass*);
  /*
    goes thru all nodes that it is connected to and prints the index of 
    these nodes.
  */
  void printTree(nodeClass *fromNode);
};
			 
///
class treeClass {

public:
  nodeClass *root;
  treeClass *nextTree;
  treeClass *prevTree;
  int numNodesInTree;
  int myTreeName;

  treeClass();
  void removeTree(treeClass*);
  void removeTree (treeClass* replacementTree,
		   const int    numNodes);
  void printTree();
};

///
class bitClass {

private:
  bitVecType *bv;
  int size;
  int numWords;

public:
  void initBitClass(int);
  void setBit(int);
  int getBit(int);
};


///
class bitmapClass {

private:
  bitClass map;
  float multFactor;
  int dimensionSize;

public:
  void initBitmap(float, float);
  void setBits(VEC);
  int checkBit(VEC);
};

///
class receptorClass {

public:
  VEC recAtom[MAX_ATOMS];
  float recAtomCharge[MAX_ATOMS];
  float recAtomSize[MAX_ATOMS];
  int numReceptorAtoms;
  bitmapClass myBitmap;
  float *elecPot;
  float *vanPot;
  int potMatSize;
  float potMatMultFact;
  int modelNumber;
  float center[3];
  float objMat[16];

  void readReceptorFile(const char*);
  void fixReceptor();
  void createBitmap(float);
  void createPotential(float, char*);
  void setVanPot(VEC, float);
  void setElecPot(VEC, float);
  int   checkCollision(atomData*, float*);
  float getEnergy(atomData*);
  void  createObjList();
  float distance(VEC vec);
  vector<int> findClosestAtoms(VEC vec, int N);
  // returns the distance from vec to the closest protein atom.

};


///
struct queStruct {

  struct queStruct *nextQEntry;
  float timeStamp;
  nodeClass *fromNode;
  nodeClass *toNode;
};
///
typedef struct queStruct queType;
///
class nodeIndexNodeDistance
{
  // each element of the pq.used in finding closest nodes
public:
  friend class prioritizeAtoms;
  nodeIndexNodeDistance (float _distance, int _index) : distance(_distance), index(_index) {};
  float getDistance()  const { return distance;}
  int getIndex() const { return index;}
  void setIndex(int _index) { index = _index;}
  void setDistance(float _dist) {distance = _dist;}
private:
    float distance;
    int index;

};

#endif
