function [ets,flags] = computeEscapeTimes(T, regions, solver, prof)
%
% Compute the escape for a set of regions (potential binding sites)
%
%    T:       transition probability matrix for the (entire) roadmap
%    regions: a cell array containing the indices of nodes in each region
%    solver:  1 for direct, 2 for naive, 3 for CG
%    prof:    flag for profiling
%    flags:   check computeOneEscTime..   

if (nargin < 2) 
  error('esctimes: too few arguments');
end;
if (nargin < 3)
  solver = 3;
end;  
if (nargin < 4)
  prof = 0;
end


maxit      = 1000000;
numRegions = size(regions, 1);
ets        = cell(numRegions, 1);
flags      = zeros(numRegions, 1);
for i=1:numRegions
  [ets{i},flags(i)] = computeOneEscTime(T, regions{i}, solver, maxit, prof);
end









