%first filters indices (according to dist. to protein surface, 
%pairwise distance, and distance to whether nodes are in the biggest connected 
%component.
%then computes funnels, and runs escape time computation scripts.
%writes the results to escapesTimesNIndices.txt.

clear;
load indices.txt
load pairwiseDistances.txt

distFromProt = textread('surfaceDistances.txt', '%*s %*s %*s %*s %*s  %f');
distFromProt
%pause
load transitionProbs.txt

% checks for the size of the read matrices..
% distFromProt should have 20 elements, pairwiseDistances should be 20x20.
[m,n] = size(pairwiseDistances);
if ((m ~= n) | (m ~= 20))
  fprintf(1, 'error! m and n have problems..\n');
  keyboard;
end

m = length(distFromProt); n = length(indices);
if ((m ~= 20) | (n ~= 20))
  fprintf(1, 'error! m or n has problem.\n');
  keyboard;
end

%end size checks


%now discarding centers which are close to   previous ones
removedFunnelCenters = []; % list of removed funnel centers.
keptFunnelCenters    = [];

for i = 1:20
  for j = i+1:20
    p = pairwiseDistances(i,j);
    if (p < 5)
      fprintf(1, 'centers %d and %d are close. removing %d\n', i,j,j);
      removedFunnelCenters = [removedFunnelCenters j];
    end
  end
end

%now discarding centers which are far from protein surface
toofarfunnelCenters = find(distFromProt > 4);
fprintf(1, 'discarding the following nodes since they are far from\n');
fprintf(1, 'protein surface...\n');
toofarfunnelCenters

removedFunnelCenters = [removedFunnelCenters toofarfunnelCenters'];

biggestCCNodes = findBiggestCC(transitionProbs);

fprintf(1, 'num nodes in biggest cc is %d\n', length(biggestCCNodes));
for i = 1:length(indices)
    debugIndex = find(biggestCCNodes == indices(i));
    if (length(debugIndex) == 0)
        fprintf(1, 'alert. index %d is not in the biggest cc.\n', indices(i));
        fprintf(1, 'removing this index from funnel center indices list\n');
%        return
	removedFunnelCenters = [removedFunnelCenters i];
    end
end

for i = 1:20
  if (length(find(removedFunnelCenters == i)) == 0) 
	%couldn't find i 
        %in the list of funnel centers to be removed...
     keptFunnelCenters = [keptFunnelCenters i];
  end
end

indices = indices(keptFunnelCenters);
fprintf(1, 'we have %d funnel centers..\n', length(indices));
% finished discarding funnel centers close to previous ones or 
% too far from protein surface.






temp     = max(transitionProbs);
numNodes = temp(1) + 1;
fprintf(1,' num nodes = %d\n',numNodes); 

debugIndices1 = find(indices <= 0);
debugIndices2 = find(indices > numNodes);
if ((length(debugIndices1) > 0) | (length(debugIndices2) > 0))
     keyboard;
end

[numConnections,n]          = size(transitionProbs);
fprintf(1, 'num connections = %d\n', numConnections);

Tsparse  = prepareTsparse (transitionProbs);

[funnels, funnel_equP] = computeFunnels(indices, 1E5, Tsparse);
% do escape time computation with all funnels

expectedTime = computeEscapeTime(Tsparse, funnels, indices);

fprintf(1, 'expected time to escape from each funnel center is as follows:\n');

escapeTimes = zeros(1, length(indices));
for i = 1:length(indices)
    funnelCenterIndex = find(funnels{i} == indices(i));
    escapeTimes(i)    = expectedTime{i}(funnelCenterIndex);
end
indices
escapeTimes

fid = fopen ('escapeTimesNIndices.txt', 'w');
fprintf     (fid, 'indices : \n');
fprintf     (fid, '%d\n', indices');
fprintf     (fid, 'escape times: \n');
fprintf     (fid, '%e\n', escapeTimes');


% do it pairwise between funnel of catalytic site and other site.

%first making sure the catalytic site is among the indices;
%if (length(find(indices == 1)) ~= 1)
%    fprintf(1, 'alert! indices does not contain 1...\n');
%    keyboard;
%end

%catalyticIndex = find(indices == 1);


%for i = 1:length(indices)

%  if (i ~= catalyticIndex)

%    pairIndices            = [1 indices(i)];
%    [funnels, funnel_equP] = computeFunnels(pairIndices, 1E5, Tsparse);

%    do escape time computation with all funnels

%    expectedTime           = computeEscapeTime(Tsparse, funnels, pairIndices);
%    fprintf(1, 'expected time to escape from each funnel center is as follows:\n');
%    escapeTimes            = zeros(1, length(pairIndices));

%    for i = 1:length(pairIndices)
%      funnelCenterIndex = find(funnels{i} == pairIndices(i));
%      escapeTimes(i)    = expectedTime{i}(funnelCenterIndex);
%    end %for
%  indices
%  escapeTimes

%  fprintf     (fid, 'indices : \n');
%  fprintf     (fid, '%d\n', pairIndices');
%  fprintf     (fid, 'escape times: \n');
%  fprintf     (fid, '%e\n', escapeTimes');

%  end %if
%end %for

fclose(fid);
