function Tsparse = prepareTsparse (transitionProbs)
%after reading transition probs, this function returns Tsparse.

T                            = transitionProbs;
numNodes1                    = max(T(:,1)) + 1;
numNodes2                    = max(T(:,2)) + 1;

numNodes                     = max(numNodes1, numNodes2);
selfConnectionProbabilities  = zeros(numNodes,3);

%T(:,3) = max(T(:,3), 1E-7);
diff   = T(:,3) - transitionProbs(:,3);


numChangedTransProbs = length(find(diff ~= 0));
numTotalTransProbs   = length(diff);

fprintf(1, 'total %d elements in the transition matrix are elevated to a probability of 1E-7, out of %d connections\n', numChangedTransProbs , numTotalTransProbs);

if ((numChangedTransProbs/numTotalTransProbs) > 0.1)
  fprintf(1, 'warning! there are too many edges that have been changed.\n');
  fprintf(1, 'type return to continue\n');
  keyboard;
end

rowIndex = 1; [numRows,dummy] = size(T);
for i = 1:numRows-1
  if (T(i,1) > T(i+1,1))
    keyboard;
  end
end



for i = 0:numNodes-1

% below lines are
% equivalent to:   links                 = find(T(:,1) == i);
% but much faster since they take advantage of the fact
% that the rows of the array are in increasing order.

  startRowIndex = rowIndex;

  while ((rowIndex<= numRows) &  (T(rowIndex, 1) == i))
    rowIndex = rowIndex + 1;
  end

  links                      = [startRowIndex:rowIndex-1]';
%end finding links.

%all the outgoing edges from node # i.
  sigmaProbability      = sum(T(links,3));
  numLinks              = length(links);

  
  if (numLinks > 0)
    if (numLinks - sigmaProbability < 0)
      fprintf('numLinks - sigmaProbability < 0');
      keyboard;
    end
    selfConnectionProbabilities(i+1,:)   = [i i (numLinks - sigmaProbability)/numLinks];
    T(links,3)                           = T(links,3) / numLinks; 
    checkForDuplicateEntries(i, T(links,2));
  else    
      selfConnectionProbabilities(i+1,:) = [i i 1];
  end
end


T = [T; selfConnectionProbabilities];

Tsparse = sparse  (T(:,1) + 1, T(:,2) + 1,  T(:,3));

function checkForDuplicateEntries(node_i, node_j_array)
sortedIndices                        = sort(node_j_array);
for i = 1:(length(sortedIndices) - 1)
  if (sortedIndices(i) == sortedIndices(i+1))
    fprintf(1, 'duplicate %d,%d found\n', node_i, sortedIndices(i));
    error('there are duplicate entries in the transition prob. file');
  end
end
