#include <iostream>
using std::cout;
using std::endl;
#include <iomanip>
#include <fstream>
using std::ifstream;
using std::ios;
#include <sstream>
#include <cstring>
#include <cmath>
#include <cstdlib>
#include <cstdio>
#include <float.h>
#include <GL/gl.h>
#include <GL/glu.h>
//#include <GL/glx.h>
#include "ldock.H"
//#include <stl_algobase.h>
#include <queue>
using std::priority_queue;
#include "transform.H"

extern int globalNumAtoms;
extern int globalRootAtom;
extern ligAtom *myLigand;
extern ligAtom **globalLigAtomPtrs;
extern int** globalNonBondAtoms;
extern int* globalNumNonBondAtoms;
extern float *globalLigAtomCharge;
extern float globalNumNLigCharge;
extern float globalNumPLigCharge;
extern float globalPrevEnergy;


extern int globalPathLength;
extern atomData* pathConfigurations[MAXNODES];



extern void printAtomData(atomData*);
extern void extStr(char*, char*, int, int);
extern void addAnimatedLigand(atomData*);

extern char aa3_aa(char*);
extern float compute_side_charge(char*, char, float);
extern float compute_C_charge(char, float);
extern float compute_N_charge(char, float);

float MAX_XYZ;
float MIN_X = 10000.0, MIN_Y = 10000.0, MIN_Z = 10000.0;
float MAX_X = -10000.0, MAX_Y = -10000.0, MAX_Z = -10000.0;
float MIN_P_ENERGY = 100000.0, MIN_N_ENERGY = 100000.0;
float MIN_VAN_ENERGY = 1000000.0;
float MIN_ENERGY;

receptorClass myReceptor;


void bitClass::initBitClass(int mysize) {

  int i;

  size = mysize;
  numWords = (mysize + WORDSIZE - 1)/WORDSIZE;

  bv = new bitVecType[numWords];

  for (i = 0; i < numWords; i++) {
    bv[i] = 0;
  }
}
  

void bitClass::setBit(int pos) {

  int index, offset;

  index = pos/WORDSIZE;
  offset = pos%WORDSIZE;

  bv[index] |= (1 << offset);
}


int bitClass::getBit(int pos) {

  int index, offset;

  index = pos/WORDSIZE;
  offset = pos%WORDSIZE;

  return (bv[index] & (1 << offset));

}



void bitmapClass::initBitmap(float angstromSize, float mfact) {

  multFactor = mfact; 
  dimensionSize = (int)(angstromSize*mfact);

  map.initBitClass(dimensionSize*dimensionSize*dimensionSize);

}




void bitmapClass::setBits(VEC mypt) {

  int xval, yval, zval, x, y, z, zindex, yindex, subfactor;
  VEC tvec;
  float divFactor, divFactorBy2;

  divFactor = 1.0/multFactor;
  divFactorBy2 = divFactor/2.0;

  xval = (int)(mypt[0]*multFactor);
  yval = (int)(mypt[1]*multFactor);
  zval = (int)(mypt[2]*multFactor);

  subfactor = (int)((3.0 + divFactor) * multFactor);

  for (z = zval - subfactor; z <= zval + subfactor; z++) {

    if ((z >= 0) && (z < dimensionSize)) {
      zindex = z * dimensionSize * dimensionSize;
      tvec[2] = ((float)(z)) * divFactor + divFactorBy2;

      for (y = yval - subfactor; y <= yval + subfactor; y++) {

	if ((y >= 0) && (y < dimensionSize)) {
	  yindex = y * dimensionSize;
	  tvec[1] = ((float)(y)) * divFactor + divFactorBy2;

	  for (x = xval - subfactor; x <= xval + subfactor; x++) {

	    if ((x >= 0) && (x < dimensionSize)) {
	      tvec[0] = ((float)(x)) * divFactor + divFactorBy2;

	      if (distanceBetweenVecs(tvec, mypt) <= 2.8) {
		map.setBit(zindex + yindex + x);
	      }
	    }
	  }
	}
      }
    }
  }
}



void receptorClass::setVanPot(VEC mypt, float mysize) {

  int xval, yval, zval, x, y, z, zindex, yindex, subfactor;
  VEC tvec;
  float divFactor, divFactorBy2;
  float mydist;


  divFactor = 1.0/potMatMultFact;
  divFactorBy2 = divFactor/2.0;

  xval = (int)(mypt[0]*potMatMultFact + divFactor);
  yval = (int)(mypt[1]*potMatMultFact + divFactor);
  zval = (int)(mypt[2]*potMatMultFact + divFactor);

  subfactor = (int)(5.0 * potMatMultFact);

  for (z = zval - subfactor; z <= zval + subfactor; z++) {

    if ((z >= 0) && (z < potMatSize)) {
      zindex = z * potMatSize * potMatSize;
      tvec[2] = ((float)(z)) * divFactor;

      for (y = yval - subfactor; y <= yval + subfactor; y++) {

	if ((y >= 0) && (y < potMatSize)) {
	  yindex = y * potMatSize;
	  tvec[1] = ((float)(y)) * divFactor;

	  for (x = xval - subfactor; x <= xval + subfactor; x++) {

	    if ((x >= 0) && (x < potMatSize)) {
	      tvec[0] = ((float)(x)) * divFactor;

	      mydist = distanceBetweenVecs(tvec, mypt);
	      
	      vanPot[zindex + yindex + x] += 0.2*(pow((mysize+1.45)/mydist,12) - 2.0*(pow((mysize+1.45)/mydist,6)));


	    }
	  }
	}
      }
    }
  }

}

void receptorClass::setElecPot(VEC mypt, float mycharge) {

  int xval, yval, zval, x, y, z, zindex, yindex, subfactor;
  VEC tvec;
  float divFactor, divFactorBy2;
  float mydist;

  divFactor = 1.0/potMatMultFact;
  divFactorBy2 = divFactor/2.0;

  xval = (int)(mypt[0]*potMatMultFact);
  yval = (int)(mypt[1]*potMatMultFact);
  zval = (int)(mypt[2]*potMatMultFact);

  subfactor = (int)(20.0 * potMatMultFact);

  for (z = zval - subfactor; z <= zval + subfactor; z++) {

    if ((z >= 0) && (z < potMatSize)) {
      zindex = z * potMatSize * potMatSize;
      tvec[2] = ((float)(z)) * divFactor + divFactorBy2;

      for (y = yval - subfactor; y <= yval + subfactor; y++) {

	if ((y >= 0) && (y < potMatSize)) {
	  yindex = y * potMatSize;
	  tvec[1] = ((float)(y)) * divFactor + divFactorBy2;

	  for (x = xval - subfactor; x <= xval + subfactor; x++) {

	    if ((x >= 0) && (x < potMatSize)) {
	      tvec[0] = ((float)(x)) * divFactor + divFactorBy2;

	      mydist = distanceBetweenVecs(tvec, mypt);
	      elecPot[zindex + yindex + x] += (5.53*mycharge)/(mydist);
	    }
	  }
	}
      }
    }
  }
}




int bitmapClass::checkBit(VEC mypt) {

  int xval, yval, zval;

  if ((mypt[0] > 0.0) && (mypt[0] < MAX_XYZ) && (mypt[1] > 0.0) && (mypt[1] < MAX_XYZ) && (mypt[2] > 0.0) && (mypt[2] < MAX_XYZ)) {

    xval = (int)(mypt[0]*multFactor);
    yval = (int)(mypt[1]*multFactor);
    zval = (int)(mypt[2]*multFactor);
    
    return map.getBit(zval*dimensionSize*dimensionSize + yval*dimensionSize + xval);
  }
  else {
    return 1;
  }

}

///
void receptorClass::readReceptorFile(const char *filename) {
  
  FILE* infile; 
  char record[10]; 
  char line[200]; 
  char temp[100]; 
  char atomname[5];
  char resname3[4];
  char resname;
  int i;

  infile = fopen(filename, "r"); 

  numReceptorAtoms = 0;

  while (fgets(line, 200, infile)) { 
    
    extStr(record, line, 1, 6); 
    
    if ((strcmp(record, "ATOM  ") == 0) || (strcmp(record, "HETATM") == 0)) { 
      
      extStr(temp, line, 31, 38); 
      sscanf(temp, "%f", &(recAtom[numReceptorAtoms][0])); 
      
      extStr(temp, line, 39, 46); 
      sscanf(temp, "%f", &(recAtom[numReceptorAtoms][1])); 
      
      extStr(temp, line, 47, 54); 
      sscanf(temp, "%f", &(recAtom[numReceptorAtoms][2])); 

      extStr(temp, line, 55, 60); 
      sscanf(temp, "%f", &(recAtomSize[numReceptorAtoms])); 

      extStr(temp, line, 61, 67); 
      sscanf(temp, "%f", &(recAtomCharge[numReceptorAtoms])); 


      //      extStr(temp, line, 13, 16); 
      //      sscanf(temp, "%s", atomname); 

      //      extStr(temp, line, 18, 20); 
      //      sscanf(temp, "%s", resname3); 
      
      //      resname = aa3_aa(resname3);

      //      recAtomCharge[numReceptorAtoms] = compute_side_charge(atomname, resname, MY_PH);

      //      if (numReceptorAtoms == 0) {
      //	recAtomCharge[numReceptorAtoms] += compute_N_charge(resname, MY_PH);
      //      }

      numReceptorAtoms++; 
      //      cout << "numReceptorAtoms = " << numReceptorAtoms << endl;
      if (numReceptorAtoms == MAX_ATOMS) {
	cout << "Reached MAX_ATOMS in receptor\n";
	break;
      }      
    }
  }

  //  recAtomCharge[numReceptorAtoms - 1] += compute_C_charge(resname, MY_PH);

  for (i = 0; i < numReceptorAtoms; i++) {

    if (recAtom[i][0] < MIN_X) 
      MIN_X = recAtom[i][0];

    if (recAtom[i][1] < MIN_Y) 
      MIN_Y = recAtom[i][1];

    if (recAtom[i][2] < MIN_Z) 
      MIN_Z = recAtom[i][2];

    if (recAtom[i][0] > MAX_X) 
      MAX_X = recAtom[i][0];

    if (recAtom[i][1] > MAX_Y) 
      MAX_Y = recAtom[i][1];

    if (recAtom[i][2] > MAX_Z) 
      MAX_Z = recAtom[i][2];
  }

  cout << "num receptor atoms = " << numReceptorAtoms << endl;

}
 

void receptorClass::fixReceptor() {

  int i;

  for (i = 0; i < numReceptorAtoms; i++) {
    recAtom[i][0] += MIN_X;
    recAtom[i][1] += MIN_Y;
    recAtom[i][2] += MIN_Z;
  }

}

void receptorClass::createBitmap(float multfact) {

  int i;

  myBitmap.initBitmap(MAX_XYZ + 1.0, multfact);

  for (i = 0; i < numReceptorAtoms; i++) {
    myBitmap.setBits(recAtom[i]);
  }

}



void receptorClass::createPotential(float gridresolution, char *filename) {

  int i, numcharged = 0, totalmatsize;
  int x, y, z, zindex, yindex;
  VEC tvec;
  float divFactor, divFactorBy2;
  float energy;
  ifstream gridfile;
  char ch;


  gridfile.open(filename, ios::in);


  ch = '0';
  while (ch != '\n') {
    gridfile.get(ch);
  }

  ch = '0';
  while (ch != '\n') {
    gridfile.get(ch);
  }

  ch = '0';
  while (ch != '\n') {
    gridfile.get(ch);
  }

  ch = '0';
  while (ch != '\n') {
    gridfile.get(ch);
  }

  ch = '0';
  while (ch != '\n') {
    gridfile.get(ch);
  }


  potMatMultFact = 1.0/gridresolution;
  potMatSize = (int) (MAX_XYZ * potMatMultFact) + 1;

  totalmatsize = potMatSize * potMatSize * potMatSize;

  elecPot = new float[totalmatsize];
  vanPot = new float[totalmatsize];

  for (i = 0; i < totalmatsize; i++) {
    elecPot[i] = 0.0;
    vanPot[i] = 0.0;
  }

  cout << "pot mats zeroed" << endl;
  

  for (i = 0; i < totalmatsize; i++) {
    if (!(gridfile >> elecPot[i])) {
      cout << "eof reached " << i << " " << totalmatsize << endl;
      exit(0);
    }
  }


  gridfile.close();

  /*  for (i = 0; i < numReceptorAtoms; i++) {
    if ((recAtomCharge[i] > 0.000001) || (recAtomCharge[i] < -0.000001)) {
      setElecPot(recAtom[i], recAtomCharge[i]);
      numcharged++;
    }
  }*/

  cout << "elec pot done" << endl;


  for (i = 0; i < numReceptorAtoms; i++) {
    setVanPot(recAtom[i], recAtomSize[i]);
  }

  cout << "van pot done" << endl;

  divFactor = gridresolution;

  for (z = 0; z < potMatSize; z++) {
      
    zindex = z * potMatSize * potMatSize;
    tvec[2] = ((float)(z)) * divFactor;
    
    for (y = 0; y < potMatSize; y++) {
      
      yindex = y * potMatSize;
      tvec[1] = ((float)(y)) * divFactor;
      
      for (x = 0; x < potMatSize; x++) {
	tvec[0] = ((float)(x)) * divFactor;
	
	energy = vanPot[zindex+yindex+x] - elecPot[zindex+yindex+x];
	if (energy < MIN_N_ENERGY) {
	  MIN_N_ENERGY = energy;
	}


	energy = vanPot[zindex+yindex+x] + elecPot[zindex+yindex+x];
	if (energy < MIN_P_ENERGY) {
	  MIN_P_ENERGY = energy;
	}


	energy = vanPot[zindex+yindex+x];
	if (energy < MIN_VAN_ENERGY) {
	  MIN_VAN_ENERGY = energy;
	}

      }
    }
  }

  //  cout << "MIN_P = " << MIN_P_ENERGY << " MIN_N = " << MIN_N_ENERGY << endl;

  cout << "min Vanpot = " << MIN_VAN_ENERGY << endl;

  
  MIN_ENERGY = ((float)(globalNumNLigCharge)) * MIN_N_ENERGY + ((float)(globalNumPLigCharge)) * MIN_P_ENERGY + ((float)(globalNumAtoms)) * (MIN_VAN_ENERGY/4.0);

  cout << "Min energy = " <<  MIN_ENERGY << endl;

  MIN_ENERGY = -60.0;

}


int receptorClass::checkCollision (atomData *adata, float *retEnergy) {
  
  float myenergy;

  myenergy = getEnergy(adata);

  *retEnergy = myenergy;

  //myrand = (((float)(random()))/((float)(MY_RAND_MAX)))*(((float)(MAX_ENERGY)) - MIN_ENERGY) + MIN_ENERGY;  

  if (myenergy < MAX_ENERGY)
    return 1;
  else
    return 0;

  /*
    if (myrand > myenergy) {
    return 1;
    }
    else {
    return 0;
    }
  */
}
    
  
float receptorClass::getEnergy(atomData *adata) {

  int i, xval, yval, zval, index;
  float retval = 0.0;
  float divFactor;

  divFactor = 1.0/potMatMultFact;

  for (i = 0; i < globalNumAtoms; i++) {
    
    if ((adata[i].xyz[0] > 0.0) && (adata[i].xyz[0] < MAX_XYZ) && (adata[i].xyz[1] > 0.0) && (adata[i].xyz[1] < MAX_XYZ) && (adata[i].xyz[2] > 0.0) && (adata[i].xyz[2] < MAX_XYZ)) {

      xval = (int)(adata[i].xyz[0]*potMatMultFact + divFactor);
      yval = (int)(adata[i].xyz[1]*potMatMultFact + divFactor);
      zval = (int)(adata[i].xyz[2]*potMatMultFact + divFactor);
      
      index = zval*potMatSize*potMatSize + yval*potMatSize + xval;
      
      retval += vanPot[index] + globalLigAtomCharge[i]*elecPot[index];
      
    }
    else {
      retval += 100000.0;
    }

  }


  /*  

  for (i = 0; i < globalNumAtoms; i++) {
    if (myBitmap.checkBit(adata[i].xyz) != 0) {
      return 0;
    }
  }*/


  int j, k;
  float mydist;


  if (retval < 10000.0) {
    for (i = 0; i < globalNumAtoms; i++) {
      for (j = 0; j < globalNumNonBondAtoms[i]; j++) {
	k = globalNonBondAtoms[i][j];
	mydist = distanceBetweenVecs(adata[i].xyz, adata[k].xyz);	  	
	retval += 0.2*(pow(2.8/mydist,12) - 2.0*(pow(2.8/mydist,6)));
	retval += (5.53*globalLigAtomCharge[i]*globalLigAtomCharge[k])/mydist;	  
      }
    }
  }
  
  return retval;

}


void findMiddleConfig(atomData *start, atomData *end, atomData *middle) {

  int i;

  for (i = 0; i < globalNumAtoms; i++) {
    
    if (globalRootAtom == i) {

      middle[i].xyz[0] = (end[i].xyz[0] + start[i].xyz[0])/2.0;
      middle[i].xyz[1] = (end[i].xyz[1] + start[i].xyz[1])/2.0;
      middle[i].xyz[2] = (end[i].xyz[2] + start[i].xyz[2])/2.0;

    }
    else {

      middle[i].angle[0] = (end[i].angle[0] + start[i].angle[0])/2.0;
      
      if ((end[i].angle[1] - start[i].angle[1] >= -PI) && (end[i].angle[1] - start[i].angle[1] <= PI)) {
	middle[i].angle[1] = (end[i].angle[1] + start[i].angle[1])/2.0;
      }

      else {
	middle[i].angle[1] = PI + (end[i].angle[1] + start[i].angle[1])/2.0;
	if (middle[i].angle[1] > PI)
	  middle[i].angle[1] = middle[i].angle[1] - 2.0 * PI;
      }

    }

  }

}


#if 0
//very old checkpath
int checkPath(atomData *start, atomData *end) {

  int i, left = 0, right = 0, split = 0;
  int retval;
  atomData *middle;

  for (i = 0; i < globalNumAtoms; i++) {
    if (distanceBetweenVecs(start[i].xyz, end[i].xyz) > EPSILON_DIST) {
      split = 1;
      break;
    }
  }

  if (split == 1) {
    middle = new atomData[globalNumAtoms];
    findMiddleConfig(start, end, middle);
    myLigand->findXYZ_root(middle);
    if (myReceptor.getEnergy(middle) < MAX_ENERGY) {
      left = checkPath(start, middle);
      if (left != 0) {
	right = checkPath(middle, end);
      }
      delete [] middle;
      if ((left != 0) && (right != 0)) {
	retval = left + right;
	return retval;
      }
      else {
	return 0;
      }
    }
    else {
      delete [] middle;
      return 0;
    }
  }
  else {
    return 1;
  }

}

#endif




/*
void displayPath(atomData *start, atomData *end) {

  int i, split = 0;
  atomData *middle;

  for (i = 0; i < globalNumAtoms; i++) {
    if (distanceBetweenVecs(start[i].xyz, end[i].xyz) > EPSILON_DIST) {
      split = 1;
      break;
    }
  }

   if (split == 1) {
    middle = new atomData[globalNumAtoms];
    findMiddleConfig(start, end, middle);
    myLigand->findXYZ_root(middle);
    displayPath(start, middle);
    displayPath(middle, end);
    delete [] middle;
  }
  else {
    addAnimatedLigand(end);
    globalPathLength++;
  }

}
*/

float receptorClass::distance(VEC vec) {

  float  minDistance = 10000.0;

  for (int i=0; i<numReceptorAtoms; i++) {
    float tempDist;

    if ((tempDist = distanceBetweenVecs(recAtom[i], vec)) < minDistance) {
      minDistance = tempDist;
    }
  }
  return minDistance;
}

// friend class with function object to compare elementTypes.
class prioritizeAtoms
{
public:
  int operator() (const  nodeIndexNodeDistance& x,
		  const nodeIndexNodeDistance & y)
    {
      return x.distance > y.distance;
    }
};


// returns indices of N closest atoms in the vector.
vector<int> receptorClass::findClosestAtoms(VEC vec, int N)
{

  vector<int> retval;

  retval.resize(0);
  priority_queue<nodeIndexNodeDistance, vector<nodeIndexNodeDistance>, prioritizeAtoms> distQueue;

  for (int i = 0; i < numReceptorAtoms; i++) 
    {
      float tempDist;

      tempDist    = distanceBetweenVecs(recAtom[i], vec);

      nodeIndexNodeDistance n(0.0, 0);
      n.setIndex(i); n.setDistance(tempDist);
      distQueue.push(n);
    }

  for (int i = 0; i < N; i++)
    {
      nodeIndexNodeDistance n          = distQueue.top();
      distQueue.pop();

      retval.push_back(n.getIndex());
      cout << "distance # " << i << "= " << n.getDistance() << endl;
    }

  return retval;
}
