#include <iostream>
using std::cout;
using std::endl;
#include <algorithm>
using std::min;
#include <fstream>
#include "ldock.H"
#include "roadmapConstruction.H"
#include "analysis.H"
#include "transform.H"
#include "roadmapConstruction.H"

extern int globalTreeName;
extern treeClass *globalFirstTree;
extern treeClass *globalLastTree;
extern nodeData  *allNodes[];
extern int numTrees;
extern int globalNumAtoms;
extern int numNodes;
extern int linkTimeArray[];
extern int MAXLINKS;
extern int numNeighbours;
extern neighbourType neighbours[];
extern ligAtom *myLigand;
extern void computeCOG(nodeData * node);
extern void sampleCOG(VEC retval, float samplingRadius, VEC origin);
extern void  VecSub(VEC Result, const VEC A, const VEC B);
extern int VecClose (const VEC A, const VEC B, float epsilon);
extern void printVec(VEC);
extern void addToQue(float, nodeClass*, nodeClass*);
extern receptorClass myReceptor;
extern atomData *origAtomData;
extern float findRMSD(atomData *a1, atomData *a2);
extern int INIT_NODES;
extern int totalExtraNodes;
extern int numSampledAround;
extern void writeMyRoadmap    (int numNodes, const char * filename, 
			       nodeData** allNodes, 
			       int numRandomNodes, 
			       int numSampledAround,
			       int totalExtraNodes,
			       //			char * relDir, 
			       const vector<int> & potBindingSiteIndices);
extern void initNodeColor(int);

//----------------------------------------------------------
//---------node creation------------------------------------



/* 
   just creates the nodes, does not set the energy or anything. 
   startIndex is the node index from which we start allocating numNodes.
   startIndex is generally 0, but is useful if the folded state is already 
   allocated at node #0, in which case numNodes parameter is actually 1 less 
   than the number of nodes we want to create.
*/

void allocateNodes (int numNodes, nodeData ** allNodes, int startIndex)
{
  int finalNodeIndex = startIndex + numNodes;

  for (int i = startIndex; i < finalNodeIndex; i++)
    {
      allNodes[i]                         = new nodeData;
      allNodes[i] ->myData                = NULL;
      allNodes[i] ->myTree                = NULL;

      allNodes[i]->numLinks               = 0;
      nodeClass * newNode                 = new nodeClass;
      newNode->myConfig                   = allNodes[i];
      allNodes[i]->myNode                 = newNode;
      allNodes[i]->index                  = i;
      
      assert (allNodes[i]->myNode->lastLink == NULL);
    }
}


/** 
    all nodes are UNIFORMLY sampled, but accepted only if they are within a 
    given rmsd of the center node.
    note: center node (node # center) should either be 0, or should already 
    have been loaded.
*/
void createNodesAroundCenter(int numToCreate, int center, float samplingRadius) {

  int numCreated = 0;
  nodeClass *newNode;
  int numNotAccepted = 0;
  float tempEnergy;
  
  assert ((numNodes == 0 ) || (center < numNodes));

  
  while ((numCreated < numToCreate) && (numNodes < MAXNODES)) 
    {
      if ((numNotAccepted > numCreated * 10000) && (numCreated != 0))
      {
	cout << "can't find nodes close to node # " << center << " by less than " << samplingRadius << endl;
	cout << "created " << numCreated << " nodes so far and rejected " << numNotAccepted << " of them." << endl;
	exit(0);
      }

      //    assert (allNodes[numNodes] == NULL);
      if (numNodes == 0) 
	{
	  allNodes[numNodes] = copyConfig(origAtomData);
	}
      else 
	{
	  allNodes[numNodes] = createNonRandomConfig(allNodes[center]->cog, samplingRadius);
	}
      
      if ((myReceptor.checkCollision(allNodes[numNodes]->myData, &tempEnergy))
	   || (numNodes == 0))
	{
	  //cout << "find rmsd: " << findRMSD(allNodes[numNodes]->myData, allNodes[center]->myData) << endl;
	  assert(findRMSD(allNodes[numNodes]->myData, 
			  allNodes[center]->myData) < samplingRadius + 10.0);//MSA: why +10 here?
	  numNeighbours                  = 0;
	  newNode                        = new nodeClass;
	  newNode->myConfig              = allNodes[numNodes];
	  allNodes[numNodes]->myNode     = newNode;
	  allNodes[numNodes]->nodeEnergy = tempEnergy;
	      
	  if (tempEnergy > MAX_ENERGY) {
	    cout << "Whats going on?" << endl;
	  }
	  numNodes++;
	  numCreated++;
	  if (numNodes % 1000 == 0)
	    cout << "num nodes sampled == " << numNodes << endl;
	}
      else {

	numNotAccepted++;

	delete [] allNodes[numNodes]->myData;
	delete allNodes[numNodes];
      }      
    }
}



/**
  samples configs uniformly, within a sphere around node # center.
*/
void sampleRoutines (float samplingRadius, int center)
{
  
  if (numNodes == 0)
    createNodesAroundCenter (INIT_NODES, center, samplingRadius);
  else // already created one node.
    createNodesAroundCenter (INIT_NODES-1, center, samplingRadius);

  cout << "created nodes, all within " << samplingRadius << " of the catalytic site." << endl;

  for (unsigned int i = 0; i < numNodes; i++)
    {
      assert ((findRMSD(allNodes[i]->myData, 
			allNodes[center]->myData) < samplingRadius + 10.0));
    }


  assert       (numNodes == INIT_NODES); 

  cout << "Sampling complete.  NumNodes = " << numNodes << endl;
      
  setIndices(numNodes); // necessary so as to have the index field of every node to be properly set.
}



/*
  nodes to be sampled such that all of them are within rmsdRadius of the
  catalytic site.
*/
void createNewRoadmapRoutines (char roadmapFilename[],
			       const vector<int> & potBindingSiteIndices,
			       float samplingRadius, int center) 
			     
{
  sampleRoutines(samplingRadius, center);
  linkNodes (0);
  computeTransitionProbabilities(allNodes, numNodes);
  setIndices(numNodes); // necessary so as to make sure any new node
// added in the linking step also have their index field properly set.
  cout << "Extra nodes = " << totalExtraNodes << " NumNodes = " << numNodes << endl;
	  
  writeMyRoadmap (numNodes, roadmapFilename, allNodes, 
		  INIT_NODES,   numSampledAround, 
		  totalExtraNodes,  //relDir, 
		  potBindingSiteIndices);
}

/** creates a new config around original within samplingradius */
nodeData* createNonRandomConfig(VEC origCOG, float samplingRadius) 
{
  const float epsilon = 1E-2;
  nodeData *retval;
  VEC desiredCOG, translationVec;

  translationVec[0] =   translationVec[1] =   translationVec[2] = 0.0;
  
  retval         = new nodeData;

  retval->myData = new atomData[globalNumAtoms];
  for (int i = 0; i < globalNumAtoms; i++)
    retval->myData[i].angle[0] = retval->myData[i].angle[1] = 0.0;

  myLigand->createRandomLigand_root(retval->myData);
  myLigand->findXYZ_root(retval->myData, translationVec);

  computeCOG (retval);
  //  printVec   (origCOG);
  //printVec   (retval->cog);
  sampleCOG  (desiredCOG, samplingRadius, origCOG);
  VecSub     (translationVec, desiredCOG, retval->cog);
  //  printVec   (translationVec);

  myLigand->findXYZ_root(retval->myData, translationVec);
  retval->myNode = NULL;
  retval->myTree = NULL;

  retval->numLinks = 0;


  computeCOG(retval);
  if (!VecClose(retval->cog, desiredCOG, epsilon))
    {
      printVec(retval->cog);
      printVec(desiredCOG);
    }
  assert (VecClose(retval->cog, desiredCOG, epsilon));

  return retval;
}


//----------------------------------------------------------
//---------node connection----------------------------------


// using the edgeweights matrix, sets up the connections.

void doConnections4MyRoadmap(const vector<vector< float> > &edgeWeights, 
			     int numNodes, nodeData ** allNodes)
{


  for (int i = 0; i < numNodes; i++)
    {
      for (int j = i+1; j < numNodes; j++)
	{
	  if (edgeWeights [i][j] != -1.0)
	    {

	      assert (edgeWeights [j][i] != -1.0);
	      //	      cout << "linking " << i << " to " << j << endl;
	      makeLinkBetweenNodes (numNodes, i, j, edgeWeights [i][j], 
				    edgeWeights [j][i]);
	    }
	}
      
      if (allNodes[i]->myTree == NULL)
	{
	  cout << " a tree created for node # " << i << endl;

	  assert (allNodes[i]->myNode != NULL);
	  //tempCreated++;
  
	  treeClass * tempTreePtr     = new treeClass;
	  tempTreePtr->root           = allNodes[i]->myNode;
	  tempTreePtr->numNodesInTree = 1;
	  tempTreePtr->prevTree       = globalLastTree;
	  tempTreePtr->myTreeName     = globalTreeName;
	  globalTreeName++;
	  
	  allNodes[i]->myTree     = tempTreePtr;
	  
	  if (globalLastTree == NULL) 
	    {
	      globalLastTree = tempTreePtr;
	      globalFirstTree = tempTreePtr;
	    }
	  else 
	    {
	      globalLastTree->nextTree = tempTreePtr;
	      globalLastTree = tempTreePtr;
	    }
	  numTrees++;
	}
    }
}


/**
  adds allNodes[count] to the global forest.
  the myTree parameter of allNodes[count] should be NULL.
*/

void addNode2Forest(int count)
{

  treeClass  *tempTreePtr;

  assert (allNodes[count]->myTree == NULL) ;
  assert (allNodes[count]->myNode != NULL);
  //tempCreated++;
  
  tempTreePtr                 = new treeClass;
  tempTreePtr->root           = allNodes[count]->myNode;
  tempTreePtr->numNodesInTree = 1;
  tempTreePtr->prevTree       = globalLastTree;
  tempTreePtr->myTreeName     = globalTreeName;
  globalTreeName++;
	    
  allNodes[count]->myTree     = tempTreePtr;
	
  if (globalLastTree == NULL) 
    {
      globalLastTree = tempTreePtr;
      globalFirstTree = tempTreePtr;
    }
  else 
    {
      globalLastTree->nextTree = tempTreePtr;
      globalLastTree = tempTreePtr;
    }
	
  numTrees++;
}

///
nodeData *copyConfig(const atomData *adata) {

  nodeData *tempNodeData;
  VEC tvec;
  int i;

  tempNodeData = new nodeData;

  tempNodeData->myData = new atomData[globalNumAtoms];

  for (i = 0; i < globalNumAtoms; i++) {
    tempNodeData->myData[i].xyz[0] = adata[i].xyz[0];
    tempNodeData->myData[i].xyz[1] = adata[i].xyz[1];
    tempNodeData->myData[i].xyz[2] = adata[i].xyz[2];
    tempNodeData->myData[i].angle[0] = adata[i].angle[0];
    tempNodeData->myData[i].angle[1] = adata[i].angle[1];
  }

  tempNodeData->myNode = NULL;
  tempNodeData->myTree = NULL;
  tempNodeData->numLinks = 0;
  tvec[0] = 0.0;
  tvec[1] = 0.0;
  tvec[2] = 0.0;

  for (i = 0; i < globalNumAtoms; i++) {
    
    tvec[0] = tvec[0] + tempNodeData->myData[i].xyz[0];
    tvec[1] = tvec[1] + tempNodeData->myData[i].xyz[1];
    tvec[2] = tvec[2] + tempNodeData->myData[i].xyz[2];

  }

  tempNodeData->cog[0] = tvec[0]/((float)(globalNumAtoms));
  tempNodeData->cog[1] = tvec[1]/((float)(globalNumAtoms));
  tempNodeData->cog[2] = tvec[2]/((float)(globalNumAtoms));

  return tempNodeData;

}



///
void linkNodes(int startindex) {

  int i, nodeDataIndex, count;
  float mydist;
  treeClass *tempTreePtr;
  int numpathschecked = 0;
  int tempint, tempCreated = 0, tempDestroyed = 0;
  int neighboursfound = 0;


  for (count = startindex; count < numNodes; count++) {
    linkTimeArray[count] = startindex;
  }

  for (count = startindex; count < numNodes; count++) {

    if (allNodes[count]->numLinks < MAXLINKS) {

      numNeighbours = 0;
      
      for (i = 0; i < linkTimeArray[count]; i++) {
	
	mydist = distanceBetweenNodes(allNodes[i]->myData, allNodes[count]->myData);
	if (mydist < DISTTHRESHOLD) {
	  neighbours[numNeighbours].index = i;
	  neighbours[numNeighbours].dist = mydist;
	  neighbours[numNeighbours].linkedBit = 0;
	  numNeighbours++;
	}
      }

      for (i = count+1; i < numNodes; i++) {
	
	mydist = distanceBetweenNodes(allNodes[i]->myData, allNodes[count]->myData);
	if (mydist < DISTTHRESHOLD) {
	  neighbours[numNeighbours].index = i;
	  neighbours[numNeighbours].dist = mydist;
	  neighbours[numNeighbours].linkedBit = 0;
	  numNeighbours++;
	}
      }
      
      qsort(neighbours, numNeighbours, sizeof(neighbourType), mycompare);      
      neighboursfound = neighboursfound + numNeighbours;

      for (i = 0; ((i < numNeighbours) && (allNodes[count]->numLinks < MAXLINKS) && (i < MAXNEIGHBOURS)); i++) {
	
	nodeDataIndex = neighbours[i].index;
	
	tempint = makeLinkBetweenNodes(count, nodeDataIndex, 0.0, 0.0);
	
	if (tempint == 1) {
	  tempCreated++;
	}
	else if (tempint == -1) {
	  tempDestroyed++;
	}
	
	neighbours[i].linkedBit = 1;
	
	//}
	numpathschecked++;
      }
      
   

      if (allNodes[count]->myTree == NULL) {
	
	tempCreated++;
	
	tempTreePtr = new treeClass;
	
	tempTreePtr->root = allNodes[count]->myNode;
	tempTreePtr->numNodesInTree = 1;
	tempTreePtr->prevTree = globalLastTree;
	tempTreePtr->myTreeName = globalTreeName;
	globalTreeName++;
	    
	allNodes[count]->myTree = tempTreePtr;
	
	if (globalLastTree == NULL) {
	  globalLastTree = tempTreePtr;
	  globalFirstTree = tempTreePtr;
	}
	else {
	  globalLastTree->nextTree = tempTreePtr;
	  globalLastTree = tempTreePtr;
	}
	
	numTrees++;
	
      }


    }

    if ((count+1)%1000 == 0) {
      cout << "Nodes Done: " << count+1 << " Num Trees: " << numTrees << endl;
      cout << "paths checked " << numpathschecked << " neihboursfound " << neighboursfound << endl;
      cout << "Trees created " << tempCreated << " trees removed " << tempDestroyed << endl;
      if (globalFirstTree != NULL) 
      	globalFirstTree->printTree();
      cout << endl;
      numpathschecked = 0;
      neighboursfound = 0;
      tempCreated = 0;
      tempDestroyed = 0;
    }	  
  }

  cout << "Nodes Done: " << count << " Num Trees: " << numTrees << endl;
  cout << "paths checked " << numpathschecked << " neighboursfound " << neighboursfound << endl;
  cout << "Trees created " << tempCreated << " trees removed " << tempDestroyed << endl;
  if (globalFirstTree != NULL) 
    globalFirstTree->printTree();
  if (allNodes[0]->myTree != NULL) 
    {
      cout << "numNodes in the tree of native structure : " << allNodes[0]->myTree->numNodesInTree << endl;
    }
}






/** 
   does join the trees. I don't know what it returns, but 
   I don't think it is used.
*/
int joinTheTreesForNodes (int n1, int n2, int numNodes)
{
  treeClass * tempTreePtr;
  int retval = 0;

  if ((allNodes[n2]->myTree == NULL) && (allNodes[n1]->myTree == NULL)) 
    {
      //      cout << " A new tree is created. " << endl;
      retval                      = 1;
      
      tempTreePtr                 = new treeClass;
      
      tempTreePtr->root           = allNodes[n2]->myNode;
      tempTreePtr->numNodesInTree = 2;
      tempTreePtr->prevTree       = globalLastTree;
      tempTreePtr->myTreeName     = globalTreeName;
      globalTreeName++;
      
      allNodes[n2]->myTree        = tempTreePtr;
      allNodes[n1]->myTree        = tempTreePtr;
      
      if (globalLastTree == NULL) 
	{
	  globalLastTree  = tempTreePtr;
	  globalFirstTree = tempTreePtr;
	}
      else 
	{
	  globalLastTree->nextTree = tempTreePtr;
	  globalLastTree           = tempTreePtr;
	}
      numTrees++;
    }
  else if ((allNodes[n2]->myTree == NULL) && (allNodes[n1]->myTree != NULL)) 
    {
      //      cout << " case 2 " << endl;
      allNodes[n2]->myTree = allNodes[n1]->myTree;
      allNodes[n1]->myTree->numNodesInTree++;
    }
  else if ((allNodes[n2]->myTree != NULL) && (allNodes[n1]->myTree == NULL))
    {
      //      cout << " case 3 " << endl;
      allNodes[n1]->myTree = allNodes[n2]->myTree;
      allNodes[n2]->myTree->numNodesInTree++;
    }
  else if ((allNodes[n2]->myTree != NULL) && 
	   (allNodes[n1]->myTree != NULL) && 
	   (allNodes[n2]->myTree != allNodes[n1]->myTree))
    {
      //      
      //      cout << "tree1 : " << allNodes[n1]->myTree->numNodesInTree << " nodes, tree2 : "  << endl;
      //      
      //      cout << allNodes[n2]->myTree->numNodesInTree << " nodes. " << endl;
      tempTreePtr = allNodes[n2]->myTree;
      tempTreePtr->removeTree (allNodes[n1]->myTree);
      //			       numNodes);
      delete tempTreePtr;
      retval = -1;
      //      
      //      cout << " joint tree : " << allNodes[n1]->myTree->numNodesInTree << " nodes." << endl;
    }
  
  return retval;
}



/** simply connects the node n1 to node n2. */
void makeOneWayLink (const int n1, const int n2, const float pW)
{
  if (allNodes[n1] -> myNode == NULL)
    {
      cout << "n1 = " << n1 << endl;
    }
  
  if (allNodes[n2] -> myNode == NULL)
    {
      cout << "n2 = " << n2 << endl;
    }
  assert (allNodes[n1] -> myNode != NULL);
  assert (allNodes[n2] -> myNode != NULL);
  
  //  cout << "adding a node link from " << n1 << " to "<< n2  << endl;
  allNodes[n1]->myNode->addNodeLink(allNodes[n2]->myNode, pW);
  allNodes[n1]->numLinks++;
}

///
int makeLinkBetweenNodes (int numNodes, const int n1, const int n2, 
			  const float forwardPathWeight, 
			  const float backwardPathWeight) 
{
  int retval;


  makeOneWayLink ( n1,  n2,   forwardPathWeight);
  makeOneWayLink ( n2,  n1,   backwardPathWeight);
  
  retval = joinTheTreesForNodes(n1, n2, numNodes);

  return retval;
}

///
int makeLinkBetweenNodes(int snode, int enode, float forwardlen, float backwardlen) {

  int retval = 0;
  treeClass *tempTreePtr;

  allNodes[snode]->myNode->addNodeLink(allNodes[enode]->myNode, forwardlen);
  allNodes[enode]->myNode->addNodeLink(allNodes[snode]->myNode, backwardlen);
  
  allNodes[snode]->numLinks++;
  allNodes[enode]->numLinks++;
  
  if ((allNodes[enode]->myTree == NULL) && (allNodes[snode]->myTree == NULL)) {
    
    retval = 1;
    
    tempTreePtr = new treeClass;
    
    tempTreePtr->root = allNodes[enode]->myNode;
    tempTreePtr->numNodesInTree = 2;
    tempTreePtr->prevTree = globalLastTree;
    tempTreePtr->myTreeName = globalTreeName;
    globalTreeName++;
    
    allNodes[enode]->myTree = tempTreePtr;
    allNodes[snode]->myTree = tempTreePtr;
    
    if (globalLastTree == NULL) {
      globalLastTree = tempTreePtr;
      globalFirstTree = tempTreePtr;
    }
    else {
      globalLastTree->nextTree = tempTreePtr;
      globalLastTree = tempTreePtr;
    }
    
    numTrees++;
    
  }
  
  else if ((allNodes[enode]->myTree == NULL) && (allNodes[snode]->myTree != NULL)) {
    allNodes[enode]->myTree = allNodes[snode]->myTree;
    allNodes[snode]->myTree->numNodesInTree++;
  }
  
  else if ((allNodes[enode]->myTree != NULL) && (allNodes[snode]->myTree == NULL)) {
    allNodes[snode]->myTree = allNodes[enode]->myTree;
    allNodes[enode]->myTree->numNodesInTree++;
  }
  
  else if ((allNodes[enode]->myTree != NULL) && (allNodes[snode]->myTree != NULL) && (allNodes[enode]->myTree != allNodes[snode]->myTree)) {
    tempTreePtr = allNodes[enode]->myTree;
    tempTreePtr->removeTree(allNodes[snode]->myTree);
    delete tempTreePtr;
    retval = -1;
  }

  return retval;
}

//----------------------------------------------------------
//---------other helpers------------------------------------


/*
  n1: # neighbors of node 1
  fTP: forward transition probability.
  E1: energy of node 1.
*/
void findTransitionProbability(float n1,float  n2, float E1, float E2, float * fTP, float * bTP) 
{
  float p = n1/n2 * exp(-(E2 - E1) * INV_BOLTZMAN_K);
  
  *fTP = min<float>(1, p);
  *bTP = min<float>(1, 1/p);

}

void  assignTransitionProbability(int i,int j, float fTP, nodeData ** allNodes)
{
  bool found;
  ptrList * linkPtr;

  linkPtr  = allNodes[i]->myNode->firstLink;
  found    = false;

  while ((linkPtr != NULL) && (!found))
    {
      int index = linkPtr->node->myConfig->index;

      if (index == j)
	{
	  found = true;
	  assert (linkPtr->pathWeight == 0.0);
	  linkPtr->pathWeight = fTP;
	}
      linkPtr = linkPtr ->nextPtr;
    }
  assert (found);

}

void computeTransitionProbabilities(nodeData ** allNodes, int numNodes)
{
  vector<int> neighbors;
  vector<int> numNeighbors;
  float fTP, bTP;
  numNeighbors.resize(numNodes);
  for (int i = 0; i < numNodes; i++)
    {
      numNeighbors[i] = findNumNeighbors(i, numNodes, allNodes);
    }
  
  for (int node1 = 0; node1 < numNodes; node1++)
    {
      findNeighbors (neighbors, node1, allNodes);
      assert (neighbors.size() == numNeighbors[node1]);
      for (int k = 0; k < numNeighbors[node1]; k++)
	{
	  int node2 = neighbors[k];
	  if (node2 > node1)
	    {
	      findTransitionProbability  (numNeighbors[node1], 
					  numNeighbors[node2], 
					  allNodes[node1]->nodeEnergy, 
					  allNodes[node2]->nodeEnergy, 
					  &fTP, &bTP);
	      assignTransitionProbability(node1,node2, fTP, allNodes);
	      assignTransitionProbability(node2,node1, bTP, allNodes);
	    }
	}
    }
}

/* 
   set index field of every node
*/   
void setIndices(int numNodes)
{
  for (int i = 0; i < numNodes; i++)
    {
      allNodes[i]->index = i;
    }
}


int mycompare(const void *vfirst, const void *vsecond) {

  neighbourType *first;
  neighbourType *second;
  
  first = (neighbourType*)(vfirst);
  second = (neighbourType*)(vsecond);

  if (first->dist < second->dist) {
    return -1;
  }
  else if (first->dist == second->dist) {
    return 0;
  }
  else {
    return 1;
  }
}

///returns the maximum distance between corresponding atoms.
float distanceBetweenNodes(atomData *a1, atomData *a2) {

  float maxDist = 0;
  float mydist;
  int i;


  for (i = 0; i < globalNumAtoms; i++) {
    mydist = distanceBetweenVecs(a1[i].xyz, a2[i].xyz);
    if (mydist > maxDist) {
      maxDist = mydist;
    }
  }

  return maxDist;

}



//---------------------------------------------------------
//-----------nodeClass methods-----------------------------


nodeClass::nodeClass() {

  firstLink = NULL;
  lastLink = NULL;
  myConfig = NULL;
  pathPtr = NULL;
  nodeColor = (char)(0);

}


void nodeClass::addNodeLink(nodeClass *thelink, float wt) {

  ptrList *tempPtrList;

  tempPtrList = new ptrList;

  tempPtrList->node       = thelink;
  tempPtrList->pathWeight = wt;
  tempPtrList->nextPtr    = NULL;

  if (lastLink == NULL) {
    
    lastLink = tempPtrList;
    firstLink = tempPtrList;

  }
  else {

    lastLink->nextPtr = tempPtrList;
    lastLink = tempPtrList;
  }
}

/*
  goes thru all nodes that it is connected to and prints the index of these nodes.
*/
void nodeClass::printTree(nodeClass *fromNode) 
{
  ptrList *linkPtr;

  nodeColor = (char)(1);
  linkPtr   = firstLink;
  cout << myConfig->index << " ";
  
  while (linkPtr != NULL) 
    {
      //      if ((linkPtr->node != fromNode) &&
      if (linkPtr->node->nodeColor == (char)(0))
      	{
	  //cout << linkPtr->node->myConfig->index << endl;
	  linkPtr->node->printTree(this);
	}
      else
	{
	  //	  cout << "node # " << linkPtr->node->myConfig->index << " also seem to be connected, but ";
	  //	  cout << "its color is not 0" << endl;
	}
      linkPtr = linkPtr->nextPtr;
    }
}


void nodeClass::replaceTreePtr(treeClass *replacementTree, nodeClass *fromNode) 
{

  ptrList *linkPtr;

  myConfig->myTree = replacementTree;
  nodeColor = (char)(1);
  linkPtr = firstLink;
  
  while (linkPtr != NULL) 
    {
      if ((linkPtr->node != fromNode) && (linkPtr->node->nodeColor == (char)(0))) 
	{
	  linkPtr->node->replaceTreePtr(replacementTree, this);
	}
      linkPtr = linkPtr->nextPtr;
    }
}



void nodeClass::setPath(nodeClass *prevNode, float cTime) {

  ptrList *linkPtr;

  pathPtr = prevNode;
  nodeTime = cTime;

  linkPtr = firstLink;

  while (linkPtr != NULL) {
    if (linkPtr->node->nodeColor == (char)(0)) {
      addToQue(cTime + linkPtr->pathWeight, this, linkPtr->node);
    }
    linkPtr = linkPtr->nextPtr;
  }

}

void nodeClass::setPath_reverse(nodeClass *prevNode, float cTime) {

  ptrList *linkPtr;

  pathPtr = prevNode;
  nodeTime = cTime;

  linkPtr = firstLink;

  while (linkPtr != NULL) {
    if (linkPtr->node->nodeColor == (char)(0)) {     
      addToQue(cTime + (linkPtr->node->reverseWeight(this)), this, linkPtr->node);
    }
    linkPtr = linkPtr->nextPtr;
  }

}


float nodeClass::reverseWeight(nodeClass *prevNode) {

  ptrList *linkPtr;
  float retval = -1.0;

  linkPtr = firstLink;

  while (linkPtr != NULL) {
    if (linkPtr->node == prevNode) {
      retval = linkPtr->pathWeight;
      break;
    }
    linkPtr = linkPtr->nextPtr;
  }

  if (retval < 0) {
    cout << "path not found" << endl;
    retval = 0.1;
  }

  return retval;
  
}

//------------------------------------------------------------------
//----ptrList methods-----------------------------------------------

ptrList::ptrList() {

  nextPtr = NULL;
  node = NULL;
  pathWeight = 0.0;

}

//------------------------------------------------------------------
//-----treeClass methods--------------------------------------------

treeClass::treeClass() {

  root = NULL;
  nextTree = NULL;
  prevTree = NULL;
  numNodesInTree = 0;

}


void treeClass::printTree() {

  cout << "Tree name: " << myTreeName << " Tree Size: " << numNodesInTree << endl;

  if (nextTree != NULL) {
    nextTree->printTree();
  }
}

void treeClass::removeTree(treeClass* replacementTree) {

  replacementTree->numNodesInTree += numNodesInTree;

  initNodeColor(0);

  root->replaceTreePtr(replacementTree, NULL);

  if (prevTree != NULL) {
    prevTree->nextTree = nextTree;
  }
  else {
    globalFirstTree = nextTree;
  }

  if (nextTree != NULL) {
    nextTree->prevTree = prevTree;
  }
  else {
    globalLastTree = prevTree;
  }

  numTrees--;

}
    

//------------------------------------------------------------------
//------------------------------------------------------------------

