#include "ldock.H"
#include <iostream.h>
#include <vector>
#include <assert.h>
#include <math.h>
#include <algorithm>
#include "analysis.H"

// checkpath without saving middle configs
// same as before, except that the transition probabilities are computed according to
// metropolis criterion, and path weights follow from these probabilities accordingly.

extern int globaldebugCheckPath;
extern float globalForwardProb;
extern float globalBackwardProb;

/*
  finds and returns both node indices that the input node is connected
  to, as well as the edge weights.
*/

/*
  n1: # neighbors of node 1
  fTP: forward transition probability.
  E1: energy of node 1.
*/
void findTransitionProbability(float n1,float  n2, float E1, float E2, float * fTP, float * bTP) 
{
  float p = n1/n2 * exp(-(E2 - E1) * INV_BOLTZMAN_K);
  
  *fTP = min<float>(1, p);
  *bTP = min<float>(1, 1/p);

}

void  assignTransitionProbability(int i,int j, float fTP, nodeData ** allNodes)
{
  bool found;
  ptrList * linkPtr;

  linkPtr  = allNodes[i]->myNode->firstLink;
  found    = false;

  while ((linkPtr != NULL) && (!found))
    {
      int index = linkPtr->node->myConfig->index;

      if (index == j)
	{
	  found = true;
	  assert (linkPtr->pathWeight == 0.0);
	  linkPtr->pathWeight = fTP;
	}
      linkPtr = linkPtr ->nextPtr;
    }
  assert (found);

}

void computeTransitionProbabilities(nodeData ** allNodes, int numNodes)
{
  vector<int> neighbors;
  vector<int> numNeighbors;
  float fTP, bTP;
  numNeighbors.resize(numNodes);
  for (int i = 0; i < numNodes; i++)
    {
      numNeighbors[i] = findNumNeighbors(i, numNodes, allNodes);
    }
  
  for (int node1 = 0; node1 < numNodes; node1++)
    {
      findNeighbors (neighbors, node1, allNodes);
      assert (neighbors.size() == numNeighbors[node1]);
      for (int k = 0; k < numNeighbors[node1]; k++)
	{
	  int node2 = neighbors[k];
	  if (node2 > node1)
	    {
	      findTransitionProbability  (numNeighbors[node1], 
					  numNeighbors[node2], 
					  allNodes[node1]->nodeEnergy, 
					  allNodes[node2]->nodeEnergy, 
					  &fTP, &bTP);
	      assignTransitionProbability(node1,node2, fTP, allNodes);
	      assignTransitionProbability(node2,node1, bTP, allNodes);
	    }
	}
    }
}
