
#include <iostream>
#include <fstream>
#include <sstream>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "transform.H"
using namespace std;

#define PI 3.1415926535897

typedef float MAT[4][4];


void printMat(MAT);
void printVec(VEC);
void setMatToIdentity(MAT);
void applyMatToVec(VEC, VEC, MAT);
void matMult(MAT, MAT, MAT);
void copyMat(MAT, MAT);
void findTransform(MAT, VEC, VEC, VEC);
void findReverseTransform(MAT, VEC, VEC, VEC);
void findAngleParameters(VEC, float*, float*, float*);
void findCoordsFromAngles(VEC, float, float, float);
void normalize(VEC);

float distanceBetweenVecs(const VEC v1, const VEC v2) {

  VEC myvec;

  myvec[0] = v1[0] - v2[0];
  myvec[1] = v1[1] - v2[1];
  myvec[2] = v1[2] - v2[2];

  return sqrt(myvec[0]*myvec[0] + myvec[1]*myvec[1] + myvec[2]*myvec[2]);

}
  
void printMat(MAT themat) {

  int i, j;

  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {
      cout << themat[i][j] << "\t";
    }
    cout << endl;
  }
  cout << endl;
}

void printVec(VEC thevec) {

  int i;

  for (i = 0; i < 3; i++) {
    cout << thevec[i] << "\t\t";
  }
  cout << endl;
  cout << endl;
}
  
void normalize(VEC myvec) {

  float len;

  len = sqrt(myvec[0]*myvec[0] + myvec[1]*myvec[1] + myvec[2]*myvec[2]);

  myvec[0] = myvec[0]/len;
  myvec[1] = myvec[1]/len;
  myvec[2] = myvec[2]/len;
}

void setMatToIdentity(MAT themat) {

  int i, j;

  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {
      if (i == j) {
	themat[i][j] = 1.0;
      }
      else {
	themat[i][j] = 0.0;
      }
    }
  }
}
	
void applyMatToVec(VEC outvec, VEC invec, MAT themat) {

  outvec[0] = invec[0]*themat[0][0] + invec[1]*themat[1][0] + invec[2]*themat[2][0] + themat[3][0];
  outvec[1] = invec[0]*themat[0][1] + invec[1]*themat[1][1] + invec[2]*themat[2][1] + themat[3][1];
  outvec[2] = invec[0]*themat[0][2] + invec[1]*themat[1][2] + invec[2]*themat[2][2] + themat[3][2];

}


void matMult(MAT myoutmat, MAT mymat1, MAT mymat2) {

  int i, j;

  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {    
      myoutmat[i][j] = mymat1[i][0]*mymat2[0][j] + mymat1[i][1]*mymat2[1][j] + mymat1[i][2]*mymat2[2][j] + mymat1[i][3]*mymat2[3][j];
    }
  }
}

void copyMat(MAT myoutmat, MAT myinmat) {

  int i, j;

  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {    
      myoutmat[i][j] = myinmat[i][j];
    }
  }
}

  

void findTransform(MAT outmat, VEC back2, VEC back1, VEC current) {

  VEC zedvec, upvec, tempvec, tempvec1;
  MAT rotx, roty, rotz, trans, tempmat1, tempmat2;
  float temp;
  float sinu, cosu, sinuu, cosuu, sinuuu, cosuuu;


  setMatToIdentity(rotx);
  setMatToIdentity(roty);
  setMatToIdentity(rotz);
  setMatToIdentity(trans);

  trans[3][0] = -current[0];
  trans[3][1] = -current[1];
  trans[3][2] = -current[2];
  

  zedvec[0] = current[0] - back1[0];
  zedvec[1] = current[1] - back1[1];
  zedvec[2] = current[2] - back1[2];
  
  upvec[0] = back2[0] - back1[0];
  upvec[1] = back2[1] - back1[1];
  upvec[2] = back2[2] - back1[2];

  normalize(zedvec);
  normalize(upvec);

  //  temp = 1.0/(sqrt(zedvec[1]*zedvec[1] + zedvec[2]*zedvec[2]));

  temp = sqrt(zedvec[1]*zedvec[1] + zedvec[2]*zedvec[2]);

  sinu = zedvec[1]/temp;
  cosu = zedvec[2]/temp;


  //  temp = 1.0/(sqrt(zedvec[0]*zedvec[0] + zedvec[1]*zedvec[1] + zedvec[2]*zedvec[2]));

  sinuu = zedvec[0];
  cosuu = temp;

  if ((cosu >= -1.05) && (cosu <= 1.05) && (sinu >= -1.05) && (sinu <= 1.05)) {
    rotx[1][1] = cosu;
    rotx[1][2] = sinu;
    rotx[2][1] = -sinu;
    rotx[2][2] = cosu;
  }
  else {
    //    cout << "u out of range " << cosu << " " << sinu << endl;
  }

  if ((cosuu >= -1.05) && (cosuu <= 1.05) && (sinuu >= -1.05) && (sinuu <= 1.05)) {
    roty[0][0] = cosuu;
    roty[0][2] = sinuu;
    roty[2][0] = -sinuu;
    roty[2][2] = cosuu;
  }
  else {
    //    cout << "uu out of range " << cosuu << " " << sinuu << endl;
  }

  applyMatToVec(tempvec1, upvec, rotx);
  applyMatToVec(tempvec, tempvec1, roty);

  temp = sqrt(tempvec[0]*tempvec[0] + tempvec[1]*tempvec[1]);
  
  sinuuu = tempvec[0]/temp;
  cosuuu = tempvec[1]/temp;

  
  if ((cosuuu >= -1.05) && (cosuuu <= 1.05) && (sinuuu >= -1.05) && (sinuuu <= 1.05)) {
    rotz[0][0] = cosuuu;
    rotz[0][1] = sinuuu;
    rotz[1][0] = -sinuuu;
    rotz[1][1] = cosuuu;
  }
  else {
    //    cout << "uuu out of range " << cosuuu << " " << sinuuu << endl;
  }

  matMult(tempmat1, trans, rotx);
  matMult(tempmat2, tempmat1, roty);
  matMult(outmat, tempmat2, rotz);

}

float p1[3] = {0.0,1.0,-1.0};
float p2[3] = {0.0,0.0,-1.0};
float p3[3] = {0.0,0.0,0.0};


void findReverseTransform(MAT outmat, VEC back2, VEC back1, VEC current) {

  VEC tp1, tp2, tp3;

  findTransform(outmat, back2, back1, current);
  
  applyMatToVec(tp1, p1, outmat);
  applyMatToVec(tp2, p2, outmat);
  applyMatToVec(tp3, p3, outmat);

  findTransform(outmat, tp1, tp2, tp3);

}


void findCoordsFromAngles(VEC outvec, float theta, float phi, float veclen) {

  float temp;

  temp = veclen * ((float)(sin(theta)));
  outvec[2] = veclen * ((float)(cos(theta)));

  outvec[0] = -temp * ((float)(sin(phi)));
  outvec[1] = temp * ((float)(cos(phi)));

}


void findAngleParameters(VEC myvec, float *theta, float *phi, float *veclen) {

  float temp;
  float tempangle;

  *veclen = sqrt(myvec[0]*myvec[0] + myvec[1]*myvec[1] + myvec[2]*myvec[2]);

  assert (*veclen > 0);

  temp = myvec[2]/(*veclen);

  if (temp > 1.0) 
    temp = 1.0;
  if (temp < -1.0) 
    temp = -1.0;


  *theta = (float)(acos(temp));
  assert ((sqrt(myvec[0]*myvec[0] + myvec[1]*myvec[1])) > 0);

  temp = myvec[1]/(sqrt(myvec[0]*myvec[0] + myvec[1]*myvec[1]));
  if (temp > 1.0) 
    temp = 1.0;
  if (temp < -1.0) 
    temp = -1.0;

  tempangle = (float)(acos(temp));

  if (myvec[0] > 0.0) {
    *phi = -tempangle;
  }
  else {
    *phi = tempangle;
  }

}

//inline 
void  VecSub(VEC Result, const VEC A, const VEC B)
{
  for (int i=0; i<3; i++)
    Result[i] = A[i] - B[i];
}

//inline 
int VecEqual (const VEC A, const VEC B) {

  if (A[0] != B[0]) return 0;
  if (A[1] != B[1]) return 0;
  if (A[2] != B[2]) return 0;
  return 1;
}

int VecClose (const VEC A, const VEC B, float epsilon) {

  if (fabs(A[0] - B[0]) > epsilon) return 0;
  if (fabs(A[1] - B[1]) > epsilon) return 0;
  if (fabs(A[2] - B[2]) > epsilon) return 0;

  return 1;
}
#if 0
void findReverseTransform(MAT outmat, VEC back2, VEC back1, VEC current) {

  VEC zedvec, upvec, tempvec;
  MAT rotx, roty, rotz, trans, tempmat1, tempmat2;
  float temp;
  float sinu, cosu, sinuu, cosuu, sinuuu, cosuuu;

  setMatToIdentity(rotx);
  setMatToIdentity(roty);
  setMatToIdentity(rotz);
  setMatToIdentity(trans);

  trans[3][0] = current[0];
  trans[3][1] = current[1];
  trans[3][2] = current[2];
  

  zedvec[0] = current[0] - back1[0];
  zedvec[1] = current[1] - back1[1];
  zedvec[2] = current[2] - back1[2];
  
  upvec[0] = back2[0] - back1[0];
  upvec[1] = back2[1] - back1[1];
  upvec[2] = back2[2] - back1[2];

  normalize(zedvec);
  normalize(upvec);

  //  temp = 1.0/(sqrt(zedvec[1]*zedvec[1] + zedvec[2]*zedvec[2]));

  temp = sqrt(zedvec[1]*zedvec[1] + zedvec[2]*zedvec[2]);

  sinu = zedvec[1]/temp;
  cosu = zedvec[2]/temp;


  //  temp = 1.0/(sqrt(zedvec[0]*zedvec[0] + zedvec[1]*zedvec[1] + zedvec[2]*zedvec[2]));

  sinuu = zedvec[0];
  cosuu = temp;

  if ((cosu > -1.0) && (cosu < 1.0) && (sinu > -1.0) && (sinu < 1.0)) {
    rotx[1][1] = cosu;
    rotx[1][2] = -sinu;
    rotx[2][1] = sinu;
    rotx[2][2] = cosu;
  }
  else {
    cout << "out of range\n";
  }

  if ((cosuu > -1.0) && (cosuu < 1.0) && (sinuu > -1.0) && (sinuu < 1.0)) {
    roty[0][0] = cosuu;
    roty[0][2] = -sinuu;
    roty[2][0] = sinuu;
    roty[2][2] = cosuu;
  }
  else {
    cout << "out of range\n";
  }

  matMult(tempmat1, rotx, roty);

  applyMatToVec(tempvec, upvec, tempmat1);

  temp = sqrt(tempvec[0]*tempvec[0] + tempvec[1]*tempvec[1]);
  
  sinuuu = tempvec[0]/temp;
  cosuuu = tempvec[1]/temp;

  
  if ((cosuu > -1.0) && (cosuu < 1.0) && (sinuu > -1.0) && (sinuu < 1.0)) {
    rotz[0][0] = cosuuu;
    rotz[0][1] = -sinuuu;
    rotz[1][0] = sinuuu;
    rotz[1][1] = cosuuu;
  }
  else {
    cout << "out of range\n";
  }

  //  matMult(tempmat1, trans, rotx);
  //  matMult(tempmat2, tempmat1, roty);
  //  matMult(outmat, tempmat2, rotz);


  //  matMult(tempmat1, rotx, roty);
  //  matMult(tempmat2, tempmat1, rotz);
  //  matMult(outmat, tempmat2, trans);

  matMult(tempmat1, rotz, roty);
  matMult(tempmat2, tempmat1, rotx);
  matMult(outmat, tempmat2, trans);

}


#endif




#if 0


void findCoordsFromAngles(VEC outvec, float theta, float phi, float veclen) {

  float temp1, temp2, temp3;

  temp1 = sin(theta);
  temp2 = (sin(phi))/(cos(phi));
  temp3 = temp1*temp1*temp2*temp2;

  outvec[0] = sqrt(veclen*veclen*temp3/(1.0 + temp3));

  if (phi < 0) {
    outvec[0] = -outvec[0];
  }


  outvec[1] = outvec[0]/temp2;

  temp1 = sqrt(veclen*veclen - outvec[0]*outvec[0]);
  outvec[2] = cos(theta) * temp1;

}


void findAngleParameters(VEC myvec, float *theta, float *phi, float *veclen) {

  float tempangle;

  *veclen = sqrt(myvec[0]*myvec[0] + myvec[1]*myvec[1] + myvec[2]*myvec[2]);

  *theta = (float)(acos(myvec[2]/(sqrt(myvec[1]*myvec[1] + myvec[2]*myvec[2]))));

  tempangle = (float)(acos(myvec[1]/(sqrt(myvec[0]*myvec[0] + myvec[1]*myvec[1]))));

  if (myvec[0] >= 0.0) {
    *phi = tempangle;
  }
  else {
    *phi = -tempangle;
  }

}
  
void findCoordsFromAngles(VEC outvec, float theta, float phi, float veclen) {

  float temp;

  outvec[0] = ((float)(sin(phi)))*veclen;

  temp = ((float)(cos(phi)))*veclen;
  
  outvec[1] = ((float)(sin(theta)))*temp;

  outvec[2] = ((float)(cos(theta)))*temp;

}


void findAngleParameters(VEC myvec, float *theta, float *phi, float *veclen) {

  float tempangle, temp;

  *veclen = sqrt(myvec[0]*myvec[0] + myvec[1]*myvec[1] + myvec[2]*myvec[2]);

  temp = sqrt(myvec[1]*myvec[1] + myvec[2]*myvec[2]);

  tempangle = (float)(acos(myvec[2]/temp));

  if (myvec[1] >= 0.0) {
    *theta = tempangle;
  }
  else {
    *theta = -tempangle;
  }


  tempangle = (float)(acos(temp/(sqrt(myvec[0]*myvec[0] + temp*temp))));

  if (myvec[0] >= 0.0) {
    *phi = tempangle;
  }
  else {
    *phi = -tempangle;
  }

}
  
#endif



#if 0
main(int argc, char **argv) {

  VEC myorigin, myzed, myup;
  MAT testmat;


  myorigin[0] = 0.0;
  myorigin[1] = 0.0;
  myorigin[2] = 0.0;

  myzed[0] = 1.0;
  myzed[1] = 0.0;
  myzed[2] = 0.0;

  myup[0] = 0.0;
  myup[1] = 1.0;
  myup[2] = 0.0;

  findTransform(testmat, myorigin, myzed, myup);

  printMat(testmat);

}

main(int argc, char **argv) {

  VEC myvec;
  float theta, phi, length;

  myvec[0] = 1.0;
  myvec[1] = 1.0;
  myvec[2] = 1.0;

  findAngleParameters(myvec, &theta, &phi, &length);
  cout << theta << "\t" << phi << endl;

  myvec[0] = 1.0;
  myvec[1] = 1.0;
  myvec[2] = -1.0;

  findAngleParameters(myvec, &theta, &phi, &length);
  cout << theta << "\t" << phi << endl;

  myvec[0] = 1.0;
  myvec[1] = -1.0;
  myvec[2] = -1.0;

  findAngleParameters(myvec, &theta, &phi, &length);
  cout << theta << "\t" << phi << endl;

  myvec[0] = 1.0;
  myvec[1] = -1.0;
  myvec[2] = 1.0;

  findAngleParameters(myvec, &theta, &phi, &length);
  cout << theta << "\t" << phi << endl;

  myvec[0] = -1.0;
  myvec[1] = 1.0;
  myvec[2] = 1.0;

  findAngleParameters(myvec, &theta, &phi, &length);
  cout << theta << "\t" << phi << endl;

  myvec[0] = -1.0;
  myvec[1] = 1.0;
  myvec[2] = -1.0;

  findAngleParameters(myvec, &theta, &phi, &length);
  cout << theta << "\t" << phi << endl;

  myvec[0] = -1.0;
  myvec[1] = -1.0;
  myvec[2] = -1.0;

  findAngleParameters(myvec, &theta, &phi, &length);
  cout << theta << "\t" << phi << endl;

  myvec[0] = -1.0;
  myvec[1] = -1.0;
  myvec[2] = 1.0;

  findAngleParameters(myvec, &theta, &phi, &length);
  cout << theta << "\t" << phi << endl;

}
#endif

