#include <iostream>
using std::cout;
using std::endl;
#include <iomanip>
#include <fstream>
#include <sstream>
#include <string.h>
#include <cmath>
#include <cstdlib>
#include <cstdio>
#include "ldock.H"
#include "transform.H"

extern void printVec(VEC);
extern void applyMatToVec(VEC, VEC, MAT);
extern void findTransform(MAT, VEC, VEC, VEC);
extern void findReverseTransform(MAT, VEC, VEC, VEC);
extern void findAngleParameters(VEC, float*, float*, float*);
extern void findCoordsFromAngles(VEC, float, float, float);
extern void setMatToIdentity(MAT);
extern void matMult(MAT, MAT, MAT);
extern void VecSub (VEC Result, const VEC A, const VEC B);
extern int VecEqual (const VEC A, const VEC B);


extern float MAX_XYZ;
extern float MIN_X, MIN_Y, MIN_Z;
extern float MAX_X, MAX_Y, MAX_Z;

atomData *origAtomData;
ligAtom **globalLigAtomPtrs;
float *globalLigAtomCharge;
int **globalNonBondAtoms;
int *globalNumNonBondAtoms;
int *globalRigidAtoms;
int globalDOFArray[MAXDOF];
int globalNumDOF;
float globalNumNLigCharge;
float globalNumPLigCharge;
int globalConnections[10000][3];
int globalNumAtoms = 0, globalNumConnects = 0;
int globalRootAtom;

ligAtom *myLigand;


void printAtomData(atomData *);


void extStr(char* out, char* in, int st, int end) { 
 
  int outcount = 0; 
  int i;

  for (i = st - 1; i < end; i++) { 
 
    out[outcount] = in[i]; 
    outcount++; 
  } 
 
  out[outcount] = '\0'; 
 
} 




ligAtom::ligAtom() {
  
  numBonds = 0;
  colorBit = 0;
  atomNum = -2;
  bonds[0] = NULL;
  bonds[1] = NULL;
  bonds[2] = NULL;
  bonds[3] = NULL;
  bondLength = NULL;
  bondAngle = NULL;
  torAngle = NULL;
  
}


ligAtom::~ligAtom() {

  int i;

  for (i = 0; i < numBonds; i++) {
    delete bonds[i];
  }
  delete  bondLength;
  delete  bondAngle;
  delete  torAngle;
  delete &numBonds;
  delete &atomNum;
}



void ligAtom::findXYZ_root(atomData *adata, VEC translationVec) {

  MAT mymat;
  VEC newdata;
  int nextAtomNum;

  clearColors();

  colorBit = 1;

  setMatToIdentity(mymat);

  adata[atomNum].xyz[0] = mymat[3][0] = translationVec[0];//adata[atomNum].xyz[0];
  adata[atomNum].xyz[1] = mymat[3][1] = translationVec[1];//adata[atomNum].xyz[1];
  adata[atomNum].xyz[2] = mymat[3][2] = translationVec[2];//adata[atomNum].xyz[2];
  //  printVec(adata[atomNum].xyz);

  nextAtomNum = bonds[0]->atomNum;
  
  findCoordsFromAngles(newdata, adata[nextAtomNum].angle[0], 
		       adata[nextAtomNum].angle[1], bondLength[0]);
  
  applyMatToVec(adata[nextAtomNum].xyz, newdata, mymat);
  //  printVec(adata[nextAtomNum].xyz);
  if (bonds[0]->numBonds > 0) 
    {      
      bonds[0]->findXYZ(adata, atomNum, -1);
    }    
}

void ligAtom::findXYZ(atomData *adata, int back1, int back2) 
{

  MAT mymat;
  VEC newdata, tvec;
  int i, nextAtomNum;


  colorBit = 1;

  if (back2 != -1) 
    {
      findReverseTransform(mymat, adata[back2].xyz, adata[back1].xyz, 
			   adata[atomNum].xyz);
    }
  else 
    {

      tvec[0] = adata[back1].xyz[0];
      tvec[1] = adata[back1].xyz[1] + 1.0;
      tvec[2] = adata[back1].xyz[2];

      findReverseTransform(mymat, tvec, adata[back1].xyz, adata[atomNum].xyz);
    
    }


  for (i = 0; i < numBonds; i++) {

    nextAtomNum = bonds[i]->atomNum;

    if (bonds[i]->colorBit == 0) {
      findCoordsFromAngles(newdata, adata[nextAtomNum].angle[0], 
			   adata[nextAtomNum].angle[1], bondLength[i]);      
      
      applyMatToVec(adata[nextAtomNum].xyz, newdata, mymat);    
      //      printVec(adata[nextAtomNum].xyz);
    }

  }

  for (i = 0; i < numBonds; i++) {

    if ((bonds[i]->numBonds > 0) && (bonds[i]->colorBit == 0)) {      
      bonds[i]->findXYZ(adata, atomNum, back1);
    }
  }
}



void ligAtom::findAngles_root(atomData *adata) {

  VEC tvec;
  int nextAtomNum;

  clearColors();

  colorBit = 1;

  nextAtomNum = bonds[0]->atomNum;
  
  tvec[0] = adata[nextAtomNum].xyz[0] - adata[atomNum].xyz[0];
  tvec[1] = adata[nextAtomNum].xyz[1] - adata[atomNum].xyz[1];
  tvec[2] = adata[nextAtomNum].xyz[2] - adata[atomNum].xyz[2];
  
  findAngleParameters(tvec, &(adata[nextAtomNum].angle[0]), &(adata[nextAtomNum].angle[1]), &(bondLength[0]));

  bondAngle[0] = adata[nextAtomNum].angle[0];
  torAngle[0] = adata[nextAtomNum].angle[1];
  
  if (bonds[0]->numBonds > 0) {
    bonds[0]->findAngles(adata, atomNum, -1);
  }

}


void ligAtom::findAngles(atomData *adata, int back1, int back2) {

  VEC tvec;
  MAT mymat;
  int i, nextAtomNum;

  colorBit = 1;

  if (back2 != -1) 
    {
      findTransform(mymat, adata[back2].xyz, 
		    adata[back1].xyz, adata[atomNum].xyz);
    }

  else {

    tvec[0] = adata[back1].xyz[0];
    tvec[1] = adata[back1].xyz[1] + 1.0;
    tvec[2] = adata[back1].xyz[2];

    findTransform(mymat, tvec, adata[back1].xyz, adata[atomNum].xyz);

  }

  for (i = 0; i < numBonds; i++) {

    nextAtomNum = bonds[i]->atomNum;

    if (bonds[i]->colorBit == 0) {
      
      applyMatToVec(tvec, adata[nextAtomNum].xyz, mymat);
      
      findAngleParameters(tvec, &(adata[nextAtomNum].angle[0]), 
			  &(adata[nextAtomNum].angle[1]), &(bondLength[i]));
      
      bondAngle[i] = adata[nextAtomNum].angle[0];
      torAngle[i]  = adata[nextAtomNum].angle[1];
      //cout << " ligand_fineSampling.C:findAngles i = " << i << " torAngle[i] = " << torAngle[i] << " ";
    }
  }
  cout << endl;
  
  for (i = 0; i < numBonds; i++) {
    if ((bonds[i]->numBonds > 0) && (bonds[i]->colorBit == 0)) 
      {      
	bonds[i]->findAngles(adata, atomNum, back1);
      }
  }
}

void ligAtom::addBond(ligAtom *mybond) {
  
  bonds[numBonds] = mybond;
  numBonds++;

}



void ligAtom::linkTree() {

  int i, tempbond = 0;

  for (i = 0; i < globalNumConnects; i++) {
      
    if (globalConnections[i][2] == 0) {

      if (globalConnections[i][0] == atomNum) {     
	//cout << atomNum << " " << globalConnections[i][1] << endl;
	globalConnections[i][2] = 1;
	tempbond = globalConnections[i][1];

	if (globalLigAtomPtrs[tempbond] == NULL) {
	  globalLigAtomPtrs[tempbond] = new ligAtom;
	  globalLigAtomPtrs[tempbond]->atomNum = tempbond;
	  addBond(globalLigAtomPtrs[tempbond]);
	  globalLigAtomPtrs[tempbond]->addBond(this);
	  globalLigAtomPtrs[tempbond]->linkTree();
	}

	else {
	  addBond(globalLigAtomPtrs[tempbond]);
	  globalLigAtomPtrs[tempbond]->addBond(this);
	}
      }
	
      else if (globalConnections[i][1] == atomNum) {
	//cout << atomNum << " " << globalConnections[i][0] << endl;
	globalConnections[i][2] = 1;
	tempbond = globalConnections[i][0];
	
	if (globalLigAtomPtrs[tempbond] == NULL) {
	  globalLigAtomPtrs[tempbond] = new ligAtom;
	  globalLigAtomPtrs[tempbond]->atomNum = tempbond;
	  addBond(globalLigAtomPtrs[tempbond]);
	  globalLigAtomPtrs[tempbond]->addBond(this);
	  globalLigAtomPtrs[tempbond]->linkTree();
	}
	  
	else {
	  addBond(globalLigAtomPtrs[tempbond]);
	  globalLigAtomPtrs[tempbond]->addBond(this);
	}
      }

    }
  }

  if ((bondLength == NULL) && (bondAngle == NULL) && (torAngle == NULL)) {
    bondLength = new float[numBonds];
    bondAngle  = new float[numBonds];
    torAngle   = new float[numBonds];
  }

  //cout << "atom num = " << atomNum << " num bonds = " << numBonds << endl;

}  
      
void ligAtom::clearColors() {
  
  int i;

  for (i = 0; i < globalNumAtoms; i++) {
    if (globalLigAtomPtrs[i] == NULL)
      cout << "global lig atom ptr # " << i << " is NULL." << endl;
  }

  for (i = 0; i < globalNumAtoms; i++) {
    assert (globalLigAtomPtrs[i] != NULL);
    globalLigAtomPtrs[i]->colorBit = 0;
  }
}

///
void ligAtom::readLigFile(const char *fname) {
  
  FILE* infile; 
  char record[10]; 
  char line[200]; 
  char temp[100]; 
  int i, foundone, validstart, validend;
  int startbond, endbond;
  int tempPdbNum[200];
  VEC tempAtom[200];
  int tempRigid[200];
  float tempCharge[200];

  infile = fopen(fname, "r"); 

  while (fgets(line, 200, infile)) { 
    
    extStr(record, line, 1, 6); 
    
    if (strcmp(record, "HETATM") == 0) { 
      
      extStr(temp, line, 7, 11); 
      sscanf(temp, "%d", &(tempPdbNum[globalNumAtoms])); 
      
      extStr(temp, line, 31, 38); 
      sscanf(temp, "%f", &(tempAtom[globalNumAtoms][0])); 
      
      extStr(temp, line, 39, 46); 
      sscanf(temp, "%f", &(tempAtom[globalNumAtoms][1])); 
      
      extStr(temp, line, 47, 54); 
      sscanf(temp, "%f", &(tempAtom[globalNumAtoms][2])); 

      extStr(temp, line, 55, 60); 
      sscanf(temp, "%f", &(tempCharge[globalNumAtoms])); 
      
      tempRigid[globalNumAtoms] = 0;

      globalNumAtoms++; 
      
    } 
    else if (strcmp(record, "ATOM  ") == 0) { 
      
      extStr(temp, line, 7, 11); 
      sscanf(temp, "%d", &(tempPdbNum[globalNumAtoms])); 
      
      extStr(temp, line, 31, 38); 
      sscanf(temp, "%f", &(tempAtom[globalNumAtoms][0])); 
      
      extStr(temp, line, 39, 46); 
      sscanf(temp, "%f", &(tempAtom[globalNumAtoms][1])); 
      
      extStr(temp, line, 47, 54); 
      sscanf(temp, "%f", &(tempAtom[globalNumAtoms][2])); 
      
      extStr(temp, line, 55, 60); 
      sscanf(temp, "%f", &(tempCharge[globalNumAtoms])); 

      tempRigid[globalNumAtoms] = 0;

      globalNumAtoms++; 
      
    } 
    else if (strcmp(record, "RIGATM") == 0) { 
      
      extStr(temp, line, 7, 11); 
      sscanf(temp, "%d", &(tempPdbNum[globalNumAtoms])); 
      
      extStr(temp, line, 31, 38); 
      sscanf(temp, "%f", &(tempAtom[globalNumAtoms][0])); 
      
      extStr(temp, line, 39, 46); 
      sscanf(temp, "%f", &(tempAtom[globalNumAtoms][1])); 
      
      extStr(temp, line, 47, 54); 
      sscanf(temp, "%f", &(tempAtom[globalNumAtoms][2])); 
      
      extStr(temp, line, 55, 60); 
      sscanf(temp, "%f", &(tempCharge[globalNumAtoms])); 

      tempRigid[globalNumAtoms] = 1;

      globalNumAtoms++; 
      
    } 
    
    else if (strcmp(record, "CONECT") == 0) { 
      
      extStr(temp, line, 7, 11); 
      sscanf(temp, "%d", &startbond); 
      
      validstart = 0;
      for (i = 0; i < globalNumAtoms; i++) {
	if (startbond == tempPdbNum[i]) {
	  startbond = i;
	  validstart = 1;
	  break;
	}
      }

      if (validstart == 1) {
	extStr(temp, line, 12, 16); 
	if (sscanf(temp, "%d", &endbond) == 1) {
	
	  validend = 0;
	  for (i = 0; i < globalNumAtoms; i++) {
	    if (endbond == tempPdbNum[i]) {
	      endbond = i;
	      validend = 1;
	      break;
	    }
	  }

	  if (validend == 1) {

	    foundone = 0;
	    
	    for (i = 0; i < globalNumConnects; i++) {
	      if (((globalConnections[i][0] == startbond) && (globalConnections[i][1] == endbond)) || ((globalConnections[i][0] == endbond) && (globalConnections[i][1] == startbond))) {
		foundone = 1;
		break;
	      }
	    }
	    
	    if (foundone == 0) {
	      globalConnections[globalNumConnects][0] = startbond;
	      globalConnections[globalNumConnects][1] = endbond;
	      globalConnections[globalNumConnects][2] = 0;
	      //cout << "want to connect " << tempPdbNum[startbond] << " to : " << tempPdbNum[endbond] << endl;
	      globalNumConnects++;
	    }	  
	  }
	}
      
      
	extStr(temp, line, 17, 21); 
	if (sscanf(temp, "%d", &endbond) == 1) {
	
	  validend = 0;
	  for (i = 0; i < globalNumAtoms; i++) {
	    if (endbond == tempPdbNum[i]) {
	      endbond = i;
	      validend = 1;
	      break;
	    }
	  }

	  if (validend == 1) {

	    foundone = 0;
	    
	    for (i = 0; i < globalNumConnects; i++) {
	      if (((globalConnections[i][0] == startbond) && (globalConnections[i][1] == endbond)) || ((globalConnections[i][0] == endbond) && (globalConnections[i][1] == startbond))) {
		foundone = 1;
		break;
	      }
	    }
	    
	    if (foundone == 0) {
	      //	      cout << "want to connect " << tempPdbNum[startbond] << " to : " << tempPdbNum[endbond] << endl;
	      globalConnections[globalNumConnects][0] = startbond;
	      globalConnections[globalNumConnects][1] = endbond;
	      globalConnections[globalNumConnects][2] = 0;
	      globalNumConnects++;
	    }	  
	  }
	}
      
	extStr(temp, line, 22, 26); 
	if (sscanf(temp, "%d", &endbond) == 1) {
	
	  validend = 0;
	  for (i = 0; i < globalNumAtoms; i++) {
	    if (endbond == tempPdbNum[i]) {
	      endbond = i;
	      validend = 1;
	      break;
	    }
	  }
	  
	  if (validend == 1) {

	    foundone = 0;
	    
	    for (i = 0; i < globalNumConnects; i++) {
	      if (((globalConnections[i][0] == startbond) && (globalConnections[i][1] == endbond)) || ((globalConnections[i][0] == endbond) && (globalConnections[i][1] == startbond))) {
		foundone = 1;
		break;
	      }
	    }
	    
	    if (foundone == 0) {
	      //	      cout << "want to connect " << tempPdbNum[startbond] << " to : " << tempPdbNum[endbond] << endl;
	      globalConnections[globalNumConnects][0] = startbond;
	      globalConnections[globalNumConnects][1] = endbond;
	      globalConnections[globalNumConnects][2] = 0;
	      globalNumConnects++;
	    }	  
	  }
	}
      
	extStr(temp, line, 27, 31); 
	if (sscanf(temp, "%d", &endbond) == 1) {
	
	  validend = 0;
	  for (i = 0; i < globalNumAtoms; i++) {
	    if (endbond == tempPdbNum[i]) {
	      endbond = i;
	      validend = 1;
	      break;
	    }
	  }
	  
	  if (validend == 1) {

	    foundone = 0;
	    
	    for (i = 0; i < globalNumConnects; i++) {
	      if (((globalConnections[i][0] == startbond) && (globalConnections[i][1] == endbond)) || ((globalConnections[i][0] == endbond) && (globalConnections[i][1] == startbond))) {
		foundone = 1;
		break;
	      }
	    }
	    
	    if (foundone == 0) {
	      //	      cout << "want to connect " << tempPdbNum[startbond] << " to : " << tempPdbNum[endbond] << endl;
	      globalConnections[globalNumConnects][0] = startbond;
	      globalConnections[globalNumConnects][1] = endbond;
	      globalConnections[globalNumConnects][2] = 0;
	      globalNumConnects++;
	    }	  
	  }      
	}
      }
    }     
  }

  cout << " globalNumAtoms = " << globalNumAtoms << " globalNumConnects = " << globalNumConnects << endl;

  origAtomData = new atomData[globalNumAtoms];
  globalLigAtomPtrs = (ligAtom**)(new ligAtom*[globalNumAtoms]);
  globalLigAtomCharge = new float[globalNumAtoms];
  globalNumNonBondAtoms = new int[globalNumAtoms];
  globalRigidAtoms = new int[globalNumAtoms];
  globalNonBondAtoms = new int*[globalNumAtoms];

  for (i = 0; i < globalNumAtoms; i++) {

    if (tempAtom[i][0] < MIN_X) 
      MIN_X = tempAtom[i][0];
    if (tempAtom[i][1] < MIN_Y) 
      MIN_Y = tempAtom[i][1];
    if (tempAtom[i][2] < MIN_Z) 
      MIN_Z = tempAtom[i][2];

    if (tempAtom[i][0] > MAX_X) 
      MAX_X = tempAtom[i][0];
    if (tempAtom[i][1] > MAX_Y) 
      MAX_Y = tempAtom[i][1];
    if (tempAtom[i][2] > MAX_Z) 
      MAX_Z = tempAtom[i][2];


    origAtomData[i].xyz[0] = tempAtom[i][0];
    origAtomData[i].xyz[1] = tempAtom[i][1];
    origAtomData[i].xyz[2] = tempAtom[i][2];
    origAtomData[i].angle[0] = -10.0;
    origAtomData[i].angle[1] = -10.0;

    globalLigAtomCharge[i] = tempCharge[i];
    globalRigidAtoms[i] = tempRigid[i];

    globalLigAtomPtrs[i] = NULL;

    globalNumNonBondAtoms[i] = 0;
    globalNonBondAtoms[i] = new int[globalNumAtoms];

  }


  if (globalNumConnects > 0) {
    /*
    cout << "printing globalConnections array. " << endl;
    for (i = 0; i < globalNumConnects; i++)
      {

	cout << globalConnections[i][0] << " ";
	cout << globalConnections[i][1] << " ";
	cout << globalConnections[i][2] << endl;
      }
    */

    globalRootAtom = globalConnections[0][0];
    globalLigAtomPtrs[globalRootAtom] = this;
    atomNum = globalRootAtom;
    
    linkTree();
  }
  cout << "in ligand_fineSampling.C::readLigFile..." << endl;
  cout << "MIN_X = " << MIN_X << " MAX_X = " << MAX_X << endl;
  cout << "MIN_Y = " << MIN_Y << " MAX_Y = " << MAX_Y << endl;
  cout << "MIN_Z = " << MIN_Z << " MAX_Z = " << MAX_Z << endl;
  
}


void ligAtom::findNumDof_root() {

  clearColors();

  colorBit = 1;

  globalDOFArray[0] = atomNum;
  globalDOFArray[1] = atomNum;
  globalDOFArray[2] = atomNum;
  globalDOFArray[3] = atomNum;
  globalDOFArray[4] = atomNum;

  globalNumDOF = 5;

  if (bonds[0]->numBonds > 0) {   
    bonds[0]->findNumDof(atomNum);
  }
}


void ligAtom::findNumDof(int back1) {

  int i; 

  colorBit = 1;

  if ((globalRigidAtoms[atomNum] == 1) && (globalRigidAtoms[back1] == 1)) {
  }
  else {
    globalDOFArray[globalNumDOF] = atomNum;
    globalNumDOF++;
  }
    
  for (i = 0; i < numBonds; i++) {
    
    if ((bonds[i]->numBonds > 1) && (bonds[i]->colorBit == 0)) {      
      bonds[i]->findNumDof(atomNum);
    }
  }

}


void ligAtom::createRandomLigand_root(atomData* adata) {

  int nextAtomNum;

  clearColors();

  colorBit = 1;

  adata[atomNum].xyz[0] = 0.0;//(((float)(random()))/((float)(MY_RAND_MAX)))*(MAX_XYZ);
  adata[atomNum].xyz[1] = 0.0;//(((float)(random()))/((float)(MY_RAND_MAX)))*(MAX_XYZ);
  adata[atomNum].xyz[2] = 0.0;//(((float)(random()))/((float)(MY_RAND_MAX)))*(MAX_XYZ);

  nextAtomNum = bonds[0]->atomNum;

  adata[nextAtomNum].angle[0] = (((float)(random()))/((float)(MY_RAND_MAX)))*PI;
  adata[nextAtomNum].angle[1] = (((float)(random()))/((float)(MY_RAND_MAX)))*(2.0*PI) - PI;


  if (bonds[0]->numBonds > 0) {   
    bonds[0]->createRandomLigand(adata, atomNum);
  }
}


void ligAtom::createRandomLigand(atomData* adata, int back1) {

  int nextAtomNum, i; 
  float count = 0.0;
  float randphi;

  colorBit = 1;

  if ((globalRigidAtoms[atomNum] == 1) && (globalRigidAtoms[back1] == 1)) {
    randphi = 0.0;
  }
  else {
    randphi = (((float)(random()))/((float)(MY_RAND_MAX)))*(2.0*PI) - PI;
  }

  for (i = 0; i < numBonds; i++) {

    nextAtomNum = bonds[i]->atomNum;

    if (bonds[i]->colorBit == 0) {
      
      adata[nextAtomNum].angle[0] = bondAngle[i];
      adata[nextAtomNum].angle[1] = torAngle[i] + randphi;

      if (adata[nextAtomNum].angle[1] > PI) {
	adata[nextAtomNum].angle[1] -= 2.0 * PI;
      }
      else if (adata[nextAtomNum].angle[1] < -PI) {
	adata[nextAtomNum].angle[1] += 2.0 * PI;
      }
    }

  }
    
  for (i = 0; i < numBonds; i++) {
    
    if ((bonds[i]->numBonds > 0) && (bonds[i]->colorBit == 0)) {      
      bonds[i]->createRandomLigand(adata, atomNum);
    }
  }

}



void ligAtom::createRandomLigFromLig_root(atomData* adata, atomData *orig) {

  int nextAtomNum;
  float randfloat;

  clearColors();
  
  colorBit = 1;

  adata[atomNum].xyz[0] = orig[atomNum].xyz[0] + (((float)(random()))/((float)(MY_RAND_MAX)))*(2.0 * XYZ_EPSILON) - XYZ_EPSILON;
  adata[atomNum].xyz[1] = orig[atomNum].xyz[1] + (((float)(random()))/((float)(MY_RAND_MAX)))*(2.0 * XYZ_EPSILON) - XYZ_EPSILON;
  adata[atomNum].xyz[2] = orig[atomNum].xyz[2] + (((float)(random()))/((float)(MY_RAND_MAX)))*(2.0 * XYZ_EPSILON) - XYZ_EPSILON;



  if (adata[atomNum].xyz[0] < 0.0) {
    adata[atomNum].xyz[0] = 0.1;
  }
  else if (adata[atomNum].xyz[0] > MAX_XYZ) {
    adata[atomNum].xyz[0] = MAX_XYZ - 0.1;
  }

  if (adata[atomNum].xyz[1] < 0.0) {
    adata[atomNum].xyz[1] = 0.1;
  }
  else if (adata[atomNum].xyz[1] > MAX_XYZ) {
    adata[atomNum].xyz[1] = MAX_XYZ - 0.1;
  }

  if (adata[atomNum].xyz[2] < 0.0) {
    adata[atomNum].xyz[2] = 0.1;
  }
  else if (adata[atomNum].xyz[2] > MAX_XYZ) {
    adata[atomNum].xyz[2] = MAX_XYZ - 0.1;
  }


  nextAtomNum = bonds[0]->atomNum;

  adata[nextAtomNum].angle[0] = orig[nextAtomNum].angle[0] + (((float)(random()))/((float)(MY_RAND_MAX)))*(2.0 * ANGLE_EPSILON) - ANGLE_EPSILON;
  adata[nextAtomNum].angle[1] = orig[nextAtomNum].angle[1] + (((float)(random()))/((float)(MY_RAND_MAX)))*(2.0 * ANGLE_EPSILON) - ANGLE_EPSILON;

  if (adata[nextAtomNum].angle[0] > PI) {
    adata[nextAtomNum].angle[0] = PI - 0.1;
  }
  else if (adata[nextAtomNum].angle[0] < 0.0) {
    adata[nextAtomNum].angle[0] = 0.1;
  }

  if (adata[nextAtomNum].angle[1] > PI) {
    adata[nextAtomNum].angle[1] -= 2.0 * PI;
  }
  else if (adata[nextAtomNum].angle[1] < -PI) {
    adata[nextAtomNum].angle[1] += 2.0 * PI;
  }



  if (bonds[0]->numBonds > 0) {   
    bonds[0]->createRandomLigFromLig(adata, atomNum, orig);
  }
}


void ligAtom::createRandomLigFromLig(atomData* adata, int back1, atomData *orig) {

  int nextAtomNum, i; 
  float randphi;

  colorBit = 1;

  if ((globalRigidAtoms[atomNum] == 1) && (globalRigidAtoms[back1] == 1)) {
    randphi = 0.0;
  }
  else {
    randphi = (((float)(random()))/((float)(MY_RAND_MAX)))*(2.0*ANGLE_EPSILON) - ANGLE_EPSILON;
  }

  for (i = 0; i < numBonds; i++) {

    nextAtomNum = bonds[i]->atomNum;

    if (bonds[i]->colorBit == 0) {
      
      adata[nextAtomNum].angle[0] = bondAngle[i];
      adata[nextAtomNum].angle[1] = torAngle[i] + randphi;

      if (adata[nextAtomNum].angle[1] > PI) {
	adata[nextAtomNum].angle[1] -= 2.0 * PI;
      }
      else if (adata[nextAtomNum].angle[1] < -PI) {
	adata[nextAtomNum].angle[1] += 2.0 * PI;
      }
    }

  }
    
  for (i = 0; i < numBonds; i++) {
    
    if ((bonds[i]->numBonds > 0) && (bonds[i]->colorBit == 0)) {      
      bonds[i]->createRandomLigFromLig(adata, atomNum, orig);
    }
  }

}

  
 

#if 0

void ligAtom::createRandomLigand(atomData* adata) {

  atomData* retvalue;
  int i;

  retvalue = new atomData[globalNumAtoms];

  for (i = 0; i < globalNumAtoms; i++) {

    if (globalAtomDepth[i] > 2) {
      retvalue[i].angle[0] = (((float)(random()))/((float)(MY_RAND_MAX)))*(2.0*THETA_MAX) - THETA_MAX;
      retvalue[i].angle[1] = (((float)(random()))/((float)(MY_RAND_MAX)))*(2.0*PHI_MAX) - PHI_MAX;
    }      
    else if (globalAtomDepth[i] == 0) {

      //      retvalue[i].xyz[0] = (((float)(random()))/((float)(MY_RAND_MAX)))*((float)(X_MAX - X_MIN)) + ((float)(X_MIN));
      //      retvalue[i].xyz[1] = (((float)(random()))/((float)(MY_RAND_MAX)))*((float)(Y_MAX - Y_MIN)) + ((float)(Y_MIN));
      //      retvalue[i].xyz[2] = (((float)(random()))/((float)(MY_RAND_MAX)))*((float)(Z_MAX - Z_MIN)) + ((float)(Z_MIN));


      retvalue[i].xyz[0] = (((float)(random()))/((float)(MY_RAND_MAX)))*(MAX_XYZ);
      retvalue[i].xyz[1] = (((float)(random()))/((float)(MY_RAND_MAX)))*(MAX_XYZ);
      retvalue[i].xyz[2] = (((float)(random()))/((float)(MY_RAND_MAX)))*(MAX_XYZ);

    }

    else if (globalAtomDepth[i] == 1) {
      retvalue[i].angle[0] = (((float)(random()))/((float)(MY_RAND_MAX)))*(2.0*PI) - PI;
      retvalue[i].angle[1] = (((float)(random()))/((float)(MY_RAND_MAX)))*(2.0*PI) - PI;
    }
    else if (globalAtomDepth[i] == 2) {
      retvalue[i].angle[0] = (((float)(random()))/((float)(MY_RAND_MAX)))*(2.0*THETA_MAX) - THETA_MAX;
      retvalue[i].angle[1] = (((float)(random()))/((float)(MY_RAND_MAX)))*(2.0*PI) - PI;
    }
    else {
      cout << "depth is incorrect\n";
    }

  }

  myLigand->findXYZ_root(retvalue);

  return retvalue;

}

#endif

void copyDegreesFromAtomData(atomData *out, atomData *in) {

  int i;

  for (i = 0; i < globalNumAtoms; i++) {

    if (i == globalRootAtom) {
      
      out[i].xyz[0] = in[i].xyz[0];
      out[i].xyz[1] = in[i].xyz[1];
      out[i].xyz[2] = in[i].xyz[2];

      out[i].angle[0] = -10.0;
      out[i].angle[1] = -10.0;

    }
    
    else {
      out[i].angle[0] = in[i].angle[0];
      out[i].angle[1] = in[i].angle[1];

      out[i].xyz[0] = 0.0;
      out[i].xyz[1] = 0.0;
      out[i].xyz[2] = 0.0;
      
    }

  }  
}


void printAtomData(atomData *adata) {

  int i;

  cout << "Root Atom: " << globalRootAtom << endl;

  for (i = 0; i < globalNumAtoms; i++) {

    cout << adata[i].xyz[0] << "\t" << adata[i].xyz[1] << "\t" << adata[i].xyz[2] << "\t" << adata[i].angle[0] << "\t" << adata[i].angle[1] << endl;

  }

  cout << endl;

}
    
ligAtom* ligAtom::findGoodRoot(ligAtom* prevAtom) {

  int i;
  ligAtom* retval = NULL;

  if (prevAtom == NULL) {
    clearColors();
  }

  colorBit = 1;

  if (numBonds == 1) {
    globalRootAtom = atomNum;
    return this;
  }
  else {
    for (i = 0; ((i < numBonds) && (retval == NULL)); i++) {
      if (bonds[i]->colorBit == 0) {
	retval = bonds[i]->findGoodRoot(this);
      }
    }
    return retval;
  }  

}

void findNonBondedPairs() {

  int i, j, k, kstop, l, lstop;
  int *tempArray;

  tempArray = new int[globalNumAtoms];


  for (i = 0; i < globalNumAtoms; i++) {
    for (j = 0; j < globalNumAtoms; j++) {
      tempArray[j] = 0;
    }

    kstop = globalLigAtomPtrs[i]->numBonds;
    for (k = 0; k < kstop; k++) {
      tempArray[globalLigAtomPtrs[i]->bonds[k]->atomNum] = 1;
      lstop = globalLigAtomPtrs[i]->bonds[k]->numBonds;
      for (l = 0; l < lstop; l++) {
	tempArray[globalLigAtomPtrs[i]->bonds[k]->bonds[l]->atomNum] = 1;
      }
    }

    for (j = i + 1; j < globalNumAtoms; j++) {
      if (tempArray[j] == 0) {	
	globalNonBondAtoms[i][globalNumNonBondAtoms[i]] = j;
	globalNumNonBondAtoms[i]++;
      }
    }
  }

  /*  for (i = 0; i < globalNumAtoms; i++) {
    cout << "Atom " << i << " ";
    for (j = 0; j < globalNumNonBondAtoms[i]; j++) {
      cout << globalNonBondAtoms[i][j] << " ";
    }
    cout << endl;
  }	*/			


}
    


    


    
void ligAtom::fixLigand() {

  int i;
  //  atomData *tempAtomData;

  //  tempAtomData = new atomData[globalNumAtoms];
  
  globalNumPLigCharge = 0.0;
  globalNumNLigCharge = 0.0;
  
  for (i = 0; i < globalNumAtoms; i++) {

    origAtomData[i].xyz[0] += MIN_X;
    origAtomData[i].xyz[1] += MIN_Y;
    origAtomData[i].xyz[2] += MIN_Z;

    if (globalLigAtomCharge[i] > 0) {
      globalNumPLigCharge += globalLigAtomCharge[i];
    }
    else if (globalLigAtomCharge[i] < 0) {
      globalNumNLigCharge += -globalLigAtomCharge[i];
    }

  }

  findAngles_root(origAtomData);

  findNumDof_root();


  //  printAtomData(origAtomData);

  //  1ldm:
  //  globalLigAtomCharge[4] = -0.5;
  //  globalLigAtomCharge[5] = -0.5;
  //  globalNumNLigCharge = 1;
  //  globalNumPLigCharge = 0;


  //  3tpi:
  //  globalLigAtomCharge[11] = -0.5;
  //  globalLigAtomCharge[15] = -0.5;
  //  globalLigAtomCharge[0] = 0.8;  
  //  globalNumNLigCharge = 1.0;
  //  globalNumPLigCharge = 0.8;
  

  //  4tln:
  //  globalLigAtomCharge[0] = 1.0;
  //  globalLigAtomCharge[3] = -1.0;
  //  globalNumNLigCharge = 1.0;
  //  globalNumPLigCharge = 1.0;

}


void moveOneDOF(atomData *adata, int dofToMove, float direction) {

  int ta;
  int i, nextAtomNum;

  ta = globalDOFArray[dofToMove];

  if (dofToMove < 3) {
    adata[ta].xyz[dofToMove] = adata[ta].xyz[dofToMove] + direction * XYZ_UNIT_STEP;
  }

  else if (dofToMove == 3)  {
    
    nextAtomNum = globalLigAtomPtrs[ta]->bonds[0]->atomNum;

    adata[nextAtomNum].angle[0] = adata[nextAtomNum].angle[0] + direction * ANGLE_UNIT_STEP;

    if (adata[nextAtomNum].angle[0] > PI) {
      adata[nextAtomNum].angle[0] = PI - 0.1;
    }
    else if (adata[nextAtomNum].angle[0] < 0.0) {
      adata[nextAtomNum].angle[0] = 0.1;
    }
  }

  else if (dofToMove == 4) {
    
    nextAtomNum = globalLigAtomPtrs[ta]->bonds[0]->atomNum;

    adata[nextAtomNum].angle[1] = adata[nextAtomNum].angle[1] + direction * ANGLE_UNIT_STEP;

    if (adata[nextAtomNum].angle[1] > PI) {
      adata[nextAtomNum].angle[1] -= 2.0 * PI;
    }
    else if (adata[nextAtomNum].angle[1] < -PI) {
      adata[nextAtomNum].angle[1] += 2.0 * PI;
    }   
  }
  else {
    for (i = 0; i < globalLigAtomPtrs[ta]->numBonds; i++) {

      nextAtomNum = globalLigAtomPtrs[ta]->bonds[i]->atomNum;      
      adata[nextAtomNum].angle[1] = adata[nextAtomNum].angle[1] + direction * ANGLE_UNIT_STEP;
      if (adata[nextAtomNum].angle[1] > PI) {
	adata[nextAtomNum].angle[1] -= 2.0 * PI;
      }
      else if (adata[nextAtomNum].angle[1] < -PI) {
	adata[nextAtomNum].angle[1] += 2.0 * PI;
      }   
    }
  }
}


/* sets cog of node */ 
void computeCOG(nodeData * node)
{
  VEC tvec;
  
  tvec[0] = 0.0;
  tvec[1] = 0.0;
  tvec[2] = 0.0;

  for (int i = 0; i < globalNumAtoms; i++) {
    
    tvec[0] = tvec[0] + node->myData[i].xyz[0];
    tvec[1] = tvec[1] + node->myData[i].xyz[1];
    tvec[2] = tvec[2] + node->myData[i].xyz[2];

  }

  node->cog[0] = tvec[0]/((float)(globalNumAtoms));
  node->cog[1] = tvec[1]/((float)(globalNumAtoms));
  node->cog[2] = tvec[2]/((float)(globalNumAtoms));
}

/* samples within samplingRadius of the origin. */
void sampleCOG(VEC retval, float samplingRadius, VEC origin)
{

  float  randomNumber, 
    radius, // actual radius
    x,y,z,temp,sinTheta, cosTheta,phi;
  
  /* first sample the radius */
  randomNumber = (((float)(random()))/((float)(MY_RAND_MAX)));

  radius = randomNumber * samplingRadius;
  
  /* second sample cos theta, spherical coords */
  
  
  randomNumber = 2 * (((float)(random()))/((float)(MY_RAND_MAX))) - 1;

  cosTheta = randomNumber;
  
  z = radius * cosTheta;
  
  /* unfortunately taking sqrt to find sin theta */
  
  sinTheta = sqrt(1 - cosTheta * cosTheta);

  /* third, sampling phi, spherical coords */
  
  randomNumber = M_PI * (2 * ((float)(random()))/((float)(MY_RAND_MAX)) - 1);

  phi = randomNumber;

  temp = radius * sinTheta;
  
  y    = temp   * sin(phi);
  x    = temp   * cos(phi);
 
  retval[0] = origin[0] + x;
  retval[1] = origin[1] + y;
  retval[2] = origin[2] + z;

  for (int i = 0; i < 3; i++)
    {
      if (retval[i] < 0.0) {
	retval[i] = 0.1;
      }
      else if (retval[i] > MAX_XYZ) {
	retval[i] = MAX_XYZ - 0.1;
      }
    }
  if (distanceBetweenVecs(origin,retval) >=  samplingRadius)
    {
      cout << "distance between vecs = " << distanceBetweenVecs(origin,retval) << endl;
      assert (distanceBetweenVecs(origin,retval) < samplingRadius);
    }
}
